//
// QSINIT
// base type
// -------------------------------
// note that this used *everywhere* (in 16/32/64-bits)
//
#ifndef QS_BASE_TYPES
#define QS_BASE_TYPES

#ifdef DOXYGEN_GEN
#include "qsxdoc.h"
#endif

#ifndef NULL
#define NULL   (0)
#endif

#if defined(__WATCOMC__) || defined(_MSC_EXTENSIONS)
typedef unsigned __int64    u64t;
typedef __int64             s64t;
#else
typedef unsigned long long  u64t;
typedef long long           s64t;
#endif
#ifndef EFIX64
typedef unsigned long       u32t;
typedef long                s32t;
#ifdef __QSINIT__
typedef s32t           ptrdiff_t;
#endif
#else
#if _MSC_EXTENSIONS
typedef unsigned __int32    u32t;
typedef __int32             s32t;
#else
typedef unsigned int        u32t;
typedef int                 s32t;
#endif
typedef s64t           ptrdiff_t;
#endif // EFIX64
typedef unsigned short      u16t;
typedef short               s16t;
typedef unsigned char        u8t;
typedef signed char          s8t;
typedef void              *pvoid;
typedef u32t               qserr;
typedef u32t            qshandle;
typedef u64t             qsclock;

typedef u16t                c16t;

#define PAGESHIFT    (12)
#define PAGESIZE     (0x1000)
#define PAGEMASK     (0x0FFF)
#define PARASHIFT    (4)

#define QS_MAXPATH   (260)

#define   _1KB       (  1UL* 1024)       // 1 kb
#define   _2KB       (  2UL* _1KB)       // 2 kb
#define   _4KB       (  4UL* _1KB)       // 4 kb
#define   _8KB       (  8UL* _1KB)       // 8 kb
#define  _16KB       ( 16UL* _1KB)       // 32 kb
#define  _32KB       ( 32UL* _1KB)       // 32 kb
#define  _64KB       ( 64UL* _1KB)       // 64 kb
#define _128KB       (  2UL*_64KB)       // 128 kb
#define _256KB       (  4UL*_64KB)       // 256 kb
#define _512KB       (  8UL*_64KB)       // 512 kb
#define _640KB       ( 10UL*_64KB)       // 640 kb
#define   _1MB       ( _1KB* _1KB)       // 1 mb
#define   _2MB       (  2UL* _1MB)       // 2 mb
#define   _4MB       (  4UL* _1MB)       // 4 mb
#define   _8MB       (  8UL* _1MB)       // 8 mb
#define  _15MB       ( 15UL* _1MB)       // 15 mb
#define  _16MB       ( 16UL* _1MB)       // 16 mb
#define  _32MB       ( 32UL* _1MB)       // 32 mb
#define  _64MB       ( 64UL* _1MB)       // 64 mb
#define _512MB       (512UL* _1MB)       // 512 mb
#define   _1GB       ( 16UL*_64MB)       // 1 gb
#define   _2GB       ( 32UL*_64MB)       // 2 gb
#define   _4GB       (  0UL)             // 4 gb (32 bit)

#define FFFF         (0xFFFFFFFF)
#define x7FFF        (0x7FFFFFFF)
#define x7FFF64      (((s64t)x7FFF<<32)+FFFF)
#define FFFF64       (((u64t)FFFF<<32)+FFFF)
#define _4GBLL       ((u64t)FFFF+1)      // 4 gb (64 bit)
#define _2TBLL       ((u64t)FFFF+1<<9)   // 2 tb (64 bit)

#define PAGEROUND(p)    (((p) + PAGEMASK) & ~PAGEMASK)
#define PAGEROUNDLL(p)  (((p) + PAGEMASK) & ~(u64t)PAGEMASK)

#define Round2(ii)   (((ii)+1)&0xFFFFFFFE)
#define Round4(ii)   (((ii)+3)&0xFFFFFFFC)
#define Round8(ii)   (((ii)+7)&0xFFFFFFF8)
#define Round16(ii)  (((ii)+0x0F)&0xFFFFFFF0)
#define Round32(ii)  (((ii)+0x1F)&0xFFFFFFE0)
#define Round64(ii)  (((ii)+0x3F)&0xFFFFFFC0)
#define Round128(ii) (((ii)+0x7F)&0xFFFFFF80)
#define Round256(ii) (((ii)+0xFF)&0xFFFFFF00)
#define Round512(ii) (((ii)+0x1FF)&0xFFFFFE00)
#define Round1k(ii)  (((ii)+0x3FF)&0xFFFFFC00)
#define Round2k(ii)  (((ii)+0x7FF)&0xFFFFF800)
#define Round4k(ii)  (((ii)+0xFFF)&0xFFFFF000)
#define Round64k(ii) (((ii)+0xFFFF)&0xFFFF0000)

#ifndef true
#define true  1
#endif
#ifndef false
#define false 0
#endif

#ifdef EFIX64
#define _std
#define _cdec
#else
#define _std  __stdcall
#define _cdec __cdecl
#endif
#ifdef __WATCOMC__
#define _fast __watcall
#else
#define _fast
#endif

#ifndef OS2_INCLUDED
// same decls exists int os2def.h
#define OFFSETOF(x)   (((u16t*)&(x))[0])
#define SELECTOROF(x) (((u16t*)&(x))[1])
#define MAKEFAR16(seg,ofs) ((u32t)(seg)<<16|(ofs))
#define MAKEFAR32(sel,ofs) ((u64t)(sel)<<32|(ofs))
#define MAKEID4(c1,c2,c3,c4)  ((u32t)((s8t)(c4))<<24|(u32t)((s8t)(c3))<<16|(u32t)((s8t)(c2))<<8|((s8t)(c1)))
#define FIELDOFFSET(type,field) ((u32t)&(((type*)0)->field))
#define BYTEOF(p,i) (((u8t*)&(p))[i])
#define WORDOF(p,i) (((u16t*)&(p))[i])
#define DWORDOF(p,i) (((u32t*)&(p))[i])
#define TRUE   (1)
#define FALSE  (0)
#endif // OS2_INCLUDED

#define Xor(v1,v2)    (((v1)?1:0)^((v2)?1:0))
#define SwapInt(x,y)  ((x)^=(y)^=(x)^=(y))

#endif // QS_BASE_TYPES
