//
// QSINIT
// start 32bit code
//
#include "stdlib.h"             // only for hlp_memcpy() flags
#include "qslocal.h"
#include "qsconst.h"
#include "qsbinfmt.h"
#include "vio.h"

#ifndef EFI_BUILD
#include "ioint13.h"
#include "parttab.h"
#include "filetab.h"
u32t                  qs16base, // base addr of 16-bit part
                     qs16basev; // the same, but fixuped in pagemode.c
extern
MKBIN_HEADER        bin_header;
extern struct Disk_BPB BootBPB; // boot disk BPB
extern u16t          pmdataseg, // dpmi service data (GDT, callbacks, etc)
                      rm16code;
extern u32t          stacksize; // 32-bit part stack size
extern
struct filetable_s   filetable;
static void       *cmdhist_ptr = 0;
static u32t        cmdhist_len;
char                 *rest_ini = 0;
void                 *init_env = 0; // EFI build has this var in the table
u8t                 forcedBMnt = BVMOUNT_AUTO;
#else
extern u32t        qd_bootdisk;
#endif // EFI_BUILD
u8t               bootio_avail; // boot partition disk i/o is available
extern
MKBIN_HEADER      *pbin_header;  // use pointer to be compatible with EFI build
extern u32t          phmembase;
extern u8t           pminitres; // result of PM init
extern u16t       puff_bufsize; // memory size need by puff.c
extern void          *puff_buf; // and pointer for it
extern u32t         *crc_table;
extern u32t         DiskBufPMF;
extern u16t    physmem_entries; // number of entries in physmem
extern physmem_block   physmem[PHYSMEM_TABLE_SIZE];
extern
cache_extptr      *cache_eproc; // cache module callbacks
u8t*                     ExCvt; // OEM case table (was used for FatFs, now for stricmp only)

#define CRC32_SIZE     (1024)
#define OEMTAB_SIZE     (128)

int  start_it(void);
void unpack_start_module(void);
void check_disks(void);
void get_ini_parm(void);
void make_crc_table(void *table);
void memmgr_init(void);
void hlp_finit(void);
void _std make_reboot(int warm);

void _std init_common(void) {
   // memmgr is ready, allocating memory for unzip tables and other misc stuff
   crc_table    = (u32t*)hlp_memallocsig(CRC32_SIZE + puff_bufsize + OEMTAB_SIZE,
                                         "serv", QSMA_READONLY);
   puff_buf     = (u8t*)crc_table + CRC32_SIZE;
   ExCvt        = (u8t*)puff_buf + puff_bufsize;
   make_crc_table(crc_table);
   push_fixup(&crc_table, 2, &puff_buf, &ExCvt);
   DiskBufPMF   = DiskBufPM;
#ifndef EFI_BUILD
   // saves mini-FSD crc32
   if (minifsd_ptr) {
      u32t crc = crc32(0, minifsd_ptr, filetable.ft_mfsdlen);
      sto_save(STOKEY_MFSDCRC, &crc, 4, 1);
   }
   /* read block, passed from the loader, who RESTARTs us.
      it should be placed in common memory, on 64k border (hlp_memalloc()) */
   if (transition_ptr && (transition_ptr&0xFFFF)==0) {
      // used just to range check here
      void *rb = hlp_memreserve(transition_ptr, sizeof(transition_data), 0);
      if (rb) {
         transition_data *td = (transition_data*)rb;
         while (td->sign==TRND_SIGN) {
            u32t  btype = td->type&~TRND_LAST;

            if (btype==0 || btype>TRND_CFGFILE) break; else {
               u32t crc = crc32(0, (u8t*)(td+1), td->length);

               if (crc==td->crc32) {
                  void *blk = 0;
               
                  switch (btype) {
                     case TRND_ENV    :
                        blk = hlp_memallocsig(td->length, "ienv", QSMA_READONLY|
                                              QSMA_RETERR|QSMA_NOCLEAR);
                        init_env = blk;
                        push_fixup(&init_env, 0);
                        break;
                     case TRND_CMDHIST:
                        blk = hlp_memalloc(td->length, QSMA_RETERR|QSMA_NOCLEAR);
                        if (blk) {
                           cmdhist_ptr = blk;
                           cmdhist_len = td->length;
                           push_fixup(&cmdhist_ptr, 0);
                        }
                        break;
                     case TRND_CFGFILE:
                        blk = hlp_memalloc(td->length, QSMA_RETERR|QSMA_NOCLEAR);
                        // there is no push_fixup() here!!! (INI uses phys addrs)
                        if (blk) rest_ini = blk;
                        break;
                     case TRND_CODEPAGE:
                        crc = *(u16t*)(td+1);
                        sto_save(STOKEY_RSTCPNUM, &crc, 4, 1);
                        break;
                     case TRND_FLAGS   : {
                        u32t flags = *(u32t*)(td+1);
                        if (flags & QRST_OSMODE) vmmode = 1; else
                           if (flags & QRST_PAE) vmmode = 2;
                        if (flags & QRST_BMNTMASK)
                           forcedBMnt = (flags&QRST_BMNTMASK) >> 2;
                        break;
                     }
                  }
                  if (blk) memcpy(blk, td+1, td->length);
               } else
                  break;
            }
            if (td->type&TRND_LAST) break;
            td = next_trblock(td);
         }
         hlp_memfree(rb);
      }
   }
#endif
   memset(ExCvt, '.', OEMTAB_SIZE);
}

// this should be called after init_pages() to save *new* pointers
static void push_storage(void) {
   u32t value = 1;
#ifndef EFI_BUILD
   if (cmdhist_ptr) sto_save(STOKEY_CMDHINIT, cmdhist_ptr, cmdhist_len, 0);
   if (lowmem) {
      sto_save(STOKEY_LOWMEM, &value, 4, 1);
      log_it(LOG_HIGH, "low memory mode!\n");
   }
#endif
   sto_save(STOKEY_PHYSMEM, &physmem, physmem_entries * sizeof(physmem_block), 0);
   sto_save(STOKEY_VOLDATA, extvol, sizeof(vol_data)*DISK_COUNT, 0);
   sto_save(STOKEY_BASEMEM, &phmembase, 4, 1);
   if (safeMode) sto_save(STOKEY_SAFEMODE, &value, 4, 1);
   sto_save(STOKEY_BOOTRW, &bootmount, 4, 1);
}

/// 32-bit main() function
#ifdef EFI_BUILD
void _std init32(void) {
   bootio_avail = qd_bootdisk!=FFFF;
   // pbin_header value filled by 64-bit code
#else
#pragma aux init32 parm [esi];
void _std init32(u32t rmstack) {
   // is boot volume available?
   bootio_avail = dd_bootflags&BF_NOMFSHVOLIO ? 0 : 1;
   pbin_header  = &bin_header;
   qs16base     = (u32t)rm16code<<PARASHIFT;
   qs16basev    = qs16base;
   // hello, world!
   if (pminitres) vio_strout("\nA20 error!\n"); else vio_charout('\n');
#endif // EFI_BUILD
   // init common memory manager
   memmgr_init();
   // init internal structs & save some keys for START module
   init_common();
   init_host();
   // init file i/o
   hlp_finit();
   // get some critical keys from .ini
   get_ini_parm();
   // restart option/EFI command line overrides the INI key
   if (forcedBMnt!=BVMOUNT_AUTO) bootmount = forcedBMnt;
#ifndef EFI_BUILD
   // just log it
   log_printf("stack for rm calls: %d bytes\n", rmstack);
   /* this better to be the last call in the function (it never returns),
      to prevent any possible address caching in registers below it */
   if (vmmode) init_pages();
#endif
   call_start();
}

/** continue of init32().
    Maked as "aborts;" so cannot use "return", jumped from init_pages() or
    the end of init32() here.

    This code should be a separate function because init_pages() above
    relocates ALL of this to another address, so better to drop any data
    in CPU registers by starting a new function. */
void call_start(void) {
   // save some keys only after switching to page mode!
   push_storage();
   // read and unpack QSINIT.LDI
   unpack_start_module();
   // minor check to see error in log (also adjust mode for the boot volume)
   check_disks();
#if !defined(EFI_BUILD) && !defined(INITDEBUG)
   // verbose build prints this from RM asm code
   log_printf("pmdata: %04X, log: %04X, iobuf: %04X, stack: %d\n", pmdataseg,
      logbufseg, DiskBufRM_Seg, stacksize);
#endif
   {  // load and launch START module
      int rc = start_it();
      if (rc) {
         char msg[80];
         if (rc!=E_MOD_NOORD)
            snprintf(msg, sizeof(msg), "start error %X...\n", rc);
         else // no required ordinal - just say about wrong file
            snprintf(msg, sizeof(msg), "%s version mismatch!\n", ldi_name);
         vio_strout(msg);
      }
   }
   // return is impossible here (function called via jmp)
   exit_pm32(0);
}

void _std exit_reboot(int warm) {
   // shut down QS services
   exit_prepare(SHUT_OFF);
   // and reboot
   make_reboot(warm);
}

/// install cache callback functions
u32t hlp_runcache(cache_extptr *fptr) {
   if (fptr && fptr->entries!=3) return 0;
   cache_eproc = fptr;
   return 1;
}

/// call installed cache callback with supplied action code
u32t _std hlp_cachenotify(u8t vol, u32t action) {
   if (cache_eproc) (*cache_eproc->cache_ioctl)(vol, action);
   return cache_eproc?1:0;
}
