//
// QSINIT
// critical ini parameters read
//
#include "qslocal.h"
#include "vio.h"

static const char *strList[]={"DBPORT", "BAUDRATE", "RESETMODE", "NOAF",
                              "MFSPRINT", "PAGEMODE", "BINNAME", "BOOTMOUNT",
                              0};
// 4 high bits for special command values (0x80 - bootmount)
static const char *valList[]={"\0""OFF", "\1""ON", "\0""NO", "\1""YES",
                              "\x82RO", "\x83RW", 0};

extern u16t   mfs_rmcmode;
extern u8t    useint13ext,
                  useAFio;
extern char    aboutstr[];
extern char     *rest_ini;

static int cmpname(const char *name, char *pos, u32t len) {
   pos--;
   while ((*pos==' '||*pos=='\t')&&len) { pos--; len--; }
   return len?strnicmp(name,pos-len+1,len):1;
}

static void copyname(char *dst, u32t size, char *pos, u32t len) {
   u32t ii;
   pos--;
   while ((*pos==' '||*pos=='\t')&&len) { pos--; len--; }
   if (size && len>size) len = size-1;
   for (ii=0; ii<len; ii++) dst[ii] = toupper(pos[-len+1+ii]);
   dst[len] = 0;
}

static long getvalue(const char *str, u8t addmask) {
   char ch;
   while (*str==' '||*str=='\t') str++;
   ch = *str;
   if (!ch) return 0;
   /* accept strings like
      "KEY = ON   \n\r"
      OFF/NO = 0, ON/YES = 1 */
   if (ch>='A' && ch<='Z' || ch>='a' && ch<='z') {
      u32t ii, len;
      for (ii=0; valList[ii]; ii++) {
         u8t code = valList[ii][0];
         // if string is a command-related - verify mask match
         if ((code&0xF0) && !(code&addmask)) continue;

         len = strlen(valList[ii]+1);
         if (!strnicmp(str,valList[ii]+1,len)) {
            ch = str[len];
            if (!ch||ch==' '||ch=='\t'||ch=='\n'||ch=='\r') return code&0xF;
         }
      }
   }
   return str2long(str);
}

void get_ini_parm(void) {
   u32t  size = 0,
         pgon = 0;
   char* cfgp = 0;
#ifndef EFI_BUILD
   if (rest_ini) {
      cfgp = rest_ini;
      size = strlen(cfgp) + 1;
   }
#endif
   if (!cfgp) {
      void *ini = 0;
#ifdef EFI_BUILD
      if (ini_path[0]) ini = hlp_freadfull(ini_path, &size, 0);
         else
#endif
      {
         ini = hlp_freadfull("QSINIT.INI", &size, 0);
         if (!ini) ini = hlp_freadfull("OS2LDR.INI", &size, 0);
      }
      if (ini) {
         cfgp = (char*)hlp_memrealloc(ini, size+1);
         cfgp[size++]=0;
      }
   }
   // we have a data (from INI or RESTART)
   if (cfgp) {
      char skipeol = 0, lstart = 1, backch = 0,
             error = 0, // error in string
            reason = 0, // collecting now: 1 - section name, 2 - key name, 3 - key
           section = 0; // 1 - config
      int    value = 0,
           errline = 1,
             strln =-1;
      // this pointer saved as physical!
      sto_save(STOKEY_INIDATA, cfgp, size, 0);

      while (size--) {
         char cc=*cfgp++;
         if (!size) cc='\n';
         if (skipeol && cc=='\n') skipeol=0;

         if (!skipeol)
         switch (cc) {
            case ' ' :
            case '\t':
               if (strln>0) strln++;
               break;
            case '\n':
               if (strln>0 && reason<3) { backch=1; error=1; break; }
               errline++;
            case '\r': {
               if (strln > 0 || value) {
                  long rc = getvalue(cfgp-1-strln, value==8?0x80:0);
                  switch (value) {
                     // "DBPORT"
                     case 1:
                        if (rc>0 && !safeMode) hlp_seroutset(rc,0);
                        break;
                     // "BAUDRATE"
                     case 2: 
                        hlp_seroutset(0,rc);
                        break;
                     // "RESETMODE"
                     case 3:
                        if (rc) {
                           if (rc!=43 && rc!=50 && rc!=60) rc=25;
                           vio_setmode(rc);
                        }
                        break;
                     // "NOAF"
                     case 4: useAFio=0; break;
                     // "MFSPRINT"
                     case 5: if (rc>0) mfs_rmcmode=0; break;
                     // "PAGEMODE"
                     case 6: if (rc>0) vmmode=rc; break;
                     // "BINNAME"
                     case 7:
                        copyname(ldi_name, LDI_NAME_LEN, cfgp-1, strln);
                        break;
                     // "BOOTMOUNT":
                     case 8:
                        if (!safeMode) {
                           // convert 0 to BVMOUNT_DISABLED and 1 to BVMOUNT_AUTO
                           if (!(rc&~1)) rc^=1;
                           bootmount = rc;
                        }
                        break;
                  }
                  strln=0; value=0;
               }
               lstart=1; reason=0;
               break;
             }
            case '=' :
               if (reason==2 && strln>0) {
                  value=0;
                  if (section==1) {     // "config"
                     int ii=0;
                     while (strList[ii])
                        if (!cmpname(strList[ii++], cfgp-1, strln)) {
                           value=ii;
                           break;
                        }
                  }
                  if (!value) skipeol=1; else reason=3;
                  strln=0;
               } else error=1;
               break;
            case ';' : if (lstart) skipeol=1; else
                  if ( strln> 0 ) strln++;
               break;
            case '[' : if (lstart) { strln=0; reason=1; section=0; } 
                  else error=1;
               break;
            case ']' :
               if (reason==1 && !lstart) {
                  if (!cmpname("CONFIG", cfgp-1, strln)) section=1; else
                     section=0;
                  strln=0; skipeol=1;
               } else
                  error=1;
               reason=0;
               break;
            default:
               if (lstart==1 && section) { reason=2; strln=0; }
               lstart=0;
               if (strln>=0) strln++;
         }

         if (backch) { *--cfgp=cc; size++; backch=0; }
         if (error) { skipeol=1; error=0; reason=0; strln=0; }
      }
      //hlp_memfree(ini);
   } else {
#ifdef INITDEBUG
      serial_init(ComPortAddr = 0x2F8);
#else
      serial_init(ComPortAddr = 0x3F8);
#endif
   }
   log_printf("hi!\n");
   log_printf("%s\n",aboutstr);
}
