//
// QSINIT "boot menu" module
// kernel start options merging
//
#include "qsshell.h"
#include "qsutil.h"
#include "ksline.h"
#include <stdlib.h>

/* options to remove from the result line.
   there is no more "DISKSIZE", but let it will be here. */
static const char *emptylist[] = { "DEFAULT", "TIMEOUT", "DISKSIZE", "UNZALL",
  "TTYMENU", "MENUPALETTE", "RESETMODE", "DISKSIZE", "USEBEEP", "NOTASK",
  "HEAPFLAGS", "DEFAULT_PARTITION", "DBFLAGS", "SHAREIRQ", "MFSPRINT",
  "NOCLOCK", "REIPL", "CTRLN", "PCISCAN_ALL", "BINNAME", "BOOTMOUNT",
  "LVMMOUNT", 0 };

int mergeopts(char *line, u32t linelen, char *args, qs_inifile ini) {
   str_list  *opts = str_split(line,","),
            *argsl = str_split(args,","),
          *comkeys = 0;
   int     ii, res;
   char   *runline, *str;
   u32t        idx;
   // delete menu name
   str_delentry(argsl,0);
   // trim all spaces around
   opts  = str_replace(opts ,0,0,STRREP_TRIM,0,-1,1);
   argsl = str_replace(argsl,0,0,STRREP_TRIM,0,-1,1);
   // query list of keys in "config" section
   if (ini) {
      comkeys = ini->getsec("config", GETSEC_NOEMPTY|GETSEC_NOEKEY|GETSEC_NOEVALUE);
      if (comkeys) argsl = str_merge(comkeys, argsl, 0);
   }
   // order is "comkeys, args, line". The latest key value will be used
   opts = str_merge(argsl, opts, STRMER_INI|STRMER_NOEMPTY);
   /* here should be only the one entry for each existing key.
      only need to resolve ALTE/PKEY issue ... */

   for (ii=0, res=0; ii<=5; ii++) {
      static const char *afx[] = { "ALTE", "ALTF1", "ALTF2", "ALTF3",
                                   "ALTF4", "ALTF5"};
      idx = 0;
      if (str_findkey(opts, afx[ii], &idx)) {
         char str[32];
         snprintf(str, sizeof(str), "PKEY=0x%04X", ii?0x6700+(ii<<8):0x1200);
         opts = str_replace(opts, 0, str, STRREP_SET, idx, 1, 1);
         res++;
      }
   }
   // do it again to select one of PKEYs
   if (res>1) opts = str_merge(opts, 0, STRMER_INI);

   // delete dbport key if negative value specified
   idx = 0;
   str = str_findkey(opts, "DBPORT", &idx);
   if (str && atoi(str)<0) str_delentry(opts, idx);

   // remove menu/misc keys from the line
   for (ii = 0; emptylist[ii]; ii++) {
      idx = 0;
      while (str_findkey(opts, emptylist[ii], &idx))
         str_delentry(opts, idx);
   }
   // run line
   runline = str_gettostr(opts, ",");
   res     = str_findkey(opts, "RESTART", 0) ? -1 :
             str_findkey(opts, "RESTARTRAW", 0) ? -2 : 1;
   if (strclast(runline)==',') runline[strlen(runline)-1] = 0;
   //log_printf("[%s]\n", runline);
   free(opts);
   strncpy(line, runline, linelen);
   line[linelen-1] = 0;
   free(runline);

   return res;
}
