//
// QSINIT
// single byte code page support module
// ChaN`s code from FatFs partially used
//
/*  437   U.S.
    720   Arabic
    737   Greek
    771   KBL
    775   Baltic
    850   Latin 1
    852   Latin 2
    855   Cyrillic
    857   Turkish
    860   Portuguese
    861   Icelandic
    862   Hebrew
    863   Canadian French
    864   Arabic
    865   Nordic
    866   Russian
    869   Greek 2
*/

#define DBCS_SUPPORT 0

#include "stdlib.h"
#include "qcl/cplib.h"
#include "qsshell.h"
#include "errno.h"
#include "qsutil.h"
#include "qsmodext.h"
#include "qssys.h"
#include "qsint.h"
#include "vio.h"

#include "cptables.h"

#if DBCS_SUPPORT
#include "cptwide.h"
#define DPAGES     (4)

static wchar_t *dtables[2][DPAGES] = {{uni2oem932, uni2oem936, uni2oem949,
   uni2oem950} , {oem2uni932, oem2uni936, oem2uni949, oem2uni950}};

static u32t      dsizes[2][DPAGES] = {{sizeof(uni2oem932), sizeof(uni2oem936),
   sizeof(uni2oem949), sizeof(uni2oem950)} , {sizeof(oem2uni932),
   sizeof(oem2uni936), sizeof(oem2uni949), sizeof(oem2uni950)}};

static u8t        *rtables[DPAGES] = { Dbsc932, Dbsc936, Dbsc949, Dbsc950 };
static u16t         dpages[DPAGES] = { 932, 936, 949, 950 };
#else
#define DPAGES     (0)
static u16t         dpages[] = { 0 };
#endif // DBCS_SUPPORT

#define CPAGES    (17)

static wchar_t *ctables[] = { Tbl0437, Tbl0720, Tbl0737, Tbl0771, Tbl0775,
   Tbl0850, Tbl0852, Tbl0855, Tbl0857, Tbl0860, Tbl0861, Tbl0862, Tbl0863,
   Tbl0864, Tbl0865, Tbl0866, Tbl0869 };

static u8t     *utables[] = { Upr0437, Upr0720, Upr0737, Upr0771, Upr0775,
   Upr0850, Upr0852, Upr0855, Upr0857, Upr0860, Upr0861, Upr0862, Upr0863,
   Upr0864, Upr0865, Upr0866, Upr0869 };

static u16t cpages[CPAGES] = { 437, 720, 737, 771, 775, 850, 852, 855, 857,
   860, 861, 862, 863, 864, 865, 866, 869 };

#define CPLIB_SIGN        0x49534C4E          // NLSI

#define instance_ret(inst,err)              \
   cpconv_data *inst = (cpconv_data*)data;  \
   if (inst->sign!=CPLIB_SIGN) return err;

#define instance_void(inst)                 \
   cpconv_data *inst = (cpconv_data*)data;  \
   if (inst->sign!=CPLIB_SIGN) return;

codepage_info  cpisys = { 0 };
u32t          classid = 0;

typedef struct {
   u32t         sign;
   u16t     codepage;
   u32t        index;
   wchar_t   *ctable;
   u8t       *dbctbl;
} cpconv_data;

cpconv_data       sys = { CPLIB_SIGN, 0, 0, 0, 0 };

static wchar_t convert(cpconv_data *cd, wchar_t chr, int dir) {
   wchar_t cc = 0;

   if (chr<0x80) cc = chr; else 
   if (cd->codepage && cd->codepage<900) {
      if (dir) { // OEMCP to Unicode
         cc = chr>=0x100 ? 0 : cd->ctable[chr-0x80];
      } else {   // Unicode to OEMCP
#if 1
         u16t *pos = memchrw(cd->ctable, chr, 0x80);
         if (pos) cc = 0x80 + (pos-cd->ctable);
#else
         for (cc = 0; cc<0x80; cc++)
            if (chr==cd->ctable[cc]) break;
         cc = cc+0x80 & 0xFF;
#endif
      }
   } else {
#if DBCS_SUPPORT
      wchar_t *pp = dtables[dir?1:0][cd->index];
      u32t     hi = (dsizes[dir?1:0][cd->index]>>2) - 1;

      if (pp) {
         u32t  li = 0, ii, nn;
         for (nn = 16; nn; nn--) {
            ii = li + (hi - li) / 2;
            if (chr==pp[ii<<1]) break;
            if (chr>pp[ii<<1]) li = ii; else hi = ii;
         }
         if (nn) cc = pp[(ii<<1) + 1];
      }
#endif
   }
   return cc;
}

// note, that this and next functions called under cli!
static u16t _std _ff_convert(u16t src, int to_unicode) {
   if (!sys.codepage) return src>=0x80?0:src;
   return convert(&sys, src, to_unicode);
}

static u16t _std _ff_wtoupper(u16t src) {
   return towupper(src);
}

#if DBCS_SUPPORT
/// test if the character is DBC 1st byte
static int _std _ff_dbc_1st(u8t src) {
   if (sys.dbctbl && src>=sys.dbctbl[0]) {
      if (src<=sys.dbctbl[1]) return 1;                       // 1st byte range 1
      if (src>=sys.dbctbl[2] && src<=sys.dbctbl[3]) return 1; // 1st byte range 2
   }
   return 0;
}

/// test if the character is DBC 2nd byte
static int _std _ff_dbc_2nd(u8t src) {
   if (sys.dbctbl && src>=sys.dbctbl[4]) {
      if (src<=sys.dbctbl[5]) return 1;                       // 2nd byte range 1
      if (src>=sys.dbctbl[6] && src<=sys.dbctbl[7]) return 1; // 2nd byte range 2
      if (src>=sys.dbctbl[8] && src<=sys.dbctbl[9]) return 1; // 2nd byte range 3
   }
   return 0;
}
#endif // DBCS_SUPPORT

static u32t set_cpdata(cpconv_data *cpd, u16t ncp) {
#if DBCS_SUPPORT
   u16t *base = ncp<900 ? cpages : dpages,
         *pos = base ? memchrw(base, ncp, ncp<900?CPAGES:DPAGES) : 0;
#else
   u16t *base = ncp<900 ? cpages : 0,
         *pos = base ? memchrw(base, ncp, CPAGES) : 0;
#endif
   if (!pos) return 0;
   // just verify CP support
   if (!cpd) return 1;
   cpd->codepage   = ncp;
   cpd->index      = pos - base;
   if (ncp<900) {
      cpd->ctable  = ctables[cpd->index];
      cpd->dbctbl  = 0;
   } else {
#if DBCS_SUPPORT
      cpd->ctable  = 0;
      switch (ncp) {
         case 932: cpd->dbctbl = Dbsc932; break;
         case 936: cpd->dbctbl = Dbsc936; break;
         case 949: cpd->dbctbl = Dbsc949; break;
         case 950: cpd->dbctbl = Dbsc950; break;
      }
#else
      return 0;
#endif
   }
   return 1;
}

u32t _exicc cpconv_setcp(EXI_DATA, u16t cp) {
   u32t rc;
   instance_ret(cpd,0);
   mt_swlock();
   rc = set_cpdata(cpd, cp);
   mt_swunlock();
   return rc;
}

wchar_t _exicc cpconv_convert(EXI_DATA, wchar_t chr, int dir) {
   instance_ret(cpd,0);
   if (!cpd->codepage) return chr;
   return convert(cpd, chr, dir);
}

char _exicc cpconv_toupper(EXI_DATA, char chr) {
   instance_ret(cpd,chr);
   if (!cpd->codepage) return toupper(chr);
   return chr>=0x80 ? utables[cpd->index][chr-0x80] : toupper(chr);
}

char _exicc cpconv_tolower(EXI_DATA, char chr) {
   wchar_t cc;
   instance_ret(cpd,chr);
   if (!cpd->codepage) return chr;
   // long way: oem->1200->lower->oem
   cc = cpconv_convert(data, 0, chr, 1);
   if (!cc) return chr;
   cc = towlower(cc);
   cc = cpconv_convert(data, 0, cc, 0);
   return cc ? cc : chr;
}

u16t* _exicc cpconv_cplist(EXI_DATA) {
   u16t* rc = (u16t*)malloc((CPAGES+1)*2);
   memcpy(rc, &cpages, CPAGES*2);
   rc[CPAGES] = 0;
   mem_localblock(rc);
   return rc;
}

u16t _exicc cpconv_getsyscp(EXI_DATA) {
   return sys.codepage;
}

u8t* _exicc cpconv_uprtab(EXI_DATA, u16t cp) {
   instance_ret(cpd,0);
   if (!cp) cp = sys.codepage;
   if (cp) {
      u16t *pos = memchrw(cpages, cp, CPAGES);
      if (!pos) return 0;
      return utables[pos - cpages];
   }
   return 0;
}

qserr _exicc cpconv_setsyscp(EXI_DATA, u16t cp) {
   u32t rc;
   mt_swlock();
   if (sys.codepage==cp) rc = 0; else {
      // check support
      rc = set_cpdata(0, cp);
      if (rc) {
         u32t askmsg[2];
         askmsg[0] = cp;
         askmsg[1] = (u32t)&rc;
         rc = 0;
         // ask all around system about new CP
         sys_notifyexex(SECB_CPASK, 8, &askmsg);
      
         if (!rc) {
            set_cpdata(&sys, cp);
            cpisys.cpnum    = cp;
            cpisys.oemupr   = utables[sys.index];
            cpisys.convert  = _ff_convert;
            cpisys.wtoupper = _ff_wtoupper;
#if DBCS_SUPPORT
            cpisys.dbc_1st  = _ff_dbc_1st;
            cpisys.dbc_2nd  = _ff_dbc_2nd;
#else
            cpisys.dbc_1st  = 0;
            cpisys.dbc_2nd  = 0;
#endif
            // system notify
            sys_notifyexec(SECB_CPCHANGE, (u32t)&cpisys);
         }
      } else
         rc = E_SYS_BADCP;
   }
   mt_swunlock();
   return rc;
}

wchar_t _exicc cpconv_towlower(EXI_DATA, wchar_t chr) {
   return towlower(chr);
}

wchar_t _exicc cpconv_towupper(EXI_DATA, wchar_t chr) {
   return towupper(chr);
}


static void *methods_list[] = { cpconv_setcp, cpconv_convert, cpconv_toupper,
   cpconv_tolower, cpconv_cplist, cpconv_getsyscp, cpconv_setsyscp, cpconv_uprtab,
   cpconv_towlower, cpconv_towupper};

static void _std cpconv_init(void *instance, void *data) {
   cpconv_data *cpd = (cpconv_data*)data;
   mt_swlock();
   // just a copy of current system settings
   memcpy(cpd, &sys, sizeof(cpconv_data));
   mt_swunlock();
}

static void _std cpconv_done(void *instance, void *data) {
   instance_void(cpd);
   cpd->sign = 0;
}

u32t _std shl_chcp(const char *cmd, str_list *args) {
   if (args->count==1 && strcmp(args->item[0],"/?")==0) {
      cmd_shellhelp(cmd,CLR_HELP);
      return 0;
   }
   cmd_printseq(0,PRNSEQ_INIT,0);

   if (args->count<=1) {
      if (args->count==0) {
         if (!sys.codepage) printf("There is no code page selected\n");
            else printf("Active code page : %hd\n", sys.codepage);
         return 0;
      } else
      if (stricmp(args->item[0],"LIST")==0) {
         char *cpstr = strdup("Supported code pages : "), buf[16];
         u32t  total = CPAGES + DPAGES, ii;

         for (ii=0; ii<total; ii++) {
            sprintf(buf, "%hd", ii<CPAGES ? cpages[ii] : dpages[ii-CPAGES]);
            if (ii<total-1) strcat(buf, ", ");
            cpstr = strcat_dyn(cpstr, buf);
         }
         cmd_printtext(cpstr,0,0);
         free(cpstr);
         return 0;
      } else {
         long cp = str2long(args->item[0]);
         if (cp>0) {
            qserr err = cpconv_setsyscp(0, 0, cp);

            if (err==E_SYS_BADCP) {
               printf("There is no code page %d\n", cp);
               return EINVAL;
            } else
            if (err) {
               cmd_shellerr(EMSG_QS, err, 0);
               return qserr2errno(err);
            } else
               return 0;
         }
      }
   }
   cmd_shellerr(EMSG_CLIB,EINVAL,0);
   return EINVAL;
}

unsigned __cdecl LibMain(unsigned hmod, unsigned termination) {
   if (!termination) {
      if (mod_query(mod_getname(hmod,0), MODQ_LOADED|MODQ_NOINCR)) {
         vio_setcolor(VIO_COLOR_LRED);
         printf("%s already loaded!\n", mod_getname(hmod,0));
         vio_setcolor(VIO_COLOR_RESET);
         return 0;
      }
      classid = exi_register("qs_cpconvert", methods_list,
         sizeof(methods_list)/sizeof(void*), sizeof(cpconv_data), 0,
            cpconv_init, cpconv_done, 0);
      if (!classid) {
         log_printf("unable to register cpconvert class!\n");
         return 0;
      }
      cmd_shelladd("CHCP",shl_chcp);
   } else {
      // remove cp class
      if (classid)
         if (exi_unregister(classid)) classid = 0;
      // DENY unload if class was not unregistered
      if (classid) return 0;

      cmd_shellrmv("CHCP",shl_chcp);
      // drop code page support
      sys_notifyexec(SECB_CPCHANGE,0);
   }
   return 1;
}
