//
// QSINIT "extcmd" module
// bldlevel command implementation
// --------------------------------------
// but also a code for any platform -
// just "wcl386 bldlevel.c" to make windows/linux bldlevel

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#ifndef __QSINIT__
// item numbers in the list
#define BLDL_VENDOR   0
#define BLDL_REV      1
#define BLDL_DATE     2
#define BLDL_HOST     3
#define BLDL_ASDF     4
#define BLDL_LANG     5
#define BLDL_CNTRY    6
#define BLDL_BUILD    7
#define BLDL_FIXPACK  8
#define BLDL_VER      9
#define BLDL_DESC    10

#define malloc_th  malloc
#define mem_localblock(x)

typedef struct {
   size_t   count;
   char*  item[1];
} str_list;

#else  // __QSINIT__
#include "qsshell.h"
#include "qssys.h"

// unpublished call, used here to handle PXE with less memory & time
u32t  _std hlp_fopen2(const char *name, int allow_pxe);
#endif // __QSINIT__

#define BLDL_ITEMS   11

static const char *prn_name[BLDL_ITEMS] = { "Vendor", "File Revision",
   "Build Time", "Build Host", "ASD Feature ID", "Language Code",
   "Country Code", 0, "FixPack", "File Version", "Description" };

/* not sure about cheking the module description, but the original
   bldlevel.exe can`t find the signature beyond the first ~500kb of the binary.
   So made the same limit here */
#define MAX_SEARCH_LEN (512*1024*1024)

static void* read_file(const char *path, int bootio, unsigned long *len) {
   void *res = 0;
   int   err = 0;

   if (bootio)
      if (path[1]==':' || strchr(path,'\\') || strchr(path,'/')) bootio = 0;

   if (!bootio) {
      long  flen;
      FILE *fsrc = fopen(path, "rb");
      if (!fsrc) return 0;

      fseek(fsrc, 0, SEEK_END);
      flen = ftell(fsrc);
      rewind(fsrc);

      if (flen>128) {
         if (flen>MAX_SEARCH_LEN) flen = MAX_SEARCH_LEN;
         res = malloc_th(flen);
         if (!res) err = ENOMEM; else
            if (fread(res, 1, flen, fsrc) != flen) {
               free(res);
               res = 0;
            } else
               *len = flen;
      } else 
         err = EINVAL;
      fclose(fsrc);
   } else {
#ifdef __QSINIT__
      u32t flen = hlp_fopen2(path,1), alen, rlen;
      if (flen!=FFFF) {
         alen = !flen ? MAX_SEARCH_LEN :
                 flen<128 ? 0 : flen>MAX_SEARCH_LEN ? MAX_SEARCH_LEN : flen;
         if (alen) {
            res  = malloc_th(alen);
            if (!res) err = ENOMEM; else {
               /* using "unofficial" way to handle PXE here - offset just
                  doesn`t work in PXE mode, but we read from 0, so it`s ok.
                  i.e. this will read only first 500k of the kernel file when
                  checking it on TFTP server (noticeable acceleration) */
               rlen = hlp_fread(0, res, alen);
               // PXE?
               if (!flen) {
                  if (rlen<128) err = EINVAL;
               } else
               if (rlen<alen) err = EIO;
            }
         }
         hlp_fclose();
         // file too small
         if (!alen) err = EINVAL;

         if (err && res) {
            free(res);
            res = 0;
         } else
            *len = rlen;
      } else
#endif
         err = ENOENT;
   }
   if (err) errno = err;
   return res;
}

char* __stdcall sys_bldlevel(const char *path, int bootio) {
   unsigned long  len;
   unsigned char *buf = read_file(path, bootio, &len), *fp,
                 *pos = buf;
   unsigned long  rem = len, ii;
   char          *res = 0;

   if (!buf) return 0;

   while (rem) {
      fp = memchr(pos, '@', rem);
      if (!fp) break;
      rem = len - (fp - buf) - 2; // dec one more byte
      pos = ++fp;
      // verify string ...
      if (*fp=='#') {
         for (ii=1; ii<rem && fp[ii]>=' ' && fp[ii]<0x7F &&
            fp[ii]!=':'; ii++);
         if (fp[ii++]==':') {
            for (; ii<rem && fp[ii]>=' ' && fp[ii]<0x7F &&
               fp[ii]!='#'; ii++);
            if (fp[ii]=='#') {
               if (!memcmp(fp+ii, "#@##built", 9)) {
                  for (ii+=4; ii<rem-1 && fp[ii]>=' ' && fp[ii]<0x7F; ii++)
                     if (fp[ii]=='@' && fp[ii+1]=='@') break;
                  if (fp[ii]!='@') continue;
               } else
               if (!memcmp(fp+ii, "#@##1##", 7)) {
                  int ddcnt = 0;
                  ii+=7+25;  // time field min width is 25
                  if (ii+8>rem) break;

                  for (; ii<rem-1 && fp[ii]>=' ' && fp[ii]<0x7F; ii++) {
                     if (fp[ii]==':') ddcnt++;
                     if (fp[ii]=='@' && fp[ii+1]=='@') break;
                  }
                  if (ddcnt!=6 || fp[ii]!='@') continue;
               }
               // check for all 3 types
               if (fp[ii+1]!='@') continue;

               for (; ii<rem && fp[ii]>=' '; ii++);
               res = malloc_th(++ii+1);
               if (!res) { errno = ENOMEM; break; }
               memcpy(res, --fp, ii);
               res[ii] = 0;
               break;
            }
         }
      }
   }
   free(buf);
   // switch ownership from thread to process
   if (res) mem_localblock(res);
   return res;
}

/* parse string mj[.mn[.rev[.misc]]]
   all fields must start with a digit, misc can have any chars at the end.
   returns -1 in rev if it`s missing,
   returns 0 in mn if it`s missing */
static size_t getver(char *src, int len, size_t *mn, int *rev, char **misc) {
   size_t major;
   char    *str;
   if (len>0) {
      str = malloc_th(len+1);
      memcpy(str, src, len);
      str[len] = 0;
   } else
      str = strdup(src);
   src = str;

   while (*str==' ') str++;
   major = strtoul(str,0,10);
   *mn   = 0;
   *rev  = -1;
   *misc = 0;
   while (*str && *str!='.') str++;
   if (*str++) {
      *mn = strtoul(str,0,10);
      while (*str && *str!='.') str++;
      if (*str++) {
         *rev = strtoul(str,0,10);
         while (*str && *str!='.') str++;
         if (*str++) {
            memmove(src, str, strlen(str)+1);
            *misc = src;
         }
      }
   }
   if (!*misc) free(src);
   return major;
}

#define COPY_STR(type, src, len) { \
   int Len = len;                  \
   if (Len) memcpy(dp, src, Len);  \
   dp[Len] = 0;                    \
   rs->item[type] = dp;            \
   dp += Len + 1;                  \
}

str_list* __stdcall sys_bldlist(const char *bldlevel) {
   str_list *rs;
   int     type = 2, ii, rev = -1, build = -1;
   char     *mp, *dp, *sp,
          *dscp = 0,
          *misc = 0;
   size_t   maj = 0,
            min = 0;

   if (*bldlevel++!='@') return 0;
   if (*bldlevel++!='#') return 0;

   // detect type of the source
   mp = strstr(bldlevel, "#@##1##");
   if (!mp) {
      mp = strstr(bldlevel, "#@##built");
      if (mp) type = 1; else {
         mp = strstr(bldlevel, "#@");
         if (mp) type = 0; else
            return 0;
      }
   }
   // 11 zeroes & 32 for BLDL_VER
   rs = malloc_th(sizeof(str_list) + (BLDL_ITEMS-1)*sizeof(char*) +
                  strlen(bldlevel) + BLDL_ITEMS + 32);
   rs->count = BLDL_ITEMS;
   dp = (char*)(rs+1) + (BLDL_ITEMS-1)*sizeof(char*);
   // init all with an empty string
   for (ii=0; ii<BLDL_ITEMS; ii++) rs->item[ii] = dp;
   *dp++ = 0;

   sp = strchr(bldlevel,':');
   // broken string?
   if (!sp || sp > mp) ii = mp-bldlevel; else {
      ii  = sp-bldlevel;
      if (type==1)
         while (sp+1!=mp && !isdigit(sp[1])) sp++;
      if (mp-sp > 1)
         maj = getver(sp+1, mp-sp-1, &min, &rev, &misc);
   }
   // copy vendor
   COPY_STR(BLDL_VENDOR, bldlevel, ii);

   if (type==0) { // type 0 : @#ven:ver#@desc
      // ptr to decription
      dscp = mp;
   } else
   if (type==1) { // type 1 : @#ven:ver#@##built time -- On host;0.1@@desc
      mp += 9;
      while (*mp==' ') mp++;
      dscp = strstr(mp, "@@");
      sp = strstr(mp, " -- ");
      if (sp && sp < dscp) {
         COPY_STR(BLDL_DATE, mp, sp-mp);
         mp = sp + 6;
         while (*mp==' ') mp++;
         sp = strchr(mp, ';');
         if (sp && sp < dscp)
            COPY_STR(BLDL_HOST, mp, sp-mp);
      }
   } else { // type 2 : @#ven:ver#@##1## date host:ASD:Lang:Cou:Bld::Fix@@desc
      mp += 7;
      dscp = strstr(mp, "@@");
      sp = mp + 25;
      while (*mp==' ' && mp!=sp) mp++;

      if (sp < dscp) {
         char *dpos[6];
         ii = sp-mp;
         while (ii && mp[ii-1]==' ') ii--;
         COPY_STR(BLDL_DATE, mp, ii);

         ii = 0;
         mp = sp;
         while (sp < dscp) {
            if (*sp==':')
               if (ii<6) dpos[ii++] = sp; else ii++;
            sp++;
         }
         // string is valid?
         if (ii==6) {
            static signed char DstIdx[6] = { BLDL_ASDF, BLDL_LANG, BLDL_CNTRY,
               BLDL_BUILD, -1, BLDL_FIXPACK};
            int idx;

            for (idx=0; idx<6; idx++)
               if (DstIdx[idx]>=0) {
                  int   len = (idx==5 ? dscp : dpos[idx+1]) - dpos[idx] - 1;
                  char *str = dpos[idx]+1;
                  while (len && *str==' ') { str++; len--; }
                  COPY_STR(DstIdx[idx], str, len);
               }
            if (rs->item[BLDL_BUILD][0])
               build = strtoul(rs->item[BLDL_BUILD],0,10);
            while (mp < dpos[0] && *mp==' ') mp++;
            ii = dpos[0] - mp;
            while (ii && mp[ii-1]==' ') ii--;
            if (ii) COPY_STR(BLDL_HOST, mp, ii);
         }
      }
   }
   // copy decription (trim spaces on both sides)
   if (dscp) {
      dscp += 2;
      while (*dscp==' ') dscp++;
      ii = strlen(dscp);
      while (ii && dscp[ii-1]==' ') ii--;
      COPY_STR(BLDL_DESC, dscp, ii);
   }
   // file revision
   sprintf(dp, rev>=0?"%u.%02u.%i":"%u.%02u", maj, min, rev);
   rs->item[BLDL_REV] = dp;
   dp += strlen(dp);
   if (misc) {
      *dp++ = '.';
      memcpy(dp, misc, mp-misc);
      dp += mp-misc;
      free(misc);
   }
   *dp++ = 0;
   // file version (all this should fit in 32 chars reserved above)
   sprintf(dp, build>=0?"%u.%u.%i":"%u.%u", maj, min, build);
   rs->item[BLDL_VER] = dp;
   dp += strlen(dp)+1;
   // build
   if (build>=0) {
      sprintf(dp, "%i", build);
      rs->item[BLDL_BUILD] = dp;
      dp += strlen(dp)+1;
   }
   // switch ownership from thread to process
   mem_localblock(rs);
   return rs;
}

#ifndef __QSINIT__
int main(int argc, char *argv[]) {
   if (argc!=2) {
      printf("bldlevel <file>\n");
      return EINVAL;
   } else {
      char *str = sys_bldlevel(argv[1],0);
      if (!str) {
         if (!errno) {
            printf("Signature not found.\n");
         } else {
            printf("Error %i opening \"%s\"\n", errno, argv[1]);
            return errno;
         }
      } else {
         str_list *lst = sys_bldlist(str);
         printf("%s\n", str);

         if (!lst) printf("Error while decoding signature.\n"); else {
            int ii;
            for (ii=0; ii<BLDL_ITEMS; ii++)
               if (prn_name[ii] && lst->item[ii][0])
                  printf("%-18s: %s\n", prn_name[ii], lst->item[ii]);
            free(lst);
         }
         free(str);
      }
   }
   return 0;
}
#else

u32t _std shl_bldlevel(const char *cmd, str_list *args) {
   int rc=-1, boot=0, q=-1, ii;
   static char *argstr   = "boot|q|q0|q1|q2|q3|q4|q5|q6|q7|q8|q9|q10";
   static short argval[] = { 1, 11, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
   char            *arg0 = 0;
   // help overrides any other keys
   if (str_findkey(args, "/?", 0)) {
      cmd_shellhelp(cmd,CLR_HELP);
      return 0;
   }
   args = str_parseargs(args, 0, SPA_RETLIST|SPA_NOSLASH, argstr, argval, &boot,
                        &q, &q, &q, &q, &q, &q, &q, &q, &q, &q, &q, &q);
   // process command
   if (q<0 && args->count==1 || q>=0 && args->count==2) {
      char *str = sys_bldlevel(args->item[q>=0?1:0], boot);
      if (!str) {
         rc = errno;
         if (!errno && q<0) printf("Signature not found.\n");
         if (q>=0) setenv(args->item[0], "", 1);
      } else {
         str_list *lst = sys_bldlist(str);
         if (q<0) {
            printf("%s\n", str);
            if (!lst) printf("Error while decoding signature.\n"); else
               for (ii=0; ii<BLDL_ITEMS; ii++)
                  if (prn_name[ii] && lst->item[ii][0])
                     printf("%-18s: %s\n", prn_name[ii], lst->item[ii]);
         } else
            setenv(args->item[0], q>=BLDL_ITEMS?str:lst?lst->item[q]:"", 1);
         rc = 0;
         if (!lst) free(lst);
         free(str);
      }
   } else // force visible error
      q = -1;
   // free args clone, returned by str_parseargs() above
   free(args);
   if (rc<0) rc = EINVAL;
   if (rc && q<0) cmd_shellerr(EMSG_CLIB,rc,0);
   return rc;
}

#endif
