//
// QSINIT runtime
// simple heap for the fixed size small blocks
//
#ifndef QSINIT_FXSHEAP
#define QSINIT_FXSHEAP

#include "qstypes.h"
// this heap can be used in 64-bit part
#ifndef EFIX64
#include "qsclass.h"
#include "stddef.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif
typedef void *qs_thp;

/** initialize heap.
    Note that heap instance calls are not thread-safe, use the shared class
    below to create thread safe instances.

    @param  blocksize   size of the heap block (<=32k)
    @param  limit       maximum number of heap blocks (will be aligned to 64)
    @param  memflags    only QSMA_LOCAL and QSMA_READONLY are accepted (r/o
                        can be used for PID 1 heaps)
    @param  [out] res   heap if success
    @return error code or 0 */
qserr  _std hlp_thpinit (u32t blocksize, u32t limit, u32t memflags, qs_thp *res);

/** delete heap.
    @param  hp          heap
    @return error code or 0 */
qserr  _std hlp_thpfini (qs_thp hp);

/** allocates a block in the heap.
    @param  hp          heap
    @return pointer to zero filled memory or 0 if error */
void*  _std hlp_thpalloc(qs_thp hp);

/** free memory block.
    @param  hp          heap
    @param  ptr         heap block pointer
    @return error code or 0 */
qserr  _std hlp_thpfree (qs_thp hp, void *ptr);

#ifdef QSINIT_EXPCLASSES
/// shared class implementation, to make it thread-safe by request
typedef struct {
   /** initialize heap.
       @param  blocksize   size of the heap block (<=32k).
       @param  limit       maximum # of heap blocks (will be aligned to 64)
       @param  flags       see hlp_thpinit() above
       @return error code or 0 */
   qserr   _exicc (*init  )(u32t blocksize, u32t limit, u32t flags);
   /** allocates a block in the heap.
       @return pointer to zero filled memory or 0 if failed */
   void*   _exicc (*alloc )(void);
   /// free memory block.
   qserr   _exicc (*free  )(void *ptr);
} _qs_theap, *qs_theap;
#endif

#ifdef __cplusplus
}
#endif
#endif // QSINIT_FXSHEAP
