//
// QSINIT runtime
// rt<->qs conversion funcs
//
#ifndef QSINIT_RTCVT
#define QSINIT_RTCVT

#ifdef __cplusplus
#include "qstypes.h"
#include "classes.hpp"
#include "qsshell.h"

/** convert TStrings to string list for usage in pure C code.
    warning! it trims spaces in the supplied source list! */
str_list* str_getlist(TStringVector &lst);
/// the same as str_getlist() but returns list one in one
str_list* str_getlistpure(TStringVector &lst);

/// convert string list to TStrings
template <class F> 
void str_getstrs(const str_list* lst, F &strs) {
   strs.Clear();
   if (!lst||!lst->count) return;
   for (u32t ll=0;ll<lst->count;ll++) strs.Add(lst->item[ll]);
}

/// convert string list to TStrings
template <class F> 
void str_getstrs2(const str_list* lst, F &strs, u32t start) {
   strs.Clear();
   if (!lst||!lst->count) return;
   for (u32t ll=start;ll<lst->count;ll++) strs.Add(lst->item[ll]);
}

/** unroll TStrings to a plain memory.
   @param lst              source list (TStringVector part)
   @param start            first entry to copy
   @param count            # of entries to copy, 0 for all
   @param dst              destination memory (0 to malloc it)
   @param [in,out] dstlen  length of destination, should be enough for the
                           list, replaced with real/needed length at exit
   @param [out] ptrlist    optional list of pointers to strings (filled at
                           exit), must be long enough for count entries
   @return dst or alloced memory or 0 if dstlen is too small or no entries
           to copy */
char* str_unroll(const TStringVector &lst, u32t start, u32t count, void *dst,
                 u32t &dstlen, char **ptrlist);

#endif // __cplusplus
#endif // QSINIT_RTCVT
