//
//  sp strings (wide version)
//  no floats and apply values (first for portability, second was never used)
//  +int removed to not confuse watcom bad type conversion
//
//  (c) dixie 1997-2023 ;)
//
#ifndef _SP_WSTRINGS_INCLUDED
#define _SP_WSTRINGS_INCLUDED

#ifndef __cplusplus
#error strings.hpp is for use with C++
#endif

#include <wchar.h>
#include "sp_defs.h"
#include <stdlib.h>
#ifndef NO_SP_ASSERTS
#include <assert.h>
#endif
#include "sp_str.h"

typedef struct {
   l        chars;
   l       maxnow;
   l       usecnt;
   d         hash;
   wchar_t buf[1];
} strinfow;
typedef strinfow *pstrinfow;

extern wchar_t *spnullstrw;

class  spstrw {
private:
  strinfow *data;
  void reallocstr(l newsize) {                // realloc memory
    if (!data||data->maxnow<=newsize) {
      data = data?(pstrinfow)REALLOC(data, newsize+sizeof(strinfow)+1 << 1):
                  (pstrinfow)MALLOC(newsize+sizeof(strinfow)+1 << 1);
      data->maxnow = newsize;
    }
  }
  void allocstr(l size) {                     // alloc new string
    reallocstr(size+1);
    data->chars  = 0;
    data->hash   = FFFF;
    data->buf[0] = 0;
    data->usecnt = 1;
  }
  void disposestr(void) { if (data) FREE(data); data=0; }  // free string
  void initstring(wchar_t const *str,d __n=0);
  void initstringa(char const *str,d __n=0);
  Bool setu8c(const char *src,l len);
  d calchash() const;
  d  Lock;
  void _unique();
public:
// constructors
  spstrw() { data=NULL; Lock=0; }
  spstrw(int FromInt) { data=NULL; Lock=0; operator= ((int)FromInt); }
  spstrw(l FromInt)   { data=NULL; Lock=0; operator= ((int)FromInt); }
  spstrw(d FromDword) { data=NULL; Lock=0; operator= (FromDword); }
  spstrw(spstrw const &,                      // from another string
          d __pos=0,                          // from __pos, first __n chars
          d __n=0);
  spstrw(spstr const &,                       // from ascii string
          d __pos=0,                          // from __pos, first __n chars
          d __n=0);
  spstrw(wchar_t []);                         // from C string, first __n chars
  spstrw(wchar_t const*, l __n=0);            // --'--'--
  spstrw(char const*, l __n=0);               // --'--'-- (no chars >127!)
  spstrw(wchar_t,d __rep=1);                  // from single character

  ~spstrw() { clear(); }                      // destructor

  void clear() {
#ifndef NO_SP_ASSERTS
     assert(!Lock);
#endif
     if (data&&--data->usecnt==0) disposestr();
     data=0;
  }

// conversion operators
  const wchar_t *operator()() const {         // to C string
     return data?data->buf:spnullstrw;
  }
  const wchar_t *c() const { return operator()();}    // to C string
  operator wchar_t () const
    { return data?data->buf[0]:*spnullstrw; }         // to single character
  operator l() const { return Int(); }
  operator d() const { return Dword(); }

  spstrw& sprintf(const wchar_t *fmt,...);    // replace old contents with string...

// assignment operators
  spstrw& operator= (spstrw const& );         // from another string
  spstrw& operator= (wchar_t const* );        // from a C string
  spstrw& operator= (int const );             // from an integer
  spstrw& operator= (d const );               // from an dword

// append operators
  spstrw& operator+= (spstrw const& );        // a string
  spstrw& operator+= (wchar_t const* );       // a C string
  spstrw& operator+= (wchar_t);               // char

  spstrw operator() (l __pos,l __len) const;  // substring operator

  wchar_t& operator[] (l __pos);              // reference-to-char operator
  // return last character
  wchar_t lastchar() const { return data?(data->chars?data->buf[data->chars-1]:0):0; }

  l       Int(l __pos=0) const;               // try to read int
  d       Dword(l __pos=0) const;             // try to read dword
#if defined(__LONGLONG_SUPPORT) && defined(STRTOUL64)
  s64     Int64(l __pos=0) const;             // try to read long long
  u64     Dword64(l __pos=0) const;           // try to read qword
#endif
  wchar_t*LockPtr(l Size=256);                // lock pointer
  void    UnlockPtr();                        // unlock pointer
  Bool    UnlockPtr(l Len);                   // unlock with size (MFC-like)
  d       Locked() const { return Lock; }     // return lock counter

// comparison operators
  friend Bool operator== (spstrw const&,spstrw const&);    // string == string?
  friend Bool operator== (spstrw const&,wchar_t const*);   // string == C string?
  friend Bool operator== (wchar_t const*,spstrw const&);   // C string == string?

  friend Bool operator!= (spstrw const &,spstrw const &);  // string != string?
  friend Bool operator!= (spstrw const &,wchar_t const* ); // string != C string?
  friend Bool operator!= (wchar_t const *,spstrw const &); // C string != string?

// concatenate string operators
  friend spstrw operator+ (spstrw const&,spstrw const&);   // string + string
  friend spstrw operator+ (spstrw const&,wchar_t const*);  // string + C string
  friend spstrw operator+ (wchar_t const*,spstrw const&);  // C string + string
  friend spstrw operator+ (spstrw const&,wchar_t);         // string + character
  friend spstrw operator+ (wchar_t,spstrw const &);        // character + string

  l length() const { return data?data->chars:0; }          // length of the string

  l pos(spstrw const &,l startfrom=0) const;  // search string
  l pos(wchar_t const *,l startfrom=0) const; // --'--'--
  l rpos(wchar_t const *) const;              // search from the right
  l cpos(wchar_t,l startfrom=0) const;        // search char
  l crpos(wchar_t) const;                     // search char from the right

  l bepos(const wchar_t *begin,const wchar_t *end,l &epos,l startfrom=0) const;
                                              // search substring by begin/end
// case conversion & misc
  spstrw& upper();                            // to UPPER CASE
  spstrw& lower();                            // to lower case
  spstrw& capitalize();                       // to Capital Case
  spstrw& trim();                             // remove both sides spaces
  spstrw& trimleft();                         // remove start spaces
  spstrw& trimright();                        // remove trailing spaces
  spstrw& expandtabs(int alignpos=8);         // tabs -> spaces
  spstrw& replace(const spstrw&,const spstrw&);            // replace
  spstrw& replace(const wchar_t*,const wchar_t*);          // replace
  spstrw& insert(const spstrw&,l pos);        // insert string
  spstrw& insert(const wchar_t*,l pos);       // insert string
  spstrw& del(l pos,d len=1);                 // delete substring
  spstrw& dellast() { return del(length()-1); }
  spstrw  deltostr(l pos,d len);              // del to new string
  spstrw& unquote(wchar_t quote='\"');        // unquote string
  spstrw& quote(wchar_t quote='\"');          // quote string if not quoted
  void unique(d len=0);                       /* make string unique with the
                                                 buffer lenght >=len */
  l words(const wchar_t *separators=(wchar_t*)L" \x9") const;
  l wordpos(l ord,const wchar_t *separators=(wchar_t*)L" \x9") const;
  spstrw word(l ord,const wchar_t *separators=(wchar_t*)L" \x9") const;
  l word_Int(l ord,const wchar_t *separators=(wchar_t*)L" \x9") const;
  d word_Dword(l ord,const wchar_t *separators=(wchar_t*)L" \x9") const;
  spstrw& replaceword(l ord,const wchar_t *separators,const wchar_t *str);

  spstrw left(l count) const;
  spstrw right(l count) const;

  d countchar(wchar_t chr) const;                  // # of this char
  spstrw& replacechar(wchar_t from,wchar_t to);    // fast replace for chars

  spstr getutf8(d pos=0,l len=-1) const;           // return a copy as utf-8 string
  Bool setutf8(const spstr &src,d pos=0,l len=-1); // replace to utf-8 text
  Bool setutf8(const char *src,l len=-1);          // --'--'--

  d hash() const { return data?(data->hash!=FFFF?data->hash:calchash()):0; }

  Bool operator! () const { return data?!data->chars:1; }  // empty string?
  spstrw& Init(l size=12)
    { data=0; Lock=false; allocstr(size); return *this; }  // constructor
                                                   // do not call directly!!
};

#endif // _SP_WSTRINGS_INCLUDED
