//
//  StringList (wide string version)
//  (c) dixie 1997-2023 ;)
//
#ifndef SP_STRINGSW
#define SP_STRINGSW

#include "sp_strw.h"
#include "classes.hpp"

typedef List<spstrw>  TStringWVector;

template <class L>
class StringsW {
protected:
  typedef List<L> TListL;
public:
  TListL       LList;
  TStringWVector Str;

  StringsW() {}
  StringsW<L>& operator=(const StringsW<L> &F) { Str=F.Str; LList=F.LList; return *this; }
  StringsW(wchar_t **List, l Count);
#if !defined(__WATCOMC__)&&!defined(__IBMCPP__)
  // StringsW<another L> assigment
  template <class F> StringsW(const F &From)
    { Str=From.Str; LList.SetCount(From.LList.Count()); }
#else
  // compilers with ugly template support
  operator TStringWVector&() { return Str; }
  StringsW(TStringWVector &From)
    { Str=From; LList.SetCount(From.Count()); }
  StringsW<L>& operator=(const TStringWVector&From)
    { Str=From; LList.SetCount(From.Count()); return *this; }
#endif
  l Add(const spstrw &s) {LList.IncCount(); return Str.Add(s);}
  l AddObject(const spstrw &s,L value) {LList.Add(value); return Str.Add(s);}

  l AddReplace(const spstrw &s) {
    l idx=IndexOf(s);
    if (idx>=0) Str[idx]=s;
    return idx<0?Add(s):idx;
  }

  void Delete(l Pos,l Count=1,Bool CleanUp=false) {
    LList.Delete(Pos,Count,CleanUp);
    Str.Delete(Pos,Count,CleanUp);
  }
  void Exchange(l Idx1,l Idx2) { LList.Exchange(Idx1,Idx2); Str.Exchange(Idx1,Idx2);}

  /// name by index
  spstrw Name(l idx);
  /// name by value
  spstrw Name(const spstrw&);
  /// value by name
  spstrw Value(const wchar_t*);
  /// value by name
  spstrw Value(const spstrw&);
  /// int value by name
  int    IntValue(const wchar_t*);
  /// dword value by name
  d      DwordValue(const wchar_t*);
  /// value by index (string number)
  spstrw Value(l);

  spstrw StrValue(l idx) const { return spstrw(Str.Value()[idx]); }

  spstrw& operator[](l idx) { return Str[idx]; }
  L& Objects(l idx) {return LList[idx]; }
  l Max() const {return Str.Max();}
  l Count() const {return Str.Count();}
//  void Compact() {LList.Compact(); Str.Compact();}
  void Clear() {LList.Clear(); Str.Clear();}
  void Sort(Bool ICase=false);
#ifdef SORT_BY_OBJECTS_ON
  void SortByObjects(Bool Forward=true);
#endif
  l TrimEmptyLines(Bool EndOnly=false);
  void TrimAllLines(l Pos=0);

  l SplitString(const spstrw &ss,const wchar_t *separators=(wchar_t*)L" \x9");
  l ParseCmdLine(const spstrw &ss,wchar_t separator=' ');

  void Insert(l Pos,spstrw Value,l Repeat=1);
  void InsertObject(l Pos,spstrw Value1,L Value2,l Repeat=1) {
    Str.Insert(Pos,Value1,Repeat);
    LList.Insert(Pos,Value2,Repeat);
  }

  void AddStrings(l Pos,const StringsW<L> &FromStrings) {
    LList.Insert(Pos,FromStrings.LList);
    Str.Insert(Pos,FromStrings.Str);
  }
  l IndexOf(const spstrw &Value,l StartFrom=0);
  l IndexOfICase(const wchar_t *SName,l StartFrom=0);  // ignore case
  l IndexOfName(const wchar_t *SName,l StartFrom=0);
  l IndexOfName(const spstrw &Name,l StartFrom=0);   // ignore case
  l IndexOfObject(const L Value,l StartFrom=0) { return LList.IndexOf(Value,StartFrom); }
  Bool Equals(const StringsW<L> &St) {
     return Str.Equal(St.Str) && LList.Equal(St.LList);
  }
  void Move(l FromIdx,l ToIdx) {
    Str[ToIdx] =Str[FromIdx];
    LList[ToIdx]=LList[FromIdx];
  }
  wchar_t *GetText(const wchar_t *eol=(wchar_t*)L"\x0D\x0A", l start=0, l count=-1) const;
  spstrw GetTextToStr(const wchar_t *eol=(wchar_t*)L"\x0D\x0A", l start=0, l count=-1) const;
  /// merge command line (add quotes around lines with spaces)
  spstrw MergeCmdLine(const wchar_t *separator=(wchar_t*)L" ", wchar_t quote='\"') const;

  /** accept both UNIX & normal text,
      special CSV multiline support: \n act as soft-CR and ignored, only \n\r
      used as CR */
  void SetText(const char*,l len=0,Bool is_csv=false,Bool is_utf8=false);
  void SetTextW(const wchar_t*,l len=0,Bool is_csv=false);

  spstrw MergeBackSlash(l FromPos=0,l *LastPos=0,const wchar_t *separator=0) const;
  l Compact(const wchar_t *FirstChar=(wchar_t*)L";",Bool Empty=true);

  friend StringsW<L>& operator<< (StringsW<L>&LL,const spstrw Value) {
    LL.Add(Value);
    return LL;
  }
  friend StringsW<L>& operator<< (StringsW<L>&LL,const wchar_t *Value) {
    LL.Add(spstrw(Value));
    return LL;
  }
#ifndef NO_LOADSAVESTRINGS
  /// leading FFFE/FEFF causes utf-16 loading, EFBBBF - utf-8, else ascii assumed
  Bool LoadFromFile(const char*, Bool is_csv=false);
  Bool SaveToFile(const char*, Bool unixmode=false, Bool utf8=false) const;
#endif

  Bool IsMember(const spstrw str) { return Str.IndexOf(str)>=0; }
  Bool IsObjectMember(const L& Value) { return LList.IndexOf(Value)>=0; }

  void Init() { LList.Init(); Str.Init(); }
};


template <class L>
StringsW<L>::StringsW(wchar_t **List, l Count) {
   LList.SetCount(Count);
   for (l ii=0;ii<Count;ii++) Str.Add(List[ii]);
}

template <class L>
void StringsW<L>::Insert(l Pos,spstrw Value,l Repeat) {
  l mx=LList.Count();
  LList.IncCount(Repeat);
  for (l ii=0;ii<Repeat;ii++) LList.Exchange(Pos+ii,mx+ii);
  Str.Insert(Pos,Value,Repeat);
}

template <class L>
spstrw StringsW<L>::GetTextToStr(const wchar_t *eol, l start, l count) const {
  l len=0,ii,inc=wcslen(eol);
  if (start<0) start=0;
  if (count<=0) count=Str.Count();
  count+=start;
  if (count>Str.Count()) count=Str.Count();
  for (ii=start;ii<count;ii++) len+=Str.Value()[ii].length()+inc;
  spstrw result;
  wchar_t  *str,*tmp=result.LockPtr(len+1);
  for (ii=start,str=tmp;ii<count;ii++) {
    memcpy(str,Str.Value()[ii](),Str.Value()[ii].length()<<1);
    str+=Str.Value()[ii].length();
    memcpy(str,eol,inc<<1); str+=inc;
  }
  result.UnlockPtr(str-tmp);
  return result;
}

template <class L>
spstrw StringsW<L>::MergeCmdLine(const wchar_t *separator, wchar_t quote) const {
  l len=0,ii,inc=wcslen(separator);
  for (ii=0;ii<Str.Count();ii++)
    len+=Str.Value()[ii].length()+inc+2+Str.Value()[ii].countchar('\"');
  spstrw result;
  wchar_t *str,*tmp=result.LockPtr(len+1);
  for (ii=0,str=tmp;ii<Str.Count();ii++) {
    spstrw qrep(Str.Value()[ii]);
    Bool nq=qrep.cpos(' ')>=0;
    qrep.replace((wchar_t*)L"\"",(wchar_t*)L"\\\"");
    if (nq) *str++=quote;
    memcpy(str,qrep(),qrep.length()<<1);
    str+=qrep.length();
    if (nq) *str++=quote;
    memcpy(str,separator,inc<<1); str+=inc;
  }
  result.UnlockPtr(str-tmp);
  return result;
}

template <class L>
wchar_t *StringsW<L>::GetText(const wchar_t *eol, l start, l count) const {
  l len=0,ii,inc=wcslen(eol);
  if (start<0) start=0;
  if (count<=0) count=Str.Count();
  count+=start;
  if (count>Str.Count()) count=Str.Count();
  for (ii=start;ii<count;ii++) len+=Str.Value()[ii].length()+inc;
#ifdef SPMALLOC_GETTEXT
  wchar_t *str,*tmp=(wchar_t*)spmalloc(len+1<<1);
#else
  wchar_t *str,*tmp=(wchar_t*)MALLOC(len+1<<1);
#endif
  for (ii=start,str=tmp;ii<count;ii++) {
    memcpy(str,Str.Value()[ii](),Str.Value()[ii].length()<<1);
    str+=Str.Value()[ii].length();
    memcpy(str,eol,inc<<1); str+=inc;
  }
  *str=0;
  return tmp;
}

template <class L>
void StringsW<L>::SetText(const char *Text,l len,Bool is_csv,Bool is_utf8) {
  if (len<=0) len=strlen(Text);
  const char *txt;
  Clear();
  d NL = is_csv?FFFF:0x0A;

  while (len) {
    txt=Text;
    while (len&&*txt!=NL&&*txt!=0x0D&&*txt!=0x0C) { len--; txt++; }

    l linelen=txt-Text;
    if (is_utf8&&linelen) {
      spstrw line;
      line.setutf8(Text,linelen);
      Add(line);
    } else
      Add((!linelen?spstrw():spstrw(Text,linelen))());

    if (len&&*txt==0x0D&&*(txt+1)==0x0A) { len--; txt++; }
    if (!len) break;
    len--; Text=++txt;
  }
}

template <class L>
void StringsW<L>::SetTextW(const wchar_t *Text,l len,Bool is_csv) {
  if (len<=0) len=wcslen(Text);
  const wchar_t *txt;
  Clear();
  d NL = is_csv?FFFF:0x0A;
  while (len) {
    txt=Text;
    while (len&&*txt!=NL&&*txt!=0x0D&&*txt!=0x0C) { len--; txt++; }
    Add((txt==Text?spstrw():spstrw(Text,txt-Text))());
    if (len&&*txt==0x0D&&*(txt+1)==0x0A) { len--; txt++; }
    if (!len) break;
    len--; Text=++txt;
  }
}

template <class L>
l StringsW<L>::IndexOfName(const wchar_t *SName,l StartFrom) {
  for (l ii=StartFrom;ii<=Str.Max();ii++)
    if (_wcsicmp(SName,Name(ii)())==0) return ii;
  return -1;
}

template <class L>
l StringsW<L>::IndexOfICase(const wchar_t *SName,l StartFrom) {
  for (l ii=StartFrom;ii<=Str.Max();ii++)
    if (_wcsicmp(SName,Str[ii]())==0) return ii;
  return -1;
}

template <class L>
l StringsW<L>::IndexOfName(const spstrw &SName,l StartFrom) {
  spstrw sName(SName);
  sName.upper();
  for (l ii=StartFrom;ii<=Str.Max();ii++)
    if (Name(ii).upper()==sName) return ii;
  return -1;
}

template <class L>
void StringsW<L>::Sort(Bool ICase) {
  if (Str.Count()>1) {
    l iCnt=Str.Count(),iOffset=iCnt/2;
    while (iOffset) {
      l iLimit=iCnt-iOffset, iSwitch;
      do {
        iSwitch=false;
        for(l iRow=0;iRow<iLimit;iRow++) {
          if ((ICase?_wcsicmp:wcscmp)
            (Str.Value()[iRow](),Str.Value()[iRow+iOffset]())>0)
          {
            Exchange(iRow,iRow+iOffset);
            iSwitch=iRow;
          }
        }
      } while (iSwitch);
      iOffset=iOffset/2;
    }
  }
}

#ifdef SORT_BY_OBJECTS_ON
template <class L>
void StringsW<L>::SortByObjects(Bool Forward) {
  if (LList.Count()>1) {
    l iCnt=LList.Count(),iOffset=iCnt/2;
    while (iOffset) {
      l iLimit=iCnt-iOffset, iSwitch;
      do {
        iSwitch=false;
        for(l iRow=0;iRow<iLimit;iRow++) {
          if (Forward) {
            if (LList.Value()[iRow]>LList.Value()[iRow+iOffset]) { Exchange(iRow,iRow+iOffset); iSwitch=iRow; }
          } else {
            if (LList.Value()[iRow]<LList.Value()[iRow+iOffset]) { Exchange(iRow,iRow+iOffset); iSwitch=iRow; }
          }
        }
      } while (iSwitch);
      iOffset=iOffset/2;
    }
  }
}
#endif

template <class L>
l StringsW<L>::TrimEmptyLines(Bool EndOnly) {
  l delcnt=0;
  if (EndOnly) {
    l ii=Str.Max();
    while (ii>=0) {
      spstrw st(Str[ii]);
      if (st.trim().length()==0) {
        Str.Delete(ii);
        LList.Delete(ii--);
        delcnt++;
      } else break;
    }
  } else {
    l ii=0;
    while (ii<=Str.Max()) {
      spstrw st(Str[ii]);
      if (st.trim().length()==0) {
        Str.Delete(ii);
        LList.Delete(ii);
        delcnt++;
      } else ii++;
    }
  }
  // return number of deleted lines
  return delcnt;
}

template <class L>
void StringsW<L>::TrimAllLines(l StartFrom) {
  for (l ii=StartFrom;ii<=Str.Max();ii++) Str[ii].trim();
  if (Str.GetIndexOfCache()>=StartFrom) Str.GetIndexOfCache()=-1;
}

template <class L>
l StringsW<L>::SplitString(const spstrw &ss, const wchar_t *separators) {
  Clear();
  // " \t" or "\t " pair has a special handling (Zero=true)
  Bool Zero=true;
  const wchar_t *cc=separators;
  while (*cc)
    if (*cc!=' '&&*cc!='\t') { Zero=false; break; } else cc++;
  cc=ss();
  if (!cc||!*cc) return 0;
  wchar_t *prev,*str=(wchar_t*)cc;
  // skip space separators at start...
  if (Zero) while (*str&&wcschr(separators,*str)) str++;
  if (*str) {
    str--;
    if (separators[1]==0) // small optimization
    do {
      prev=++str;
      str=wcschr(prev,*separators);
      spstrw toadd;
      if (str!=prev) toadd=spstrw(prev,str?str-prev:0);
      Str.Add(toadd);
      if (str&&Zero) while (wcschr(str+1,*separators)==str+1) str++; // skip same chars
    } while (str);
      else
    do {
      prev=++str;
      str=wcspbrk(prev,separators);
      spstrw toadd;
      if (str!=prev) toadd=spstrw(prev,str?str-prev:0);
      Str.Add(toadd);
      if (str&&Zero) while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
    } while (str);
  } else
  if (Zero) Str.Add(spstrw());
  if (Str.Count()) LList.IncCount(Str.Count());
  return Str.Count();
}

template <class L>
l StringsW<L>::ParseCmdLine(const spstrw &ss, wchar_t separator) {
  const wchar_t *pp=ss();
  spstrw tmp;
  wchar_t *al=tmp.LockPtr(ss.length()+260), *cl=al;
  int qout=0,argc=0;
  Clear();
  while (true) {
    if (*pp) while (*pp==separator||separator==' '&&*pp==9) pp++;
    if (*pp==0) break;
    argc++;
    while (true) {
      int copy    =1;
      int numslash=0;
      while (*pp=='\\') { ++pp; ++numslash; }
      if (*pp=='\"') {
        if ((numslash&1)==0) {
          if (qout) {
            if (pp[1]=='\"') pp++; else copy=0;
          } else copy=0;
          qout=!qout;
        }
        numslash>>=1; // divide numslash by two
      }
      while (numslash--) *al++='\\';
      if (*pp==0||(!qout&&(*pp==separator||separator==' '&&*pp==9))) break;
      if (copy) *al++=*pp;
      ++pp;
    }
    *al=0;
    Add(spstrw(cl));
    al=cl;
  }
  tmp.UnlockPtr();
  return argc;
}

template <class L>
spstrw StringsW<L>::Name(l idx) {
  const wchar_t *str=Str[idx]();
  wchar_t *ps=wcschr((wchar_t*)str,'=');
  if (ps) {
    int len=ps-str,st=0;
    while (st<len&&(str[st]==' '||str[st]==9)) { st++; len--; }
    while (st<len&&(str[len-1]==' '||str[len-1]==9)) len--;
    if (len) return spstrw(str+st,len);
  }
  return spstrw(Str[idx]);
}

template <class L>
spstrw StringsW<L>::Name(const spstrw &SValue) {
  for (l ii=0;ii<=Str.Max();ii++) {
    if (Value(ii)==SValue) return Name(ii);
  }
  return spstrw();
}

template <class L>
spstrw StringsW<L>::Value(const spstrw &SName) {
  l idx=IndexOfName(SName);
  return idx>=0?Value(idx):spstrw();
}

template <class L>
spstrw StringsW<L>::Value(const wchar_t *SName) {
  l idx=IndexOfName(SName);
  return idx>=0?Value(idx):spstrw();
}

template <class L>
int StringsW<L>::IntValue(const wchar_t *SName) {
  l idx=IndexOfName(SName);
  if (idx<0) return 0;
  l pos=Str[idx].cpos('=');
  if (pos<0) return 0;
  return Str[idx].Int(pos+1);
}

template <class L>
d StringsW<L>::DwordValue(const wchar_t *SName) {
  l idx=IndexOfName(SName);
  if (idx<0) return 0;
  l pos=Str[idx].cpos('=');
  if (pos<0) return 0;
  return Str[idx].Dword(pos+1);
}

template <class L>
spstrw StringsW<L>::Value(l idx) {
  const wchar_t *str=Str[idx]();
  wchar_t *ps=wcschr((wchar_t*)str,'=');
  if (ps) return spstrw(ps+1); else return spstrw();
}

template <class L>
spstrw StringsW<L>::MergeBackSlash(l FromPos,l *LastPos,const wchar_t *separator) const {
  if (FromPos>Str.Max()) return spstrw();
  l len=0,ii=FromPos,ok=0,
    seplen=separator?wcslen(separator):0;
  while (Str.Value()[ii].lastchar()=='\\'&&ii<Str.Count())
    len+=Str.Value()[ii++].length()-1+seplen;
  len+=Str.Value()[ii].length()+seplen;
  spstrw res;
  wchar_t *cc=res.LockPtr(len),*csave=cc;
  ii=FromPos;
  do {
    const spstrw *str=&Str.Value()[ii++];
    l first=0,last=str->length()-1;
    while ((*str)()[first]==' ') first++;
    while ((*str)()[last]==' '&&last) last--;
    if ((*str)()[last]!='\\') ok=1; else
      if (--last>0)
        while ((*str)()[last]==' '&&last) last--;
    if (last>=first) {
      l cnt=last-first+1;
      memcpy(cc,(*str)()+first,cnt<<1);
      cc+=cnt;
      if (!ok&&separator) {
        memcpy(cc,separator,seplen<<1);
        cc+=seplen;
      }
    }
  } while (!ok);
  if (LastPos) *LastPos=ii-1;
//  *cc=0;
  res.UnlockPtr(cc-csave);
  return res;
}

template <class L>
l StringsW<L>::Compact(const wchar_t *FirstChar,Bool Empty) {
  l ii=0,rc=Str.Count(),st=-1,ok=0,sch=FirstChar&&*FirstChar;
  while (Str.Count()>ii) {
    const wchar_t *str=Str[ii]();
    if (!str) ok=Empty; else
    if (*str==' '||*str=='\x9') {
      int ps=1;
      while (str[ps]&&(str[ps]==' '||str[ps]=='\x9')) ps++;
      ok=str[ps]==0;
    } else
    if (sch)
      if (wcschr(FirstChar,*str)) ok=1;
    if (ok&&st<0) st=ii; else
    if (!ok&&st>=0) {
      int cnt=ii-st;
      Str.Delete(st,cnt);
      LList.Delete(st,cnt);
      ii-=cnt;
    }
    ii++;
  }
  return rc-Str.Count();
}

template <class L>
l StringsW<L>::IndexOf(const spstrw &Value,l StartFrom) {
  l idx=Str.GetIndexOfCache();
  if (idx>=StartFrom)
    if (Value.hash()==Str[idx].hash())
      if (Value==Str[idx]) return idx;
  for (l ii=StartFrom;ii<=Str.Max();ii++)
    if (Str[ii].hash()==Value.hash())
      if (Str[ii]==Value) {
        Str.GetIndexOfCache()=StartFrom?-1:ii;
        return ii;
      }
  return -1;
}

#ifndef NO_LOADSAVESTRINGS
template <class L>
Bool StringsW<L>::LoadFromFile(const char *FName,Bool is_csv) {
  Clear();
#ifdef STRW_HRTIO
  d     rf=fOpen(FName,"rbdws+");
#else
  FILE *rf=fopen(FName,"rb");
#endif
  if (rf) {
    d    fsz=STRW_fsize(rf),ii;
    char *cc=(char*)MALLOC(fsz);
    if (STRW_fread(rf,cc,fsz)==fsz) {
      // swap byte order
      if (fsz>=2 && *(w*)cc==0xFFFE)
        for (ii=0;ii<fsz>>1;ii++) ((w*)cc)[ii] = _swapword(((w*)cc)[ii]);
      // UTF-16, UTF-8 and ascii
      if (fsz>=2 && *(w*)cc==0xFEFF)
        SetTextW((wchar_t*)(cc+2), fsz-2>>1, is_csv);
      else {
        Bool utf8=fsz>=3&&b(cc[0])==0xEF&&b(cc[1])==0xBB&&b(cc[2])==0xBF;
        if (utf8) fsz-=3;
        SetText(cc+(utf8?3:0),fsz,is_csv,utf8);
      }
    }
    FREE(cc);
    STRW_fclose(rf);
  }
  return rf!=0;
}

template <class L>
Bool StringsW<L>::SaveToFile(const char *FName, Bool unixmode, Bool utf8) const {
#ifdef STRW_HRTIO
  d     rf=fOpen(FName,"wbd+s+");
#else
  FILE *rf=fopen(FName,"wb");
#endif
  Bool res=false;
  if (rf) {
    if (utf8) {
      const b utf8sig[3] = {0xEF,0xBB,0xBF};
      res=STRW_fwrite(rf,&utf8sig,3)==3;
      if (res) {
        spstrw txt=GetTextToStr((wchar_t*)(unixmode?L"\n":L"\r\n"));
        spstr  out=txt.getutf8();
        txt.clear();
        res=STRW_fwrite(rf,(const ptr)(out()),out.length())==(size_t)out.length();
      }
    } else {
      // works in both orders
      w utf16sig = 0xFEFF;
      res=STRW_fwrite(rf,&utf16sig,2)==2;
      if (res) {
        spstrw txt=GetTextToStr((wchar_t*)(unixmode?L"\n":L"\r\n"));
        res=STRW_fwrite(rf,(const ptr)(txt()),txt.length()*2)==(size_t)txt.length()*2;
      }
    }
    STRW_fclose(rf);
  }
  return res;
}
#endif // NO_LOADSAVESTRINGS

typedef StringsW<l> TStringsW;
typedef StringsW<Pointer> TPtrStringsW;

#endif // SP_STRINGSW
