//
// QSINIT "fslib" module
// EXT2/3/4 structs
//
#ifndef QSINIT_EXT2INFO
#define QSINIT_EXT2INFO

#include "qstypes.h"

#ifdef __cplusplus
extern "C" {
#endif
#pragma pack(1)

#define EXT4_SUPER_MAGIC             0xEF53
#define EXT4_ROOT_INO                     2
#define PATH_MAX                       1024

#define EXT4_SUPER_OFS                 1024

#define EXT4_MIN_BLOCK_SIZE            1024
#define EXT4_MAX_BLOCK_SIZE           65536
#define EXT4_MIN_BLOCK_LOG_SIZE          10
#define EXT4_MAX_BLOCK_LOG_SIZE          16
#define EXT4_MAX_CLUSTER_LOG_SIZE        30

typedef struct {
   u32t                  inodes;  // inodes count
   u32t               blocks_lo;  // blocks count
   u32t             r_blocks_lo;  // reserved blocks count
   u32t          free_blocks_lo;  // free blocks count
   u32t             free_inodes;  // free inodes count
   u32t        first_data_block;  // first data block
   u32t          log_block_size;  // block size
   u32t         o_log_frag_size;  // obsoleted fragment size
   u32t        blocks_per_group;  // # blocks per group
   u32t       o_frags_per_group;  // obsoleted fragments per group
   u32t        inodes_per_group;  // # inodes per group
   u32t                   mtime;  // mount time
   u32t                   wtime;  // write time
   u16t               mnt_count;  // mount count
   u16t           max_mnt_count;  // maximal mount count
   u16t                   magic;  // magic signature
   u16t                   state;  // file system state
   u16t                  errors;  // behaviour when detecting errors
   u16t               minor_rev;  // minor revision level
   u32t               lastcheck;  // time of last check
   u32t           checkinterval;  // max. time between checks
   u32t              creator_os;  // OS
   u32t               rev_level;  // revision level
   u16t              def_resuid;  // default uid for reserved blocks
   u16t              def_resgid;  // default gid for reserved blocks
   // These fields are for EXT4_DYNAMIC_REV superblocks only.
   u32t               first_ino;  // first non-reserved inode
   u16t              inode_size;  // size of inode structure
   u16t          block_group_nr;  // block group # of this superblock
   u32t          feature_compat;  // compatible feature set
   u32t        feature_incompat;  // incompatible feature set
   u32t       feature_ro_compat;  // readonly-compatible feature set
   u8t                 uuid[16];  // 128-bit uuid for volume
   char         volume_name[16];  // volume name
   char        last_mounted[64];  // directory where last mounted
   u32t         al_usage_bitmap;  // compression
   // Performance hints.
   u8t          prealloc_blocks;  // # of blocks to try to preallocate
   u8t      prealloc_dir_blocks;  // # to preallocate for dirs
   u16t     reserved_gdt_blocks;  // Per group desc for online growth
   // Journaling support valid if EXT4_FEATURE_COMPAT_HAS_JOURNAL set.
   u8t         journal_uuid[16];  // uuid of journal superblock
   u32t            journal_inum;  // inode number of journal file
   u32t             journal_dev;  // device number of journal file
   u32t             last_orphan;  // start of list of inodes to delete
   u32t            hash_seed[4];  // HTREE hash seed
   u8t         def_hash_version;  // default hash version to use
   u8t                 reserved;
   u16t               desc_size;  // size of group descriptor
   u32t      default_mount_opts;
   u32t           first_meta_bg;  // first metablock block group
   u32t               mkfs_time;  // when the filesystem was created
   u32t          jnl_blocks[17];  // backup of the journal inode
   // 64bit support valid if EXT4_FEATURE_COMPAT_64BIT
   u32t               blocks_hi;  // blocks count
   u32t             r_blocks_hi;  // reserved blocks count
   u32t          free_blocks_hi;  // free blocks count
   u16t         min_extra_isize;  // all inodes have at least # bytes
   u16t        want_extra_isize;  // new inodes should reserve # bytes
   u32t                   flags;  // miscellaneous flags
   u16t             raid_stride;  // RAID stride
   u16t            mmp_interval;  // # seconds to wait in MMP checking
   u64t               mmp_block;  // block for multi-mount protection
   u32t       raid_stripe_width;  // blocks on all data disks (N*stride)*/
   u8t      log_groups_per_flex;  // FLEX_BG group size
   u8t                reserved2;
   u16t               reserved3;
   u64t          kbytes_written;  // nr of lifetime kilobytes written
   u32t          reserved4[160];  // Padding to the end of the block
} ext4_superblock;

// Codes for operating systems
#define EXT4_OS_LINUX       0
#define EXT4_OS_HURD        1
#define EXT4_OS_MASIX       2
#define EXT4_OS_FREEBSD     3
#define EXT4_OS_LITES       4

// Revision levels
#define EXT4_GOOD_OLD_REV   0   // The good old (original) format
#define EXT4_DYNAMIC_REV    1   // V2 format with dynamic inode sizes

#define EXT4_GOOD_OLD_INODE_SIZE 128

// Feature set definitions
#define EXT4_FEATURE_COMPAT_DIR_PREALLOC      0x00001
#define EXT4_FEATURE_COMPAT_IMAGIC_INODES     0x00002
#define EXT4_FEATURE_COMPAT_HAS_JOURNAL       0x00004
#define EXT4_FEATURE_COMPAT_EXT_ATTR          0x00008
#define EXT4_FEATURE_COMPAT_RESIZE_INODE      0x00010
#define EXT4_FEATURE_COMPAT_DIR_INDEX         0x00020
#define EXT4_FEATURE_COMPAT_SPARSE_SUPER2     0x00200
#define EXT4_FEATURE_COMPAT_FAST_COMMIT       0x00400
#define EXT4_FEATURE_COMPAT_STABLE_INODES     0x00800
#define EXT4_FEATURE_COMPAT_ORPHAN_FILE       0x01000 // orphan file exists

#define EXT4_FEATURE_RO_COMPAT_SPARSE_SUPER   0x00001
#define EXT4_FEATURE_RO_COMPAT_LARGE_FILE     0x00002
#define EXT4_FEATURE_RO_COMPAT_BTREE_DIR      0x00004
#define EXT4_FEATURE_RO_COMPAT_HUGE_FILE      0x00008
#define EXT4_FEATURE_RO_COMPAT_GDT_CSUM       0x00010
#define EXT4_FEATURE_RO_COMPAT_DIR_NLINK      0x00020
#define EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE    0x00040
#define EXT4_FEATURE_RO_COMPAT_QUOTA          0x00100
#define EXT4_FEATURE_RO_COMPAT_BIGALLOC       0x00200
#define EXT4_FEATURE_RO_COMPAT_METADATA_CSUM  0x00400
#define EXT4_FEATURE_RO_COMPAT_READONLY       0x01000
#define EXT4_FEATURE_RO_COMPAT_PROJECT        0x02000
#define EXT4_FEATURE_RO_COMPAT_VERITY         0x08000
#define EXT4_FEATURE_RO_COMPAT_ORPHAN_PRESENT 0x10000 // orphan file may be non-empty

#define EXT4_FEATURE_INCOMPAT_COMPRESSION     0x00001
#define EXT4_FEATURE_INCOMPAT_FILETYPE        0x00002
#define EXT4_FEATURE_INCOMPAT_RECOVER         0x00004 // needs recovery
#define EXT4_FEATURE_INCOMPAT_JOURNAL_DEV     0x00008 // journal device
#define EXT4_FEATURE_INCOMPAT_META_BG         0x00010
#define EXT4_FEATURE_INCOMPAT_EXTENTS         0x00040 // extents support
#define EXT4_FEATURE_INCOMPAT_64BIT           0x00080
#define EXT4_FEATURE_INCOMPAT_MMP             0x00100
#define EXT4_FEATURE_INCOMPAT_FLEX_BG         0x00200
#define EXT4_FEATURE_INCOMPAT_EA_INODE        0x00400 // EA in inode
#define EXT4_FEATURE_INCOMPAT_DIRDATA         0x01000 // data in dirent
#define EXT4_FEATURE_INCOMPAT_CSUM_SEED       0x02000
#define EXT4_FEATURE_INCOMPAT_LARGEDIR        0x04000 // >2GB or 3-lvl htree
#define EXT4_FEATURE_INCOMPAT_INLINE_DATA     0x08000 // data in inode
#define EXT4_FEATURE_INCOMPAT_ENCRYPT         0x10000
#define EXT4_FEATURE_INCOMPAT_CASEFOLD        0x20000

// Ext4 directory file types. Only the low 3 bits are used
#define EXT4_FT_UNKNOWN         0
#define EXT4_FT_REG_FILE        1
#define EXT4_FT_DIR             2
#define EXT4_FT_CHRDEV          3
#define EXT4_FT_BLKDEV          4
#define EXT4_FT_FIFO            5
#define EXT4_FT_SOCK            6
#define EXT4_FT_SYMLINK         7
#define EXT4_FT_MAX             8

// * --- inode --- *

#define EXT4_NDIR_BLOCKS        12
#define EXT4_IND_BLOCK          EXT4_NDIR_BLOCKS
#define EXT4_DIND_BLOCK         (EXT4_IND_BLOCK + 1)
#define EXT4_TIND_BLOCK         (EXT4_DIND_BLOCK + 1)
#define EXT4_N_BLOCKS           (EXT4_TIND_BLOCK + 1)

#define EXT4_SECRM_FL           0x00000001 // secure deletion
#define EXT4_UNRM_FL            0x00000002 // undelete
#define EXT4_COMPR_FL           0x00000004 // compress file
#define EXT4_SYNC_FL            0x00000008 // synchronous updates
#define EXT4_IMMUTABLE_FL       0x00000010 // immutable file
#define EXT4_APPEND_FL          0x00000020 // writes to file may only append
#define EXT4_NODUMP_FL          0x00000040 // do not dump file
#define EXT4_NOATIME_FL         0x00000080 // do not update atime
// Reserved for compression usage...
#define EXT4_DIRTY_FL           0x00000100
#define EXT4_COMPRBLK_FL        0x00000200 // one or more compressed clusters
#define EXT4_NOCOMPR_FL         0x00000400 // don't compress
#define EXT4_ECOMPR_FL          0x00000800 // compression error
// End compression flags --- maybe not all used
#define EXT4_INDEX_FL           0x00001000 // hash-indexed directory
#define EXT4_IMAGIC_FL          0x00002000 // aFS directory
#define EXT4_JOURNAL_DATA_FL    0x00004000 // file data should be journaled
#define EXT4_NOTAIL_FL          0x00008000 // file tail should not be merged
#define EXT4_DIRSYNC_FL         0x00010000 // dirsync behaviour (directories only)
#define EXT4_TOPDIR_FL          0x00020000 // top of directory hierarchies*/
#define EXT4_HUGE_FILE_FL       0x00040000 // set to each huge file
#define EXT4_EXTENTS_FL         0x00080000 // inode uses extents
#define EXT4_EA_INODE_FL        0x00200000 // inode used for large EA
#define EXT4_EOFBLOCKS_FL       0x00400000 // blocks allocated beyond EOF
#define EXT4_INLINE_DATA_FL     0x10000000 // inode has inline data
#define EXT4_PROJINHERIT_FL     0x20000000 // create with parents projid
#define EXT4_CASEFOLD_FL        0x40000000 // casefolded file
#define EXT4_RESERVED_FL        0x80000000 // reserved for ext4 lib


struct {
   u16t                    mode;  // file mode
   u16t                     uid;  // low 16 bits of Owner Uid
   u32t                 size_lo;  // size in bytes
   u32t                   atime;  // access time
   u32t                   ctime;  // change time
   u32t                   mtime;  // modification time
   u32t                   dtime;  // deletion Time
   u16t                     gid;  // low 16 bits of Group Id
   u16t             links_count;  // links count
   u32t               blocks_lo;  // blocks count
   u32t                   flags;  // file flags
   // OS dependent 1
   union {
      u32t              version;
      u32t           translator;
      u32t            reserved1;
   } osd1;
   u32t    block[EXT4_N_BLOCKS];  // pointers to blocks
   u32t              generation;  // file version (for NFS)
   u32t             file_acl_lo;  // file ACL
   u32t                 size_hi;
   u32t              obso_faddr;
   // OS dependent 2
   union {
      struct {
         u16t       blocks_high;
         u16t         mode_high; // or file_acl_high
         u16t          uid_high;
         u16t          gid_high;
         u32t            author;
      };
      struct {
         u16t         reserved2;
         u16t     file_acl_high;
         u32t      reserved3[2];
      };
   } osd2;
   u16t             extra_isize;
   u16t               reserved4;
   u32t             ctime_extra;  // extra change time      (nsec << 2 | epoch)
   u32t             mtime_extra;  // extra modification time(nsec << 2 | epoch)
   u32t             atime_extra;  // extra access time      (nsec << 2 | epoch)
   u32t                  crtime;  // file creation time
   u32t            crtime_extra;  // extra creation time (nsec << 2 | epoch)
   u32t              version_hi;  // high 32 bits for 64-bit version
} ext4_inode;

#pragma pack()
#ifdef __cplusplus
}
#endif
#endif // QSINIT_EXT2INFO
