//
// QSINIT help module
//
#include "stdlib.h"
#include "qsbase.h"
#include "qsmodext.h"
#include "qshelp.h"

const char *selfname = MODNAME_HELP;

extern unsigned __stdcall _wc_static_init();
qshandle    cmux = 0;

// grab global help mutex
void help_lock(void) { if (cmux) mt_muxcapture(cmux); }
// release global help mutex
void help_unlock(void) { if (cmux) mt_muxrelease(cmux); }

int  register_class(void);

static void _std nofity_cb(sys_eventinfo *info) {
   if (!info || info->eventtype==SECB_MTMODE) {
      if (mt_muxcreate(0, "__help_mux__", &cmux) ||
         io_setstate(cmux, IOFS_DETACHED, 1))
            log_it(0, "help mutex init error!\n");
      return;
   }
   if (info->eventtype==SECB_LOWMEM && (info->info&QSMA_HIGH)==0)
      help_release();
}

void* malloc_local(u32t size) {
   void *res = malloc_thread(size);
   if (res) mem_localblock(res);
   return res;
}

unsigned __cdecl LibMain(unsigned hmod, unsigned termination) {
   if (!termination) {
      u32t mt;
      // second load?
      if (mod_query(selfname, MODQ_LOADED|MODQ_NOINCR)) {
         log_printf("%s already loaded!\n", selfname);
         return 0;
      }
      // init static classes (since we redefine default LibMain())
      _wc_static_init();
      // error registering shared class?
      if (!register_class()) return 0;
      // catch MT mode
      mt = mt_active() ? 0 : SECB_MTMODE;
      // ask for notifications
      sys_notifyevent(mt|SECB_LOWMEM|SECB_GLOBAL, nofity_cb);
      // MT mode is ON now
      if (!mt) nofity_cb(0);
      // install shell command
      cmd_shelladd("HELP", shl_help);
      // exec shell commands on load
      cmd_execint(selfname, 0);
      // set system flag on self to prevent unloading
      ((module*)hmod)->flags |= MOD_SYSTEM;
   }
   return 1;
}
