//
// QSINIT help module
// PC test executable
// ------------------------------------------
// help         - help_info() test
// help topic   - help_draw() test
//
#ifdef __OS2__
#define INCL_VIO
#include <os2.h>
#endif
#include "qstypes.h"
#include "sp_ini.h"
#include "../hc/qshelp.h"

#define MAX_LLEN 512
qsh_file hf = 0;

u32t screen_width() {
#ifdef __OS2__
   VIOMODEINFO info;
   info.cb = sizeof(VIOMODEINFO);
   VioGetMode(&info, 0);
   return info.col;
#elif defined(__NT__)
   CONSOLE_SCREEN_BUFFER_INFO sbInfo;
   GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &sbInfo);
   return sbInfo.dwSize.X;
#else
   return 80;
#endif
}

int print_buffer(qsh_topic *out) {
   const u32t x = 0,
              y = 3;
   u32t     wdt = out->width>MAX_LLEN ? MAX_LLEN : out->width;
#ifdef __OS2__
   for (u32t ln=0; ln<out->lines; ln++) {
      u8t *sp = (u8t*)(out->text + ln*out->width);
      VioWrtCellStr((PCH)sp, wdt*2, y+ln, x, 0);
   }
   return 1;
#elif defined(__NT__)
   static HANDLE outh = 0;
   static COORD  from = {0,0};
   if (!outh) outh = GetStdHandle(STD_OUTPUT_HANDLE);

   for (u32t ln=0; ln<out->lines; ln++) {
      CHAR_INFO  cbuf[MAX_LLEN],
                 *cbp = cbuf;
      SMALL_RECT   to = {x, y+ln, x+out->width-1, y+ln};
      u8t         *sp = (u8t*)(out->text + ln*out->width);

      for (u32t ii=0; ii<wdt; ii++,cbp++) {
         cbp->Char.AsciiChar = *sp++;
         cbp->Attributes     = *sp++;
      }
      COORD bsize = {wdt,1};
      WriteConsoleOutput(outh, cbuf, bsize, from, &to);
   }
   return 1;
#else
   return 0;
#endif
}

char* _std buftoansi(u16t *src, u32t srcwidth, u32t width, u32t lines) {
   if (!src || !srcwidth || !width || !lines) return 0;
   // fix width
   if (width > srcwidth) width = srcwidth;
   spstr out;
   for (u32t ln=0, col=FFFF; ln<lines; ln++) {
      // per-line loop
      u16t *lb = src + srcwidth * ln;

      for (u32t xx=0; xx<width; xx++,lb++) {
         spstr ln;
         u8t attr = *lb>>8;
         if (attr!=col) {
            static u8t  cvt[8] = {0,4,2,6,1,5,3,7};
            char      cstr[80], *cp = cstr;
            *cp++ = '\x1B';
            *cp++ = '[';

            switch (attr) {
               case 0x0F: strcpy(cp,"0;1"); break;
               case 0x07: *cp++='0'; *cp=0; break;
               case 0x70: *cp++='7'; *cp=0; break;
               case 0xF0: strcpy(cp,"0;5"); break;
               default: {
                  int reset = col==FFFF || (col&8)&&!(attr&8) ||
                             (col&0x80)&&!(attr&0x80);
                  *cp = 0;
                  if (reset) strcpy(cp,"0;");
                  if ((attr&8) && (reset || !(col&8))) strcat(cp,"1;");
                  if ((attr&0x80) && (reset || !(col&0x80))) strcat(cp,"5;");
                  cp += strlen(cp);
                  if (reset || (attr&7)!=(col&7))
                     cp += sprintf(cp, "%u;", 30+cvt[attr&7]);
                  if (reset || (attr&0x70)!=(col&0x70))
                     cp += sprintf(cp, "%u;",  40+cvt[attr>>4&7]);
                  // drop last ';'
                  *--cp = 0;
               }
            }
            if (*cp) cp+=strlen(cp);
            *cp++ = 'm';
            *cp++ = 0;
            out  += cstr;
            col   = attr;
         }
         out+=(char)*lb;
      }
      out+='\n';
   }
   return strdup(out());
}

void errexit(spstr err) {
   printf("%s\n", err());
   exit(1);
}

int main(int argc, char *argv[]) {
   if (argc>1 && !strcmp(argv[1],"/?")) {
      printf("helptest             - show basic information\n"
             "helptest topic       - show topic\n"
             "helptest topic file  - save topic as an ANSI text file\n");
      return 1;
   }
   spstr self(argv[0]);
   l dot = self.crpos('.');
   if (dot>0) self.del(dot+1, x7FFF);
   self += "ini";
   TINIFile ini(self(), true);
   spstr hfname = ini.ReadStr("common", "helpfile").trim();
   if (!hfname) errexit("Help file name is missing in .ini");

   qserr err = help_open(hfname(), &hf);
   if (err)
      errexit(spstr().sprintf("Error %X opening help file \"%s\"", err, hfname()));

   qsh_info *hi = help_info(hf);
   if (!hi) errexit("Unable to get help file info");
   if (argc<2) {
      printf("File %s: %u topics, %u scheme(s), minw %u", hfname(),
         hi->ntopics, hi->nschemes, hi->minwidth);
      if (!hi->index) printf(", no index topic\n"); else
         printf(", index topic \"%s\"\n", hi->item[hi->index-1]);
   } else {
      u32t      id = help_topicid(hf, argv[1]);
      if (!id) errexit(spstr().sprintf("Unable to find topic \"%s\"", argv[1]));
      qserr    err;
      qsh_topic *t = help_draw(hf, 0, id, screen_width()-1, &err);
      if (!t)
         errexit(spstr().sprintf("Error %X processing topic \"%s\"", err, argv[1]));

      if (argc>=3) {
         char *text = buftoansi(t->text, t->width, t->width, t->lines);
         if (!text) errexit("Error in conversion to ANSI");

         TStrings out;
         out.SetText(text);
         free(text);

         if (!out.SaveToFile(argv[2]))
           errexit(spstr().sprintf("Error saving file \"%s\"", argv[2]));
      } else {
         printf("%u x %u text\n", t->width, t->lines);
         print_buffer(t);
      }

      free(t);
   }
   free(hi);


   help_close(hf);
   return 0;
}
