;
; QSINIT
; JFS boot sector code (micro-FSD)
;
;  * uses disk number from DL
;  * can read partition on 2Tb border ;) (i.e. only boot sector must have
;    32-bit sector number)
;  * incoming names must be in UPPER case (should be from any loader)
;  * supported sector size is fixed to 512 bytes (well, this can be fixed
;    for QSINIT, but not for OS/2)
;  * presence of OS2BOOT is not mandatory (QSINIT works fine without it, but
;    system boot will also fail)
;
                include inc/qstypes.inc
                include inc/parttab.inc
                include inc/filetab.inc
                include inc/seldesc.inc
                include jsegdef.inc

                NUM_RETRIES  = 4                                ;
                LOC_7C00h    = 7C00h                            ;
                BOOT_OEM_LEN = 11                               ; jmp + label

JCODE_TEXT      segment
                extrn   _mfsd_init:near
                extrn   _mfsd_open:near
                extrn   _mfsd_read:near
                extrn   _mfsd_close:near
JCODE_TEXT      ends

; this group should be linked as a separate object (to make offsets zero-based
; and use it freely over code!)
JCODE_START     segment
                assume  cs:JFSCODE, ds:JFSCODE, es:nothing, ss:JFSCODE

                org     0h

                public  jfs_bsect
                public  jfs_bsect_end
                public  jfs_bsname
                public  _code_nsec
                public  _bfname
                public  _mfsd_bits
jfs_bsect:
                jmp     @@bt_h_start                            ; BPB space
                nop                                             ;
                db      'QSBT 5.0'                              ;
                dw      200h                                    ;
                db      0                                       ;
                dw      0                                       ;
                db      0                                       ;
                dw      0                                       ;
                dw      0                                       ;
                db      0F0h                                    ;
                dw      0                                       ;
bpb_spt         dw      18                                      ; sector per track
bpb_numheads    dw      2                                       ;
bpb_ptstart     dd      0                                       ;
                dd      0                                       ;
bpb_disk        db      0h                                      ; 80h
                db      0                                       ;
                db      29h                                     ;
                dd      0                                       ;
                db      '           '                           ;
                db      'JFS     '                              ;
user_sp         dw      0                                       ;
user_ss         dw      0                                       ;
stack_pos       dw      0                                       ;
_code_nsec      db      0                                       ; # of whole sectors
fsd_mode        db      0                                       ;
i13x_on         db      0                                       ;
_mfsd_bits      db      BF_MICROFSD                             ;
@@bt_h_start:
                mov     si,LOC_7C00h                            ;
                xor     ax,ax                                   ;
                mov     ds,ax                                   ;
                mov     ss,ax                                   ;
                mov     sp,si                                   ;
                mov     [si+bpb_disk],dl                        ; update BPB

                int     12h                                     ; get memory for
                sub     ax,64 + SELF_SEG_LEN                    ; OS2LDR + micro
                and     ax,0FFF0h                               ; round down
                shl     ax,6                                    ; kb -> para
                mov     es,ax                                   ;
                cld                                             ;
                xor     di,di                                   ;
                mov     cx,100h                                 ;
            rep movsw                                           ;
                push    es                                      ;
                push    offset @@bt_h_home                      ;
                retf                                            ;
@@bt_h_home:
                mov     ds,ax                                   ;
                mov     ss,ax                                   ; point stack to the
                mov     sp,SELF_SEG_LEN shl 10                  ; end of own seg

                mov     ah,41h                                  ;
                mov     bx,55AAh                                ;
                mov     dl,bpb_disk                             ;
                int     13h                                     ;
                jc      @@bt_h_findname                         ;
                cmp     bx,0AA55h                               ;
                jnz     @@bt_h_findname                         ;
                and     cl,1                                    ;
                mov     i13x_on,cl                              ;
@@bt_h_findname:
                mov     bx,200h                                 ;
                xor     eax,eax                                 ; 2nd sector
                inc     al                                      ;
                mov     cx,offset jfs_bsect_end                 ; append own code
                add     cx,511                                  ;
                shr     cx,9                                    ;
                mov     _code_nsec,cl                           ;
                dec     cl                                      ;
                call    read_sector                             ; from next sectors

                cmp     word ptr _ftab,3                        ; old CHKDSK reverts all
                jz      main_code                               ; starting from sector 1
                mov     si,offset errmsg_badcode                ;
                jmp     _panic                                  ;
;----------------------------------------------------------------
panic_ioerr     label   near                                    ; read errror
                mov     si,offset errmsg_ioerr                  ;
                jmp     _panic                                  ;
                public  _panic_nofile,_panic                    ;
_panic_nofile   label   near                                    ; no OS2LDR
                mov     si,offset errmsg_nofile                 ;
_panic          label   near                                    ;
                cld                                             ;
                lodsb                                           ;
                or      al,al                                   ;
                jz      @@bt_h_err                              ;
                mov     ah,0Eh                                  ;
                mov     bx,7                                    ;
                int     10h                                     ;
                jmp     _panic                                  ;
@@bt_h_err:
                xor     ax,ax                                   ;
                int     16h                                     ;
                int     19h                                     ;
;----------------------------------------------------------------
; void _fmemcpy(void far *dst, const void far *src, u16t length);
                public  __fmemcpy
__fmemcpy       label   near                                    ;
                movzx   esp,sp                                  ;
                push    cx                                      ;
                mov     cx,[esp+2+10]                           ; length
                jcxz    @@fmemcpy0                              ;
                push    di                                      ;
                push    es                                      ;
                les     di,[esp+6+2]                            ; dst
                push    si                                      ;
                push    ds                                      ;
                lds     si,[esp+10+6]                           ; src
                cld                                             ;
                push    cx                                      ;
                shr     cx,2                                    ;
                jz      @@fmemcpyb                              ;
            rep movsd                                           ;
@@fmemcpyb:
                pop     cx                                      ;
                and     cx,3                                    ;
            rep movsb                                           ;
                pop     ds                                      ;
                pop     si                                      ;
                pop     es                                      ;
                pop     di                                      ;
@@fmemcpy0:
                pop     cx                                      ;
                ret     10                                      ;

;----------------------------------------------------------------
; read sector
; in    : eax - start, cl - count, es:bx - target
;         function cannot cross segment border!
; saves : dx, si, di, bp
; return: es:bx - points to the end of data
;         CF set on error (in micro-FSD mode only, else panic)
read_sector     label   near
@@bt_h_sloop:
                mov     ch,NUM_RETRIES                          ;
@@bt_h_rloop:
                pushad                                          ;
                mov     esi,bpb_ptstart                         ;
                xor     edx,edx                                 ;
                push    edx                                     ;
                mov     di,sp                                   ;
                cmp     i13x_on,dh                              ; cmp with 0
                jz      @@bt_h_no_ext_read                      ;
                add     eax,esi                                 ;
                rcl     edx,1                                   ; ;)
                push    edx                                     ;
                push    eax                                     ; pos to read
                push    es                                      ; target addr
                push    bx                                      ;
                xor     ch,ch                                   ;
                push    cx                                      ; # of sectors
                push    10h                                     ;
                mov     ah,42h                                  ;
                mov     si,sp                                   ;
                mov     [di],cl                                 ; read cl sectors
                jmp     @@bt_h_read_call                        ;
@@bt_h_no_ext_read:
                inc     byte ptr [di]                           ; read 1 sector
                add     eax,esi                                 ; sector #
                movzx   ecx,bpb_spt                             ;
                div     ecx                                     ;
                inc     dl                                      ;
                mov     cl,dl                                   ;
                mov     edx,eax                                 ;
                shr     edx,10h                                 ;
                div     bpb_numheads                            ;
                xchg    dl,dh                                   ;
                mov     ch,al                                   ;
                shl     ah,6                                    ;
                or      cl,ah                                   ;
                mov     ax,201h                                 ;
@@bt_h_read_call:
                mov     dl,bpb_disk                             ;
                int     13h                                     ;
                mov     sp,di                                   ;
                pop     ecx                                     ;
                jc      @@bt_h_read_err                         ;
                mov     di,sp                                   ;
                sub     word ptr [di].pa_ecx,cx                 ; upd. sector count
                add     [di].pa_eax,ecx                         ; and pos
                shl     cx,SECTSHIFT                            ; and offset in bx
                add     word ptr [di].pa_ebx,cx                 ;
                popad                                           ; esp restored here
                or      cl,cl                                   ;
                jnz     @@bt_h_sloop                            ;
                ret                                             ; CF cleared after "or"
@@bt_h_read_err:
                popad                                           ; esp restored here
                dec     ch                                      ; # retries
                jnz     @@bt_h_rloop                            ;
                or      ch,fsd_mode                             ; we`re in micro-FSD mode?
                jz      panic_ioerr                             ; error if still not
                stc                                             ;
                ret                                             ;

;----------------------------------------------------------------
; normalize pointer in di:si to seg:000x form (i.e. 64k-16 adressing)
normalize       label   near                                    ;
                push    ax                                      ;
                mov     ax,si                                   ;
                and     si,PARAMASK                             ;
                shr     ax,PARASHIFT                            ;
                add     di,ax                                   ;
                pop     ax                                      ;
                ret                                             ;

;----------------------------------------------------------------
; _printf16 char printing ...
; ah must be preserved
                public  prnchar                                 ;
prnchar         label   near                                    ;
                push    ax                                      ;
                mov     ah,0Eh                                  ;
                int     10h                                     ;
                pop     ax                                      ;
                ret                                             ;

; ---------------------------------------------------------------
errmsg_badcode  db      'Boot code is broken.',0                ;
errmsg_ioerr    db      'Disk i/o error.',0                     ;
errmsg_nofile   db      'Unable to find '                       ;
jfs_bsname:
_bfname:
                db      'OS2LDR',0,0,0,0,0,0,0,0,0,0            ; 15 chars + 0

@@bt_h_end:
bt_h_reserved   equ     510 - (@@bt_h_end - jfs_bsect)          ;
if bt_h_reserved GT 0
                db      bt_h_reserved dup (0)                   ;
endif
                dw      0AA55h                                  ;
main_code       label   near
                mov     et_museg,cs                             ; update filetable
                mov     et_muOpenSeg,cs                         ;
                mov     et_muReadSeg,cs                         ;
                mov     et_muCloseSeg,cs                        ;
                mov     et_muTermSeg,cs                         ;

                call    _mfsd_init                              ;

                mov     stack_pos,sp                            ; save stack pos
                mov     fsd_mode,1                              ; switch i/o to FSD mode

                push    BOOT_SEG                                ; loader address
                push    0                                       ;
                mov     dh,_mfsd_bits                           ; boot flags
                mov     dl,bpb_disk                             ;
                mov     si,BOOT_OEM_LEN                         ; ds:si - BPB
                push    cs                                      ;
                pop     es                                      ;
                mov     di,offset _ftab                         ; es:di - filetable
                retf                                            ; jump to the loader
; ---------------------------------------------------------------
; volume read
; in    : dx:ax - start sector, cl - count, di:si - target
;         function cannot cross segment border!
; out   : al - success (1/0)
; saves : bx
                public  _volread                                ;
_volread        label   near                                    ;
                push    bx                                      ;
                call    normalize                               ;
                mov     es,di                                   ;
                mov     bx,si                                   ;
                push    dx                                      ;
                push    ax                                      ;
                pop     eax                                     ;
                call    read_sector                             ;
                pop     bx                                      ;
                setnc   al                                      ;
                ret                                             ;
; ---------------------------------------------------------------
micro_open:
                mov     ax,offset _mfsd_open                    ;
                mov     cx,8                                    ; args size
                jmp     micro_start                             ;
micro_read:
                mov     ax,offset _mfsd_read                    ;
                mov     cx,12                                   ;
                jmp     micro_start                             ;
micro_close:
                mov     ax,offset _mfsd_close                   ;
                xor     cx,cx                                   ;
                jmp     micro_start                             ;
micro_terminate:
                retf                                            ;
micro_start:
                push    ebx                                     ; 20 bytes in user stack
                push    esi                                     ;
                push    edi                                     ;
                push    ebp                                     ;
                push    ds                                      ;
                push    es                                      ;
                push    fs                                      ;

                mov     dx,cs                                   ;
                mov     ds,dx                                   ;
                mov     es,dx                                   ; ds=es=ss -> self
                mov     bx,ss                                   ;
                mov     user_sp,sp                              ;
                mov     user_ss,bx                              ;
                cmp     bx,dx                                   ; these censored call
                jz      @@entry_no_stack_set                    ; us with OUR stack?
                mov     ss,dx                                   ; if not - switch it
                mov     sp,stack_pos                            ;
@@entry_no_stack_set:
                jcxz    @@entry_no_args                         ;
                mov     fs,bx                                   ; user ss
                mov     si,user_sp                              ;
                add     si,4+22                                 ; far call + pushes above
                sub     sp,cx                                   ;
                mov     di,sp                                   ;
                cld                                             ;
            rep movs    byte ptr es:[di],byte ptr fs:[si]       ;
@@entry_no_args:
                call    ax                                      ;
                lss     sp,dword ptr user_sp                    ; always restore ss:sp
                pop     fs                                      ;
                pop     es                                      ;
                pop     ds                                      ;
                pop     ebp                                     ;
                pop     edi                                     ;
                pop     esi                                     ;
                pop     ebx                                     ;
                retf                                            ;
; ---------------------------------------------------------------
; micro-fsd export table
                public  _ftab
_ftab           label   near
                dw      3
                dw      BOOT_SEG
et_ldrlen       dd      0
et_museg        dw      0
et_mulen        dd      SELF_SEG_LEN shl 10
                dw      MFSD_SEG
et_mfsdlen      dd      0
                dw      0
                dd      0
                dw      offset micro_open
et_muOpenSeg    dw      0
                dw      offset micro_read
et_muReadSeg    dw      0
                dw      offset micro_close
et_muCloseSeg   dw      0
                dw      offset micro_terminate
et_muTermSeg    dw      0
                dd      0
                dd      0
JCODE_START     ends

JCODE_END       segment
jfs_bsect_end   label   near
JCODE_END       ends
                end
