//
// QSINIT "partmgr" module
// JFS formatting code
//
#ifndef QSINIT_MKJFSH
#define QSINIT_MKJFSH

#include "qsbase.h"
#include <stdlib.h>
#define JFS_DMAP
#define JFS_LOGFORMAT
#include "sys/fs/jfs.h"
#include "pscan.h"
#include "rt/minmax.h"

#define MKJFS_DBLEVEL   0    // 1 - verbose, 2 - each 1000s cluster is bad

#define INIT_NUM_AGGR_INODES  (BADBLOCK_I + 1)

typedef struct xtree_buf {
  struct xtree_buf  *prev;  ///< parent (up)
  struct xtree_buf  *next;  ///< next rightmost child
  jfs_xtreepage     *page;
  int                 dyn;
} xtlist;

typedef struct {
   dmap_t      **maps,
           *emptypage;
   xtlist    *bbpages;  ///< inside j_write_bb() only
   u32t         nmaps;
   u32t       curdmap;  ///< inside j_write_bm() only
   u32t     lastalloc;  ///< max used cluster (to fill cpage.maxag later)
   dbmap_t      cpage;
} mapdata;


/// @name flags for j_io()
//@{
#define JIO_READ          0x01 ///< read operation
#define JIO_BO            0x02 ///< offset in bytes
#define JIO_BS            0x04 ///< size in bytes
//@}

qserr j_io       (disk_volume_data *di, u8t flags, u64t ofs, u64t size,
                  void *buf);

#define MARK_ALLOC  0x80000000
#define MARK_BAD   (0x40000000|MARK_ALLOC)
#define MARK_NMASK  0x3FFFFFFF

/** mark/free block in block map.
   @param clnum         cluster to mark
   @param flags         flags (MARK_*) and # of clusters
   @param pmap          map state/data buffer
   @return error code or 0 */
qserr j_mark     (mapdata *pmap, u32t clnum, u32t flags);

/** calc size of a block map and initialize memory and control page.
   
   Map file is ALLOCATED as a single extent of physical pages covering
   full control level (L2) tree control pages for the dmap pages required:
   the tree will be INITIALIZED to cover only the the dmap pages required;

   @param aggi          array of aggregate inodes
   @param sizecl        number of clusters
   @param clshift       bit shift of cluster size
   @param [out] agsize  allocation group size in clusters
   @param pmap          zero filled map state/data buffer
   @param stamp         inode stamp value to be used
   @return error code or 0 */
qserr j_calcmap  (jfs_dinode *aggi, u32t sizecl, u8t clshift, u32t *agsize,
                  mapdata *pmap, time_t stamp);

/** write the block map to disk.
   Update tree part of block map to match rest of map and then write the block
   map to disk. Also write the block map inode to disk.

   @param di            volume information
   @param pmap          map state/data buffer
   @param nclus         number of blocks (clusters)
   @return error code or 0 */
qserr j_write_bm (disk_volume_data *di, mapdata *pmap, u32t nclus);


/// write bad block list.
qserr j_write_bb (disk_volume_data *di, mapdata *pmap, u8t clshift,
                  u32t secvtab, jfs_dinode *bb_inode, dd_list bclist);

#define INO_TYPE_INLINE        0
#define INO_TYPE_EXTENT        1
#define INO_TYPE_MAXEXTENT     2
#define INO_TYPE_NODATA        3

/** initialize inode fields for an inode.
   @param ni            pointer to inode to be initialized
   @param fileset       fileset number
   @param inode         inode number
   @param clsize        # of allocated clusters
   @param size          size in bytes
   @param first_block   offset of first block of inode's extent
   @param mode          mode for inode
   @param type          type of inode to be initialized (INO_TYPE_*)
   @param inoext_addr   address of inode extent containing this inode
   @param clshift       bit shift of cluster size
   @param stamp         stamp used to identify inode as belonging to fileset */
void j_fill_inode(jfs_dinode *ni, u32t fileset, u32t inode, u32t clsize,
                  u64t size, u64t first_block, u32t mode, u32t type,
                  u64t inoext_addr, u8t clshift, u32t stamp);


/// read/write volume inode.
qserr j_rwinode  (disk_volume_data *di, jfs_dinode *ino, u32t inum, int write,
                  u32t secvtab, u8t clshift);

#endif // QSINIT_MKJFSH
