/* Notes:
   1. unique() call allocstr()
   2. allocstr() MUST BE CALLED BEFORE reallocstr()

   spstrw==char & spstrw!=char are also removed to not confuse ...
*/

#include <stdarg.h>
#include "sp_strw.h"
#include <string.h>
#include <stdlib.h>
#include "minmax.h"

static wchar_t emptystr[1] = {0};
wchar_t     *spnullstrw    = emptystr;

spstrw::spstrw(spstrw const &from,d __pos,d __n) { //from another string
#ifndef NO_SP_ASSERTS
  assert(!from.Lock);
#endif
  data=NULL;
  Lock=0;
  l tocopy = from.length()-(l)__pos;
  if (!from.data || tocopy<=0) return; else
  if (!__pos&&!__n) {
    from.data->usecnt++;
    data = from.data;
  } else
    initstring(from.data->buf+__pos, min(__n,(d)tocopy));
}

spstrw::spstrw(spstr const &from,d __pos,d __n) { // from ascii string
#ifndef NO_SP_ASSERTS
  assert(!from.Locked());
#endif
  data=NULL;
  Lock=0;
  l   slen = from.length(),
    tocopy = slen-(l)__pos;
  if (!slen || tocopy<=0) return;
  initstringa(from()+__pos, min(__n,(d)tocopy));
}


void spstrw::initstring(wchar_t const *str, d __n) {
  l len;
  if (__n&&str) {
    wchar_t*pos=(wchar_t*)memchrw((w*)str,0,__n);
    len=pos?pos-str:__n;
  } else len=str?wcslen(str):0;
  if (len>0) {
    allocstr(len);
    if (len) memcpy(data->buf,str,len<<1);
    *(data->buf+len)=0;
    data->chars=len;
    data->hash =FFFF;
  } else data=NULL;
}

// ascii only ...
static d atow(wchar_t *dst, const char *src, d maxlen=0) {
  d len=0;
  if (--maxlen)
    while (*src) {
      wchar_t ch = *src++;
      if (ch>127) ch='_';
      *dst++=ch;
      len++;
      if (--maxlen==0) break;
    }
  *dst=0;
  return len;
}

/* prepare ascii string for integer conversion (max len - 24 bytes)
   (because at least wcstoll() presence is questionable ;)) */
static void wtoai(char *dst, wchar_t *src) {
  d len=0;
  while (*src&&*src==' ') src++;
  // prevent octal conversion!
  while (*src=='0'&&src[1]>='0'&&src[1]<='9') src++;

  if (*src=='0' && (src[1]=='x' || src[1]=='X')) {
     *dst++ = '0';
     *dst++ = 'x';
     src+=2; len+=2;
     while (*src>='0'&&*src<='9' || *src>='a'&&*src<='f' || *src>='A'&&*src<='F') {
        *dst++=*src++;
        if (++len>=23) break;
     }
  } else
     while (*src>='0'&&*src<='9') {
        *dst++ = *src++;
        if (++len>=23) break;
     }
  *dst++ = 0; 
}

void spstrw::initstringa(char const *str, d __n) {
  l len;
  if (__n&&str) {
    char*pos=(char*)memchr(str,0,__n);
    len=pos?pos-str:__n;
  } else len=str?strlen(str):0;
  if (len>0) {
    allocstr(len);
    atow(data->buf, str, len+1);
    data->chars=len;
    data->hash =FFFF;
  } else data=NULL;
}

spstrw::spstrw(wchar_t const *str, l __n) {
  data=NULL;
  Lock=0;
  initstring(str,__n);
}

spstrw::spstrw(char const *str, l __n) {
  data=NULL;
  Lock=0;
  initstringa(str,__n);
}

spstrw::spstrw(wchar_t str[]) {
  data=NULL;
  Lock=0;
  initstring(str,0);
}

spstrw::spstrw(wchar_t ch, d __rep) {
  data=NULL;
  Lock=0;
  if (__rep>0) {
    allocstr(__rep);
    if (__rep>1) memsetw(data->buf,ch,__rep); else *data->buf=ch;
    *(data->buf+__rep)=0;
    data->chars=__rep;
    data->hash =d(ch)*__rep;
  }
}

wchar_t& spstrw::operator[] (l __pos) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  _unique();
  if (__pos>=0&&__pos<data->chars) return data->buf[__pos];
  /* return a _safe_ pointer to the zero char/any char after the last.
     Al least, overwrite here will be lost, but will be safe too */
  unique(data->chars+2);
  data->buf[data->chars+1] = 0;
  return data->buf[data->chars+1];
}

spstrw & spstrw::operator = (spstrw const &from) {
#ifndef NO_SP_ASSERTS
  assert(!Lock&&!from.Lock);
#endif
  if (from.data) from.data->usecnt++;    // don't reorder this two lines! ;)
  if (data&&--data->usecnt==0) disposestr();
  data=from.data;
  return *this;
}

spstrw & spstrw::operator = (wchar_t const *from)  {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&--data->usecnt==0) disposestr();
  data=0;
  initstring(from,0);
  return *this;
}

#ifndef NO_SPSTR_SPRINTF
spstrw& spstrw::sprintf(const wchar_t *fmt,...) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&--data->usecnt==0) disposestr();
  data=NULL;
  allocstr(wcslen(fmt)*5);
  va_list arglist;
  va_start(arglist,fmt);
  l sz;
  do {
    sz=_vsnwprintf(data->buf,data->maxnow-1,fmt,arglist);
    if (sz<0) reallocstr(data->maxnow*2);
  } while (sz<0);
  va_end(arglist);
  if (sz==data->maxnow-1) data->buf[sz]=0;
  data->chars=wcslen(data->buf);
  data->hash =FFFF;
  return *this;
}
#endif // NO_SPSTR_SPRINTF

spstrw& spstrw::operator= (int const ll) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&--data->usecnt==0) disposestr();
  data=NULL;
  char istr[32];
  ITOA(ll,istr,10);
  initstringa(istr);
  return *this;
}

spstrw& spstrw::operator= (d const ll) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&--data->usecnt==0) disposestr();
  data=NULL;
  char istr[32];
  UTOA(ll,istr,10);
  initstringa(istr);
  return *this;
}

spstrw & spstrw::operator += (spstrw const &from) {
#ifndef NO_SP_ASSERTS
  assert(!from.Lock);
#endif
  if (!from.data) return *this;
  if (!data) {
    data=from.data;
    data->usecnt++;
    return *this;
  }
  unique(data->chars+from.data->chars+1);
  memcpy(data->buf+data->chars,from.data->buf,from.data->chars+1<<1);
  data->chars+=from.data->chars;
  if (data->hash!=FFFF)
    if (from.data->hash!=FFFF) data->hash+=from.data->hash;
  return *this;
}

spstrw & spstrw::operator += (wchar_t const *from) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  l len=wcslen(from);
  unique((data?data->chars:0)+len);
  memcpy(data->buf+data->chars,from,len+1<<1);
  data->chars+=len;
  data->hash =FFFF;
  return *this;
}

spstrw& spstrw::operator+= (wchar_t cc) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  unique((data?data->chars:14)+1);
  data->buf[data->chars++]=cc;
  data->buf[data->chars]  = 0;
  if (data->hash!=FFFF) data->hash+=d(cc);
  return *this;
}

void spstrw::unique(d len) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) allocstr(len?len:10); else
  if (data->usecnt>1) {
    pstrinfow olddata;
    olddata=data; olddata->usecnt--;
    data=NULL;
#ifdef LOMEM_STR
    allocstr(len>olddata->chars?len:olddata->chars);
#else
    allocstr(len>olddata->maxnow?len:olddata->maxnow);
#endif
    memcpy(data->buf,olddata->buf,olddata->chars<<1);
    *(data->buf+olddata->chars)=0;
    data->chars=olddata->chars;
    data->hash =olddata->hash;
    if (!olddata->usecnt) FREE(olddata);
  } else
  if (len) reallocstr(len);
}

void spstrw::_unique() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) allocstr(10);
  if (data->usecnt>1) {
    pstrinfow olddata;
    olddata=data; olddata->usecnt--;
    data=NULL;
#ifdef LOMEM_STR
    allocstr(olddata->chars);
#else
    allocstr(olddata->maxnow);
#endif
    memcpy(data->buf,olddata->buf,olddata->chars<<1);
    *(data->buf+olddata->chars)=0;
    data->chars=olddata->chars;
    data->hash =olddata->hash;
    if (!olddata->usecnt) FREE(olddata);
  }
}

Bool operator == (spstrw const &s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock&&!s2.Lock);
#endif
  register strinfow *S1=s1.data, *S2=s2.data;
  if (S1==S2) return true;
  if (!S1||!S2) return s1.length()==s2.length();
//  if (S1->hash!=FFFF&&S2->hash!=FFFF&&S1->hash!=S2->hash) return false;
  return (S1->chars==S2->chars&&(wcscmp(S1->buf,S2->buf)==0));
}

Bool operator == (spstrw const &s1, wchar_t const *s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock);
#endif
//  if (s1.data) *(s1.data->buf+s1.data->chars)=0;
  return (wcscmp(s1(),s2)==0);
}

Bool operator == (wchar_t const *s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s2.Lock);
#endif
//  if (s2.data) *(s2.data->buf+s2.data->chars)=0;
  return (wcscmp(s2(),s1)==0);
}

Bool operator != (spstrw const &s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock&&!s2.Lock);
#endif
  if (!s1.data||!s2.data) return s1.length()!=s2.length();
  if (s1.data->hash!=FFFF&&s2.data->hash!=FFFF&&s1.data->hash!=s2.data->hash)
    return true;
//  *(s1.data->buf+s1.data->chars)=0;
//  *(s2.data->buf+s2.data->chars)=0;
  return (s1.data->chars!=s2.data->chars||wcscmp(s1.data->buf,s2.data->buf));
}

Bool operator != (spstrw const &s1, wchar_t const *s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock);
#endif
//  if (s1.data) *(s1.data->buf+s1.data->chars)=0;
  return wcscmp(s1(),s2)!=0;
}

Bool operator != (wchar_t const *s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s2.Lock);
#endif
//  if (s2.data) *(s2.data->buf+s2.data->chars)=0;
  return wcscmp(s2(),s1)!=0;
}

spstrw operator+(spstrw const &s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock&&!s2.Lock);
#endif
  l len=s1.length()+s2.length();
  spstrw str;
  if (!len) return str;
  str.unique(len);
  if (s1.length()) memcpy(str.data->buf,s1.data->buf,s1.length()<<1);
  if (s2.length()) memcpy(str.data->buf+s1.length(),s2.data->buf,s2.length()<<1);
  str.data->chars=len;
  str.data->hash =FFFF;
  str.data->buf[str.data->chars]=0;
  return str;
}

spstrw operator+(spstrw const &s1, wchar_t const *s2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock);
#endif
  if (*s2==0) return s1;
  l s2len=wcslen(s2),len=s1.length()+s2len;
  spstrw str;
  str.unique(len);
  if (s1.length()) memcpy(str.data->buf,s1.data->buf,s1.length()<<1);
  memcpy(str.data->buf+s1.length(),s2,s2len+1<<1);
  str.data->chars=len;
  str.data->hash =FFFF;
  return str;
}

spstrw operator+(wchar_t const *s1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s2.Lock);
#endif
  spstrw str(s1);
  return operator+(str,s2);
}

spstrw operator+(spstrw const &s1, wchar_t c2) {
#ifndef NO_SP_ASSERTS
  assert(!s1.Lock);
#endif
  char ch[2];
  ch[0]=c2; ch[1]=0;
  return operator+(s1,ch);
}

spstrw operator+(wchar_t c1, spstrw const &s2) {
#ifndef NO_SP_ASSERTS
  assert(!s2.Lock);
#endif
  spstrw ss(c1);
  return operator+(ss,s2);
}

l spstrw::pos(spstrw const &s,l startfrom) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock&&!s.Lock);
#endif
  if (!data||!s.data||data->chars-startfrom<s.data->chars) return -1; else {
//    *(data->buf+data->chars)=0;
//    *(s.data->buf+s.data->chars)=0;
    wchar_t *tmp=wcsstr(data->buf+startfrom,s.data->buf);
    return tmp?tmp-data->buf:-1;
  }
}

l spstrw::pos(wchar_t const *s,l startfrom) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return -1;
  if (startfrom<0||startfrom>=data->chars) return -1;
  wchar_t *tmp=wcsstr(data->buf+startfrom,s);
  return tmp?tmp-data->buf:-1;
}

// search for string by it begin and end
l spstrw::bepos(const wchar_t *begin,const wchar_t *end,l &epos,l startfrom) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock&&begin&&end);
#endif
  if (!data) return -1;
  if (startfrom<0||startfrom>=data->chars) return -1;
  startfrom=pos(begin,startfrom);
  if (startfrom<0) return -1;
  epos=pos(end,startfrom+1);
  if (epos<0) return -1;
  return startfrom;
}

l spstrw::cpos(wchar_t cc,l startfrom) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return -1;
  if (startfrom<0||startfrom>=data->chars) return -1;
  wchar_t *cp=wcschr(data->buf+startfrom,cc);
  return cp?cp-data->buf:-1;
}

l spstrw::crpos(wchar_t cc) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return -1;
  wchar_t *cp=wcsrchr(data->buf,cc);
  return cp?cp-data->buf:-1;
}

l spstrw::rpos(wchar_t const *ss) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return -1;
  d len=wcslen(ss);
  if (!len) return -1;
  wchar_t *res=NULL,*tmp=data->buf-len;
  do {
    tmp=wcsstr(tmp+len,ss);
    if (tmp) res=tmp;
  } while (tmp);
  return res?res-data->buf:-1;
}

spstrw& spstrw::upper() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  _unique();
  for (l ii=0;ii<data->chars;ii++) {data->buf[ii]=towupper(data->buf[ii]);}
  data->hash=FFFF;
  return *this;
}

spstrw& spstrw::lower() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  _unique();
  for (l ii=0;ii<data->chars;ii++) {data->buf[ii]=towlower(data->buf[ii]);}
  data->hash=FFFF;
  return *this;
}

spstrw& spstrw::capitalize() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  _unique();
  if (data->chars>1) data->buf[0]=towupper(data->buf[0]);
  data->hash=FFFF;
  return *this;
}

spstrw& spstrw::insert(const spstrw &s1,l pos) {
#ifndef NO_SP_ASSERTS
  assert(!Lock&&!s1.Lock);
#endif
  return insert(s1(),pos);
}

spstrw& spstrw::insert(const wchar_t *ss,l pos) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  l len=wcslen(ss);
  unique((data?data->chars:0)+len);
  if (pos<0) pos=0;
  if (pos>=data->chars) pos=data->chars; else
    memmove(data->buf+pos+len,data->buf+pos,data->chars-pos<<1);
  memcpy(data->buf+pos,ss,len<<1);
  data->chars+=len;
  *(data->buf+data->chars)=0;
  data->hash =FFFF;
  return *this;
}

#ifdef SP_BEBO
#define IsSpaceSeparator() (*(d*)separators==0x20000000|| \
  *(d*)separators==0x09000000|| *(d*)separators==0x20000900)
#else
#define IsSpaceSeparator() (*(d*)separators==' '||*(d*)separators=='\t'|| \
  *(d*)separators==0x090020)
#endif

l spstrw::words(const wchar_t *separators) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||!data->chars) return 0; else {
    l count=0;
//    *(data->buf+data->chars)=0;
    Bool    Zero=IsSpaceSeparator();
    wchar_t *str=data->buf;
    if (Zero) while (*str&&wcschr(separators,*str)) str++; // skip separators at start...
    if (!str) return 0;
    do {
      str=wcspbrk(str,separators);
      if (str&&Zero) {
        while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
      }
      count++;
    } while (str++);
    return count;
  }
}

spstrw spstrw::word(l ord, const wchar_t *separators) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||ord<=0||!data->chars) return spstrw();
  wchar_t *prev,*str=data->buf;
//  *(data->buf+data->chars)=0;
  Bool Zero=IsSpaceSeparator();
  if (Zero) while (*str&&wcschr(separators,*str)) str++;    // skip separators at start...
  if (!str) return spstrw();
  str--;
  do {
    prev=++str;
    str=wcspbrk(prev,separators);
    ord--;
    if (str&&Zero&&ord) while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
  } while (str&&ord);
  if (!ord&&str!=prev) return spstrw(*this,prev-data->buf, str?str-prev:0);
    else return spstrw();
}

l spstrw::word_Int(l ord,const wchar_t *separators) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||ord<=0||!data->chars) return 0;
  wchar_t *prev,*str=data->buf;
  Bool Zero=IsSpaceSeparator();
  if (Zero) while (*str&&wcschr(separators,*str)) str++;    // skip separators at start...
  if (!str) return 0;
  str--;
  do {
    prev=++str;
    str=wcspbrk(prev,separators);
    if (str&&Zero) while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
    ord--;
  } while (str&&ord);
  if (!ord) return Int(prev-data->buf); else return 0;
}

d spstrw::word_Dword(l ord,const wchar_t *separators) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||ord<=0||!data->chars) return 0;
  wchar_t *prev,*str=data->buf;
  Bool Zero=IsSpaceSeparator();
  if (Zero) while (*str&&wcschr(separators,*str)) str++;    // skip separators at start...
  if (!str) return 0;
  str--;
  do {
    prev=++str;
    str=wcspbrk(prev,separators);
    if (str&&Zero) while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
    ord--;
  } while (str&&ord);
  if (!ord) return Dword(prev-data->buf); else return 0;
}

l spstrw::wordpos(l ord, const wchar_t *separators) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||ord<=0||!data->chars) return -1;
  wchar_t *prev,*str=data->buf;
//  *(data->buf+data->chars)=0;
  Bool Zero=IsSpaceSeparator();
  if (Zero) while (*str&&wcschr(separators,*str)) str++;    // skip separators at start...
  if (!str) return spstrw();
  str--;
  do {
    prev=++str;
    str=wcspbrk(prev,separators);
    if (str&&Zero) while (wcspbrk(str+1,separators)==str+1) str++; // skip same chars
    ord--;
  } while (str&&ord);
  if (!ord) return prev-data->buf; else return -1;
}

spstrw& spstrw::replaceword(l ord,const wchar_t *separators,const wchar_t *str) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  l pos=wordpos(ord,separators);
  if (pos>=0) {
    if (word(ord,separators)!=str) {
      _unique();
      l len=wcspbrk(data->buf+pos,separators)-data->buf-pos;
      if (len) del(pos,len);
      insert(str,pos);
    }
  }
  return *this;
}

spstrw& spstrw::trim() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  l cc=0;
  while (*(data->buf+cc)==' '&&cc<data->chars) cc++;
  if (cc) {
    _unique();
    memmove(data->buf,data->buf+cc,data->chars-cc<<1);
    data->chars-=cc;
    while (data->chars&&*(data->buf+data->chars-1)==' ') data->chars--;
    *(data->buf+data->chars)=0;
    data->hash=FFFF;
  } else {
    cc=data->chars;
    while (cc&&*(data->buf+cc-1)==' ') cc--;
    if (cc!=data->chars) {
      _unique();
      data->chars=cc;
      *(data->buf+cc)=0;
      data->hash=FFFF;
    }
  }
  return *this;
}

spstrw& spstrw::trimleft() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  l cc=0;
  while (*(data->buf+cc)==' '&&cc<data->chars) cc++;
  if (cc) {
    _unique();
    memmove(data->buf,data->buf+cc,data->chars-cc<<1);
    data->chars-=cc;
    *(data->buf+data->chars)=0;
    data->hash=FFFF;
  }
  return *this;
}

spstrw& spstrw::trimright() {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  l cc=data->chars;
  while (cc&&*(data->buf+cc-1)==' ') cc--;
  if (cc!=data->chars) {
    _unique();
    data->chars=cc;
    *(data->buf+cc)=0;
    data->hash=FFFF;
  }
  return *this;
}

spstrw& spstrw::expandtabs(int alignpos) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  if (!data->chars) return *this;

  wchar_t *cp=wcschr(data->buf,'\t');
  l ps=cp?cp-data->buf:-1;
  if (ps>=0) {
    _unique();
    data->hash=FFFF;
    do {
      data->buf[ps++]=' ';
      l len=RoundN(alignpos,ps)-ps;
      if (len) {
        if (data->chars+len>data->maxnow) reallocstr(data->chars+alignpos*4);

        memmove(data->buf+ps+len,data->buf+ps,data->chars-ps<<1);
        memsetw(data->buf+ps,' ',len);
        data->chars+=len;
        data->buf[data->chars]=0;
      }
      cp=wcschr(data->buf+ps+len,'\t');
      ps=cp?cp-data->buf:-1;
    } while (ps>=0);
  }
  return *this;
}

spstrw& spstrw::replace(const spstrw &s1,const spstrw &s2) {
#ifndef NO_SP_ASSERTS
  assert(!Lock&&!s1.Lock&&!s2.Lock);
#endif
  return replace(s1(),s2());
}

spstrw& spstrw::replace(const wchar_t *src,const wchar_t *dst) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  l ii=pos(src),len=wcslen(src), dlen=wcslen(dst);
  while (ii>=0) {
    del(ii,len);
    insert(dst,ii);
    ii=pos(src,ii+dlen);
  }
  return *this;
}

spstrw& spstrw::del(l pos,d len) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&pos>=0&&pos<data->chars) {
    _unique();
    if ((d)pos+len>=(d)data->chars) data->chars=pos; else {
      memmove(data->buf+pos,data->buf+pos+len,data->chars-len-pos<<1);
      data->chars-=len;
    }
    *(data->buf+data->chars)=0;
    data->hash =FFFF;
  }
  return *this;
}

spstrw spstrw::deltostr(l pos,d len) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return spstrw();
  spstrw rc(*this,pos,len);
  del(pos,len);
  return rc;
}

spstrw& spstrw::unquote(wchar_t quote) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (data&&data->chars>1)
    if (data->buf[0]==quote&&data->buf[data->chars-1]==quote) {
      _unique();
      memmove(data->buf,data->buf+1,data->chars-2<<1);
      data->chars-=2;
      *(data->buf+data->chars)=0;
      if (data->hash!=FFFF) data->hash-=d(quote)*2;
    }
  return *this;
}

spstrw& spstrw::quote(wchar_t quote) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data||data->chars<2||data->buf[0]!=quote||data->buf[data->chars-1]!=quote) {
    _unique();
    unique((data?data->chars:0)+2);
    if (data->chars)
      memmove(data->buf+1,data->buf,data->chars<<1);
    data->chars+=2;
    data->buf[0]=quote;
    data->buf[data->chars-1]=quote;
    *(data->buf+data->chars)=0;
    if (data->hash!=FFFF) data->hash+=d(quote)*2;
  }
  return *this;
}

spstrw spstrw::operator() (l __pos,l __len) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  return data&&__len>0&&__pos>=0&&__pos<data->chars?spstrw(*this,__pos,__len):spstrw();
}

l spstrw::Int(l __pos) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  int res=0;
  if (data&&data->chars-__pos>0) {
    char sb[32];
//    *(data->buf+data->chars)=0;
    wtoai(sb,data->buf+__pos);
    res=strtol(sb,NULL,0);
  }
  return res;
}

d spstrw::Dword(l __pos) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  d res=0;
  if (data&&data->chars-__pos>0) {
    char sb[32];
//    *(data->buf+data->chars)=0;
    wtoai(sb,data->buf+__pos);
    res=strtoul(sb,NULL,0);
  }
  return res;
}

#if defined(__LONGLONG_SUPPORT) && defined(STRTOUL64)
s64 spstrw::Int64(l __pos) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  s64 res=0;
  if (data&&data->chars-__pos>0) {
    char sb[32];
    wtoai(sb,data->buf+__pos);
    res=STRTOL64(sb,NULL,0);
  }
  return res;
}

u64 spstrw::Dword64(l __pos) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  u64 res=0;
  if (data&&data->chars-__pos>0) {
    char sb[32];
    wtoai(sb,data->buf+__pos);
    res=STRTOUL64(sb,NULL,0);
  }
  return res;
}
#endif // __LONGLONG_SUPPORT

wchar_t *spstrw::LockPtr(l Size) {
  if (!Lock) {
    unique(Size);
    Lock=1;
  } else
  if (Size>data->maxnow) return NULL; else Lock++;
  return data->buf;
}

void spstrw::UnlockPtr() {
  if (Lock>0)
    if (--Lock==0) {
      data->chars=wcslen(data->buf);
      data->hash =FFFF;
    }
}

Bool spstrw::UnlockPtr(l Len) {
  if (Lock<=0||Len<0||Len>=data->maxnow) return 0;
  if (--Lock==0) {
    data->chars=Len;
    data->hash =FFFF;
    data->buf[Len]=0;
  }
  return true;
}

spstrw spstrw::left(l count) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (count&&data) return spstrw(*this,0,count);
    else return spstrw();
}

spstrw spstrw::right(l count) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return spstrw();
  l from=data->chars-count;
  from=range(from,0,data->chars);
  return spstrw(*this,from,data->chars-from);
}

d spstrw::calchash() const {
  if (!data) return 0;
  register d value=0;
  register wchar_t *cc=data->buf;
  while (*cc) value+=d(*cc++);
  return data->hash=value;
}

d spstrw::countchar(wchar_t chr) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return 0;
  if (!data->chars) return 0;
  d cnt=0;
  wchar_t *ch=wcschr(data->buf,chr);
  while (ch++) {
    cnt++;
    ch=wcschr(ch,chr);
  }
  return cnt;
}

spstrw& spstrw::replacechar(wchar_t from,wchar_t to) {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) return *this;
  if (!data->chars) return *this;
  if (from==to) return *this;
  _unique();
  wchar_t *ch=wcschr(data->buf,from);
  while (ch) {
    *ch++=to;
    if (*ch!=from) ch=wcschr(ch,from);
  }
  return *this;
}

#define IsUtfSurr(c)  ((c)>=0xD800 && (c)<=0xDFFF)
#define IsUtfSurrH(c) ((c)>=0xD800 && (c)<=0xDBFF)
#define IsUtfSurrL(c) ((c)>=0xDC00 && (c)<=0xDFFF)

spstr spstrw::getutf8(d pos,l len) const {
#ifndef NO_SP_ASSERTS
  assert(!Lock);
#endif
  if (!data) len=0; else {
    if (len<0) len=data->chars;
    if (pos>=data->chars) len=0; else
      if (pos+len>data->chars) len=data->chars-pos;
  }
  spstr out;
  if (data&&len) {
    d needlen=0,ii;
    // pre-calc string size
    for (ii=0; ii<len; ii++) {
      wchar_t ch = data->buf[pos+ii];
      if (ch<0x80) needlen++; else
        if(ch<0x800) needlen+=2; else
          if (IsUtfSurr(ch)) needlen+=2; else needlen+=3; // SurrH+SurrL = 4
    }
    char *cps=out.LockPtr(needlen+10), *cpp=cps;
    
    for (ii=0; ii<len; ii++) {
      d ch = data->buf[pos+ii];
      if (IsUtfSurrL(ch)) continue; // ignore broken pairs
      if (IsUtfSurrH(ch)) {
        if (++ii==data->chars) break;
        d chl = data->buf[pos+ii];
        if (!IsUtfSurrL(chl)) ch = chl; else
          ch = ch<<16|chl;
      }
      // ignore errors and assume that buffer is always enough
      d len = chPutUtf8(ch,cpp,6);
      if (len) cpp+=len;
    }
    out.UnlockPtr(cpp-cps);
  }
  return out;
}

// common private call for 2 functions below
Bool spstrw::setu8c(const char *src,l len) {
  clear();
  allocstr(len+10);

  wchar_t *lpp=data->buf;
  const char *spp=src;
  Bool res=True;
  // pair use 4 bytes, so we always safe here
  while (spp-src<len&&*spp) {
    d chv=chGetUtf8(&spp);
    if (chv==FFFF) { spp++; res=False; } else {
      if (chv>=0x10000) *lpp++=(wchar_t)(chv>>16);
      *lpp++=(wchar_t)chv;
    }
  }
  *lpp=0;
  data->chars=lpp-data->buf;
  data->hash =FFFF;
  // result just informs about utf-8 errors ...
  return res;
}

Bool spstrw::setutf8(const char *src,l len) {
  d slen=strlen(src);
  if (len<0) len=slen;
  return setu8c(src,len);
}

Bool spstrw::setutf8(const spstr &src,d pos,l len) {
  d slen=src.length();
  if (pos>=slen) return True;
  if (len<0) len=slen;
  if (pos+len>slen) len=slen-pos;
  return setu8c(src()+pos,len);
}
