//
// QSINIT "start" module
// secondary LE/LX code + a lot of support code
//
#define LOG_INTERNAL
#define STORAGE_INTERNAL
#include "dpmi.h"              // must be before qsutil.h
#include "qsbase.h"
#include "qslxfmt.h"
#include "malloc.h"
#include "qsconst.h"
#include "filetab.h"
#include "qsint.h"
#include "sysio.h"
#include "vioext.h"
#include "start_ord.h"
#include "syslocal.h"
#include "zip/unzip.h"
#include "qcl/bitmaps.h"
#include "qcl/qslist.h"
#include "qcl/cplib.h"
#include "classes.hpp"
#include "efnlist.h"
#include "qecall.h"
#include "rmeinfo.h"
#include "ioint13.h"
#include "parttab.h"

static u8t              *M3Buffer = 0;
u8t                     mod_delay = 1;   ///< exe delayed unpack
u8t                      lvmmount = 0;
static TStrings        *DelayList = 0;
static io_handle       mfs_handle = 0;
u32t                    mh_qsinit = 0;   ///< QSINIT module handle
extern mod_addfunc       exptable;       // just a forward

extern "C" {
extern u32t          M3BufferSize;
extern u16t               IODelay;
extern char            aboutstr[];
extern char      aboutstr_local[];
extern
struct LdrRstData       LdrRstTab;
extern boot_data        boot_info;
extern struct Disk_BPB    BootBPB;

#pragma aux aboutstr       "_*";
#pragma aux IODelay        "_*";
#pragma aux LdrRstTab      "_*";
#pragma aux boot_info      "_*";
#pragma aux BootBPB        "_*";

void         check_version(void);
void         check_codepage(void);
void         setup_loader_mt(void);
void         setup_loader(void);
int          unpack_ldi  (void);
/// call trace on parocess start
void         trace_start_hook(module *mh);
/// call trace on module free
void         trace_unload_hook(module *mh);
/// call trace just after process start
void         trace_pid_start(void);

u64t _std    gettimediff (u64t time33, u32t tics);
// decompression routines
u32t _std    DecompressM2(u32t DataLen, u8t* Src, u8t *Dst);
u32t _std    DecompressM3(u32t DataLen, u8t* Src, u8t *Dst, u8t *Buffer);
void _std    log_memtable(printf_function pfn, void*, void*, u8t*, void*,
                          u32t, process_context **, u32t*, u32t);
// cache routines
void _std    cache_ioctl (u8t vol, u32t action);
u32t _std    cache_read  (u32t disk, u64t pos, u32t ssize, void *buf);
u32t _std    cache_write (u32t disk, u64t pos, u32t ssize, void *buf);
u16t _std    io_mfs_open (const char *name, u32t *filesize);
u32t __cdecl io_mfs_read (u32t offset, void *buf, u32t readsize);
void __cdecl io_mfs_close(void);
}

#define checkexps(x) if (expsmax<=(x)) { \
   exp = mh->exps = (mod_export*)realloc(mh->exps,(expsmax*=2)*sizeof(mod_export)); \
   memset(exp+(expsmax>>1), 0, (expsmax>>1)*sizeof(mod_export)); \
}

/* build export table (full version, short in qsinit supports only 32bit and no
   forwards) */
int _std mod_buildexps(module *mh, lx_exe_t *eh) {
   lx_exp_b    *ee = (lx_exp_b*)((u8t*)eh+eh->e32_enttab);
   u32t        idx = 1, cnt=0, expsmax=32, thcnt = 0;
   mod_export *exp = mh->exps = (mod_export*)calloc(expsmax, sizeof(mod_export));

   while (ee->b32_cnt) {
      u32t e32=0, bsize=0;
      switch (ee->b32_type&~TYPEINFO) {
         case EMPTY   : idx+=ee->b32_cnt; bsize=2; break;
         case ENTRY32 :
            e32 = 2;
         case ENTRY16 :
            if (ee->b32_obj==0||ee->b32_obj>eh->e32_objcnt) // invalid object?
               return E_MOD_BADEXPORT; else
            {
               u8t *entry = (u8t*)ee + (bsize=sizeof(lx_exp_b));
               while (ee->b32_cnt--) {
                  mod_object *obj = mh->obj + (ee->b32_obj-1);
                  exp[cnt].ordinal = idx++;
                  if (!e32) {
                     exp[cnt].is16 = 1;
                     exp[cnt].sel  = obj->sel;
                     exp[cnt].address = *(u16t*)++entry;
                  } else {
                     exp[cnt].address = *(u32t*)++entry + (u32t)obj->address;
                     // save object - for thunking code below
                     if (obj->flags & OBJEXEC) {
                        exp[cnt].direct = ee->b32_obj;
                        thcnt++;
                     }
                  }
                  entry+=2+e32; bsize+=3+e32;
                  checkexps(++cnt);
               }
            }
            break;
         case GATE16  : // ring 2 does not supported
            return E_MOD_UNSUPPORTED;
         case ENTRYFWD: // forward entry
            {
               u8t *entry = (u8t*)ee + (bsize=sizeof(lx_exp_b));
               while (ee->b32_cnt--) {
                  // forwarding to the name table is not supported
                  if ((*entry++&1)==0) return E_MOD_UNSUPPORTED;
                  exp[cnt].ordinal = idx++;
                  exp[cnt].forward = *(u16t*)entry;
                  // invalid module ordinal?
                  if (!exp[cnt].forward||exp[cnt].forward>eh->e32_impmodcnt)
                     return E_MOD_BADEXPORT;
                  entry+=2;
                  exp[cnt].address = *(u32t*)entry;
                  entry+=4; bsize+=7;
                  checkexps(++cnt);
               }
            }
            break;
         default:
            return E_MOD_BADEXPORT;
      }
      ee=(lx_exp_b*)((u8t*)ee+bsize);
   }
   mh->exports = cnt;
   /* creating thunks.
      It will be aligned to 16 bytes in memory because memmgr return such
      blocks. Thunks created only for entries, which points to 32bit CODE
      objects.
      Trace code will check it in the same way:
         if (entry - module->thunks <= 16*index) ... // this is thunk.
   */
   if (thcnt) {
      u32t ii;
      u8t *thunks = mh->thunks = (u8t*)malloc(EXPORT_THUNK*thcnt);
      memset(thunks, 0xCC, EXPORT_THUNK*thcnt);

      for (ii=0; ii<cnt; ii++) {
         u32t addr = exp[ii].address;
         if (exp[ii].direct) {
            u8t* thc        = thunks;
            exp[ii].address = (u32t)thc;
            *thc++          = 0xE9;
            *(u32t*)thc     = addr - ((u32t)thc+4);
            thunks         += EXPORT_THUNK;
         }
         exp[ii].direct = addr;
      }
   }
   /* parse name tables.
      (tables validated in mdt.c) */
   char *src = (char*)mh->rtab;
   idx = 0;
   while (*src) {
      u8t  len = *src++&0x7F;
      u16t ord = *(u16t*)(src+len);
      // log_it(2, "ord %u len %u\n", ord, len);
      if (ord)
         for (u32t ii=0; ii<cnt; ii++)
            if (exp[ii].ordinal==ord) {
               // make zero-term string (overwrite ord)
               src[len] = 0;
               exp[ii].name = src;
               break;
            }
      src+=len+2;
      // switch to the non-res table
      if (!*src && !idx && mh->nrtab) { idx=1; src=(char*)mh->nrtab; }
   }
   return 0;
}

u32t _std mod_searchload(const char *name, u32t flags, qserr *error) {
   char path[QS_MAXPATH+1],
       mname[E32MODNAME+5], *epos;
   strncpy(mname,name,E32MODNAME);
   mname[E32MODNAME] = 0;
   epos = strrchr(mname,'.');
   if (!epos||strlen(epos)>4) {
      epos = mname + strlen(mname);
      strcpy(epos,".DLL");
   } else
      epos = 0;
   strupr(mname);

   if (error) *error = 0;
   *path=0;
   _searchenv(mname,"LIBPATH",path);
   if (!*path) _searchenv(mname,"PATH",path);
   if (!*path) _searchenv(name,"LIBPATH",path);
   if (!*path) _searchenv(name,"PATH",path);
   if (!*path&&epos) {
      strcpy(epos,".EXE");
      _searchenv(mname,"PATH",path);
   }
   if (!*path) return 0;
   return mod_load(path, flags, error, 0);
}

/* unpack code works as IBM one and it search for zero at the end of iterated
   data, but who guarantee its presence? it really can be missed, at least
   in ITERDATA pages.
   so we add it ;) we can save 4 bytes after source data because both mdt.c and
   bootos2.exe code allocate buffer with additional dword (even if it is an
   impossible case to reach the end of memory block)

   note: this comment applied to mod_unpack2 code too. */

u32t _std mod_unpack1(u32t datalen, u8t* src, u8t *dst) {
   struct LX_Iter *ir = (struct LX_Iter*)src;
   u8t *page = dst;
   u32t save;
   if (datalen>=PAGESIZE) return 0;
   save = *(u32t*)(src+datalen);
   *(u32t*)(src+datalen) = 0;

   while (ir->LX_nIter) {
      switch (ir->LX_nBytes) {
         case 1:
            memset(dst, ir->LX_Iterdata, ir->LX_nIter);
            dst+=ir->LX_nIter;
            break;
         case 2:
            memsetw((u16t*)dst, *(u16t*)&ir->LX_Iterdata, ir->LX_nIter);
            dst+=ir->LX_nIter<<1;
            break;
         case 4:
            memsetd((u32t*)dst, *(u32t*)&ir->LX_Iterdata, ir->LX_nIter);
            dst+=ir->LX_nIter<<2;
            break;
         default: {
            u32t cnt;
            for (cnt=0; cnt<ir->LX_nIter; cnt++) {
               memcpy(dst,&ir->LX_Iterdata,ir->LX_nBytes);
               dst+=ir->LX_nBytes;
            }
            break;
         }
      }
      ir = (struct LX_Iter*)((u8t*)ir+4+ir->LX_nBytes);
   }
   *(u32t*)(src+datalen) = save;
   return dst - page;
}

u32t _std mod_unpack2(u32t datalen, u8t* src, u8t *dst) {
   u32t save, rc;
   if (datalen>=PAGESIZE) return 0;
   save = *(u32t*)(src+datalen);
   *(u32t*)(src+datalen) = 0;
   rc = DecompressM2(datalen, src, dst);
   *(u32t*)(src+datalen) = save;
   return rc;
}

u32t _std mod_unpack3(u32t datalen, u8t* src, u8t *dst) {
   if (!M3Buffer)
      M3Buffer = (u8t*)hlp_memallocsig(M3BufferSize, "M3", QSMA_READONLY);
   return DecompressM3(datalen,src,dst,M3Buffer);
}

u32t _std mod_getpid(void) { return mod_context()->pid; }

u32t _std mod_appname(char *name, u32t parent) {
   process_context *pq = mod_context();
   if (!name||!pq||parent>1) return 0; else {
      module *src = parent?pq->parent:pq->self;
      if (!src) return 0;
      strncpy(name,src->name,E32MODNAME);
      name[E32MODNAME] = 0;
      strupr(name);
      return pq->pid;
   }
}

char *_std mod_getname(u32t mh, char *buffer) {
   module *md = (module*)mh;
   char  *res = 0;
   // function is r/o, so it`s safe to use this mutex here
   if (in_mtmode) mt_muxcapture(mod_secondary->ldr_mutex);
   if (md->sign==MOD_SIGN) {
      if (buffer) {
         strncpy(buffer, md->name, E32MODNAME);
         buffer[E32MODNAME] = 0;
         strupr(buffer);
         res = buffer;
      } else
      if (md->name_buf) res = md->name_buf; else {
         res = strdup(md->name);
         mem_modblockex(res, mh);
         strupr(res);
         md->name_buf = res;
      }
   }
   if (in_mtmode) mt_muxrelease(mod_secondary->ldr_mutex);
   return res;
}

static void release_ldi(void) {
   mt_swlock();
   void *mem = sto_data(STOKEY_ZIPDATA);
   sto_del(STOKEY_ZIPDATA);

   mod_secondary->ldi_gone = 1;
   mt_swunlock();

   if (mem) hlp_memfree(mem);
}

/// called just after MT mode start in the system thread
void unzip_delaylist(void) {
   u32t md_count = 0;
   if (DelayList) {
      md_count = DelayList->Count();
      if (md_count) {
         spstr cmd;
         cmd.sprintf("/o /q /key %s b:\\ ", STOKEY_ZIPDATA);
         cmd += DelayList->MergeCmdLine();
         cmd_shellcall(shl_unzip, cmd(), 0);
      }
      delete DelayList;
      DelayList = 0;
   }
   release_ldi();
   if (md_count)
      log_it(2, "Delayed unpack (%d modules) done\n", md_count);
}

/* unzip one of delayed EXE/DLL from the saved QSINIT.LDI data.
   After all files unpacked - LDI data will be released */
int _std unzip_ldi(void *mem, u32t size, const char *path) {
   char drv = toupper(path[0]);
   // should not be called in MT mode
   if (in_mtmode) return 0;
   // we should receive a full path here, but it can be both 1: & B:
   if (drv!='B' && drv!='1' || path[1]!=':') return 0;

   spstr callstr, cpath(path+2);
   cpath.replacechar('/','\\');
   while (cpath[0]=='\\') cpath.del(0);

   callstr.sprintf("/o /q /key %s b:\\ %s", STOKEY_ZIPDATA);
   callstr+=cpath;
   u32t  dm_cnt = DelayList?DelayList->Count():0;
   if (DelayList) {
      l idx = DelayList->IndexOfICase(cpath());
      if (idx>=0) {
         DelayList->Delete(idx);
         if (DelayList->Count()==0) delete DelayList;
      } else
         dm_cnt = 0;
   }
   if (!dm_cnt) return 0;

   log_it(2, "unpack %s, %d bytes (%d)\n", cpath(), size, dm_cnt);
   if (cmd_shellcall(shl_unzip, callstr(), 0)) return 0;

   u32t  bsize = 0;
   void *fdata = freadfull(path, &bsize);
   if (bsize==size && fdata) {
      // all delayed modules unpacked! then free ZIP data (650k now!)
      if (dm_cnt<=1) release_ldi();
      memcpy(mem, fdata, size);
   } else bsize=0;
   if (fdata) hlp_memfree(fdata);

   return bsize?1:0;
}

int unpack_ldi(void) {
   void     *zip = sto_data(STOKEY_ZIPDATA);
   u32t    zsize = sto_size(STOKEY_ZIPDATA);
   ZIP        zz;
   char  *zfpath;
   u32t   zfsize = 0;
   int    errors = 0, errprev;

   if (zip_open(&zz, zip, zsize)) return 0;

   while (zip_nextfile(&zz, &zfpath, &zfsize)==0) {
      int      skip = 0;
      void *filemem = 0;
      char     path[QS_MAXPATH+1];
      if (!zfsize) continue;
      // skip START module - because we are here already ;)
      if (stricmp(MODNAME_START,zfpath)==0) continue;
      // full path (x:\name)
      snprintf(path, QS_MAXPATH, "%c:\\%s", 'A'+DISK_LDR, zfpath);
      errprev = errors;

      if (mod_delay) {
         char *ep = strchr(zfpath,'.');
         // delayed unpack (for large modules only)
         if (ep && (stricmp(ep,".EXE")==0 || stricmp(ep,".DLL")==0)) {
            if (!DelayList) DelayList = new TStrings;
            DelayList->AddObject(spstr(zfpath).replacechar('/','\\'), zz.dostime);
            skip = 1;
         }
      }
      if (!skip) {
         zip_readfile(&zz, &filemem);
         if (!filemem) errors++;
      }
      if (skip || filemem) {
         qserr res = 0;
         while (1) {
            io_handle fh;
            res = io_open(path, IOFM_CREATE_ALWAYS|IOFM_WRITE, &fh, 0);
            // log_printf("io_open(%s) err %08X!\n", path, res);
            if (!res) {
               if (skip) {
                  res = io_setsize(fh,zfsize);
                  if (!res) {
                     u32t data = UNPDELAY_SIGN;
                     if (io_write(fh, &data, 4)!=4) res = E_DSK_ERRWRITE;
                  }
               } else
               if (io_write(fh, filemem, zfsize)!=zfsize) res = E_DSK_ERRWRITE;
               io_close(fh);
               break;
            } else {
               char dir[QS_MAXPATH+1];
               _splitfname(path, dir, 0);
               res = io_mkdir(dir);
               if (res) break; else continue;
            }
         }
         if (filemem) hlp_memfree(filemem);
         if (res) errors++;
         /* ugly way to detect verbose build.
            (these messages scroll up actual info in it) */
         u32t lines;
         vio_getmode(0, &lines);
         if (lines<50)
            log_printf("%s %s, %d bytes\n", skip?"skip ":"unzip", path, zfsize);
         // set file time on success
         if (errprev!=errors) log_printf("%s: error (%08X)!\n", path, res); else {
            io_handle_info fi;
            io_timetoio(&fi.ctime, dostimetotime(zz.dostime));
            fi.atime = fi.ctime;
            fi.wtime = fi.ctime;
            io_setinfo(path, &fi, IOSI_CTIME|IOSI_ATIME|IOSI_WTIME);
         }
      }
   }
   zip_close(&zz);
   // free memory / set flag
   if (!mod_delay) release_ldi();
#if 0
   str_list *lst = str_getlist(DelayList->Str);
   log_printlist("list:", lst);
   free(lst);
#endif
   if (errors && !mod_delay && !hlp_insafemode()) {
      log_printf("%u errors in QSINIT.LDI!\n", errors);
      return 0;
   }
   return 1;
}

/** micro-FSD file open.
    All io_mfs_* functions called under __microfsd_mux__ and
    have protection from second open in caller. */
u16t _std io_mfs_open(const char *name, u32t *filesize) {
   if (mfs_handle) return 1; else {
      char     *fnn;
      u32t   action;
      u64t      fsz;
      qserr      rc;
      /* simulate UEFI mfs open: \EFI\BOOT by default and direct path
         by / or \\ */
      if (hlp_hosttype()==QSHT_EFI) {
         if (name[0]=='/' || name[0]=='\\')
            fnn = sprintf_dyn("A:%s", name);
         else {
            char *eroot = (char*)sto_data(STOKEY_EFIROOT);
            fnn = sprintf_dyn("A:%s\\%s", eroot ? eroot : "\\EFI\\BOOT", name);
         }
      } else
         fnn = sprintf_dyn("A:\\%s", name);

      rc = io_open(fnn, IOFM_READ|IOFM_OPEN_EXISTING|IOFM_SHARE_READ|
         IOFM_SHARE_REN|IOFM_SHARE_DEL, &mfs_handle, &action);
      if (rc) log_printf("mfs open(%s) = %X\n", fnn, rc);
      free(fnn);
      if (!rc) rc = io_size(mfs_handle, &fsz);
      if (!rc && fsz>_2GB+_1GB) rc = E_SYS_TOOLARGE;
      if (!rc) rc = io_setstate(mfs_handle, IOFS_DETACHED, 1);
      if (rc) {
         if (mfs_handle) { io_close(mfs_handle); mfs_handle = 0; }
         return rc;
      }
      *filesize = fsz;
      return 0;
   }
}

u32t __cdecl io_mfs_read(u32t offset, void *buf, u32t readsize) {
   if (!mfs_handle || !readsize || !buf) return 0;
   if (io_seek(mfs_handle, offset, IO_SEEK_SET)==FFFF64) return 0;
   return io_read(mfs_handle, buf, readsize);
}

void __cdecl io_mfs_close(void) {
   if (mfs_handle) { io_close(mfs_handle); mfs_handle = 0; }
}

static u32t _std bitfind(void *data, u32t size, int on, u32t len, u32t hint) {
   bit_map bmp = NEW(bit_map);
   if (!bmp) return FFFF;
   bmp->init(data,size);
   u32t rc = bmp->find(on, len, hint);
   DELETE(bmp);
   return rc;
}

/** module loaded callback.
    Called before returning ok to user. But module is not guaranteed to
    be in loaded list at this time, it can be the one of loading imports
    for someone.
    ldr_mutex is captured at this time */
void _std mod_loaded(module *mh) {
   trace_start_hook(mh);
}

/** module free callback.
    ldr_mutex is captured at this time */
void _std mod_freeexps(module *mh) {
   mh->exports = 0;
   // free exports, not used below
   if (mh->exps) { free(mh->exps); mh->exps=0; }
   /* unchain ordinals first, because it rebuild thunks internally,
      then free thunks ;) */
   mod_apiunchain((u32t)mh,0,0,0);
   if (mh->thunks) { free(mh->thunks); mh->thunks=0; }
   // free trace info
   trace_unload_hook(mh);
   // free other possible function thunks
   mod_freethunk((u32t)mh,0);
   // free heap blocks, belonging to this module
   mem_freepool(QSMEMOWNER_COLIB, (u32t)mh);
}

/** process start callback.
    Function is not covered by any locks!
    This is the first call in the main thread of a new process - before main()
    itself.
    Called from mod_exec() in non-MT mode and by MTLIB app start code in MT. */
void _std mod_startcb(void) {
   // inherit/create file handle
   io_inherit();
   // clib stdio
   init_stdio(mod_context());
   // clib signals
   init_signals();
   // inform trace (maybe it should be 1st, because all of above it untraced!)
   trace_pid_start();
}

/** process exit callback.
    note, that this callback is in the CALLER context!
    To catch something in CALLEE context use mt_pexitcb() for tid 1.
    Function is not covered by MT lock, nor by ldr_mutex! */
s32t _std mod_exitcb(process_context *pq, s32t rc) {
   int errtype = 0;
   // !!!
   fcloseall_as(pq->pid);
   io_close_as(pq->pid, IOHT_FILE|IOHT_DIR|IOHT_MUTEX|IOHT_QUEUE|IOHT_EVENT|
                        IOHT_STREAM|IOHT_THREAD);
   // free PUSHD stack
   if (pq->rtbuf[RTBUF_PUSHDST]) {
      pushd_free((void*)pq->rtbuf[RTBUF_PUSHDST]);
      pq->rtbuf[RTBUF_PUSHDST] = 0;
   }
   // free ansi specific data
   if (pq->rtbuf[RTBUF_ANSIBUF]) {
      free((void*)pq->rtbuf[RTBUF_ANSIBUF]);
      pq->rtbuf[RTBUF_ANSIBUF] = 0;
   }

   if (!mem_checkmgr()) errtype = 1; else
   if (!exi_checkstate()) errtype = 2;

   if (errtype) {
      char   prnbuf[256];
      int  ses_exit = in_mtmode && sys_con && se_sesno()==sys_con;
      // can`t use printf during common exit here, if was closed above ;)
      snprintf(prnbuf, 256, "\n Application \"%s\"\n made unrecoverable damage in"
        " the system %s structures...\n\n This will produce trap or deadlock in nearest time.\n"
           " Reboot or %s to continue...\n", pq->self->name, errtype==1?"heap manager":
              "shared class", ses_exit?"switch back to your session":"press any key");
      panic_no_memmgr(prnbuf);
   }
   log_printf("memory check done\n");
   mod_apiunchain((u32t)pq->self,0,0,0);
   mod_fnunchain((u32t)pq->self,0,0,0);
   return rc;
}

void check_version(void) {
   int len = strlen(aboutstr_local);
   // at every place on Earth we can find a hero, who gets such a message ;)
   if (strncmp(aboutstr, aboutstr_local, len)) {
      char *about = strdup(aboutstr), *msg;
      about[len] = 0;
      msg = sprintf_dyn("Boot module is \"%s\",^but LDI version is \"%s\"."
                         "^^Continue at your own risk.", about, aboutstr_local);
      vio_msgbox("QSINIT.LDI version mismatch!", msg, MSG_OK|MSG_RED|MSG_WIDE, 0);
      free(about);
      free(msg);
   }
}

// set codepage, inherited from the loader, who RESTARTs us
void check_codepage(void) {
   u32t cp = sto_dword(STOKEY_RSTCPNUM);
   sto_del(STOKEY_RSTCPNUM);

   if (cp) {
      qs_cpconvert cplib = NEW(qs_cpconvert);
      if (cplib) {
         cplib->setsyscp(cp);
         DELETE(cplib);
      }
   }
}

void _std exit_with_popup(const char *msg, int errorcode) {
   char buf[70], appname[128];
   /* try to avoid malloc in _this_ thread context because of unknown source
      of the error */
   u32t pid = mod_appname(appname, 0);
   snprintf(buf, sizeof(buf), "Critical failure in PID %u (%s)", pid, appname);
   // called in the separate thread in MT mode
   vio_msgbox(buf, msg, MSG_OK|MSG_RED|MSG_WIDE|MSG_POPUP, 0);

   _exit(errorcode);
}

u32t _std mod_getmodpid(u32t module, u32t *parent) {
   if (!in_mtmode) {
      process_context *pq = mod_context();
      while (pq) {
         if ((u32t)pq->self==module) {
            if (parent) *parent = pq->pctx?pq->pctx->pid:0;
            return pq->pid;
         }
         pq = pq->pctx;
      }
   } else
      return get_mtlib()->getmodpid(module,parent);
   return 0;
}

u32t _std mod_checkpid(u32t pid) {
   if (pid)
      if (!in_mtmode) {
         process_context *pq = mod_context();
         while (pq) {
            if (pq->pid==pid) return 1;
            pq = pq->pctx;
         }
      } else
         return get_mtlib()->checkpidtid(pid,0,0)==0 ? 1 : 0;
   return 0;
}

/** return process context for the specified process id.
    Function exported, but unpublished in headers because it dangerous.
    It TURNS ON MT lock if MT mode active and returned value is non-zero */
process_context* _std mod_pidctx(u32t pid) {
   if (pid)
      if (!in_mtmode) {
         process_context *pq = mod_context();
         while (pq) {
            if (pq->pid==pid) return pq;
            pq = pq->pctx;
         }
      } else {
         // this call turns MT lock on in any case
  	     mt_prcdata      *pd = mtmux->getpd(pid);
  	     process_context *pq = pd?pd->piContext:0;
  	     // check pq<->pd match
  	     if ((mt_prcdata*)pq->rtbuf[RTBUF_PROCDAT]!=pd)
  	        sys_exfunc4(xcpt_prcerr, pq->self->name, pq->pid);

  	     if (!pq) mt_swunlock();
  	     return pq;
      }
   return 0;
}

/** query list of running processes.
    Function replaced by MTLIB on its load.
    @return pid list with zero in the last entry - in the process owned heap
            block. */
extern "C" u32t* _std START_EXPORT(mod_pidlist)(void) {
   u32t *rc = 0;
   if (!in_mtmode) {
      u32t cnt = 0, pass, ii = 0;

      for (pass=0; pass<2; pass++) {
         process_context *pq = mod_context();
         while (pq) {
            if (pass) rc[ii++] = pq->pid; else cnt++;
            pq = pq->pctx;
         }
         if (!pass) rc = (u32t*)malloc_local(sizeof(u32t)*(cnt+1)); else {
            // sort in in ascending order
            bswap(rc, sizeof(u32t), cnt);
            rc[cnt] = 0;
         }
      }
   }
   /* function should be catched and replaced by MTLIB, so just return 0 to
      hang in user code in a case of malfunction. */
   return rc;
}

/** get shell history list.
    read internal shell (cmd.exe) user commands history list.
    @return string list in the process owned heap block or 0 if list is empty. */
str_list* _std cmd_historyread(void) {
   // this handle in shared over system!
   ptr_list history = (ptr_list)sto_data(STOKEY_CMDSTORY);
   if (!history) return 0;
   // use embedded mutex to lock entire call, so array() can be used here
   exi_lock(history);
   u32t     cnt = history->count();
   str_list *sl = cnt ? sl = str_fromptr((char**)history->array(), cnt) : 0;
   exi_unlock(history);
   return sl;
}

/// clear shell history list.
qserr _std cmd_historywipe(void) {
   // this handle in shared over system!
   ptr_list history = (ptr_list)sto_data(STOKEY_CMDSTORY);
   if (!history) return E_SYS_NOTFOUND;
   // use embedded mutex to lock entire call
   exi_lock(history);
   u32t     cnt = history->count(), ii;

   for (ii=0; ii<cnt; ii++) {
      char *str = (char*)history->value(ii);
      // free it with care!
      if (str && mem_getobjinfo(str,0,0)) free(str);
   }
   history->clear();
   exi_unlock(history);
   return 0;
}

/// get shell command history (process owned block)
static char* _std shl_history(void) {
   str_list *sl = cmd_historyread();
   char     *rc = str_gettostr(sl,"\n");
   free(sl);
   return rc;
}

/// fill a data transition block for the next loader
static void add_trblock(transition_data *td, void *data, u32t len, u32t type) {
   td->sign   = TRND_SIGN;
   td->length = len;
   td->type   = type;
   memcpy(td+1, data, len);
   td->crc32  = crc32(0, (u8t*)data, len);
   // log_printf("rst %u %X %u!\n", type, td->crc32, len);
}

/* just a temporary space to place moveton`s PXE micro filetable
   during restart (it`s abandoned after file open) */
#define FILETAB_SEG 0x3500

/// exit and restart another os2ldr
qserr _std exit_restart(const char *loader, void *env, char *ini, u32t flags) {
   if (hlp_hosttype()==QSHT_EFI) return E_SYS_EFIHOST;

   disk_volume_data di;
   u32t        ldrsize = 0,
                envlen = env ? env_length((const char*)env) : 0, hlen, ilen;
   qserr           err;
   void       *ldrdata = 0;
   char         *hdata;
   void        *trdata;
   u8t          repvol,
                bflags = LdrRstTab.LRstBootFlags,
                 bdisk = LdrRstTab.LRstBootDisk;
   int           VolFs = -1;

   // BPB replacement ...
   repvol = (u8t)(flags>>QRST_BPBSHIFT);
   if (repvol==DISK_LDR) return E_DSK_NOTPHYS;
   if (repvol) {
      err = io_volinfo(repvol, &di);
      if (err) return err;
      // cannot change volume to HPFS/JFS
      if (di.FsVer==FST_OTHER && (!strcmp(di.FsName,"HPFS") ||
         !strcmp(di.FsName,"JFS"))) return E_DSK_FSMISMATCH;
      // deny ram disks ...
      if (hlp_diskmode(di.Disk,HDM_QUERY) & HDM_EMULATED)
         return E_DSK_NOTPHYS;
      // get BIOS disk number
      u32t biosdisk = hlp_diskbios(di.Disk, 1);
      if (biosdisk==FFFF) return E_DSK_NOTPHYS;

      bdisk = (u8t)biosdisk;
      VolFs = di.FsVer;
   }
   // boot drive first, then try it as a path
   if (loader[1]!=':') {
      if (repvol) {
         char *fp = sprintf_dyn("%c:\\%s", repvol+'A', loader);
         ldrdata  = freadfull(fp, &ldrsize);
         free(fp);
      } else
         ldrdata = hlp_freadfull(loader, &ldrsize, 0);
   }
   if (!ldrdata) ldrdata = freadfull(loader, &ldrsize);
   if (!ldrdata || !ldrsize) return E_SYS_NOFILE;

   u8t srcvol = repvol?repvol:DISK_BOOT;
   // setup real mode destination and registers
   struct rmcallregs_s regs;
   memset(&regs, 0, sizeof(regs));
   regs.r_cs  = LdrRstTab.LRstCS;
   regs.r_ds  = LdrRstTab.LRstBPBSeg;
   regs.r_es  = LdrRstTab.LRstRootSeg;
   regs.r_ss  = LdrRstTab.LRstSS;
   regs.r_sp  = LdrRstTab.LRstSP;
   regs.r_ebx = LdrRstTab.LRstFirstData;
   regs.r_edi = LdrRstTab.LRstRootOfs;
   regs.r_esi = LdrRstTab.LRstBPBOfs;

   err = 0;
   // need boot sector data (another volume or FAT)
   if (repvol || !boot_info.minifsd_ptr) {
      if (VolFs<0) VolFs = hlp_volinfo(srcvol,&di);

      struct Boot_Record *bs = (struct Boot_Record*)malloc_th(di.SectorSize);
      struct Common_BPB *bpb = &bs->BR_BPB;

      if (!hlp_diskread(QDSK_VOLUME|srcvol,0,1,bs)) err = E_DSK_ERRREAD; else {
         /* rewrite registers with general non-IFS boot data if volume was
            changed */
         if (repvol) {
            // we cannot switch to mini-FSD, so only drop it!
            bflags     = VolFs==FST_EXFAT ? BF_NEWBPB : 0;

            regs.r_ds  = 0;
            regs.r_esi = 0x7C00;
            if (bflags) { // exFAT case
               struct Disk_NewPB exfb;
               exfb.NPB_BytesPerSec = di.SectorSize;
               exfb.NPB_VolStart    = di.StartSector;
               exfb.NPB_VolSize     = di.TotalSectors;
               exfb.NPB_Reserved    = 0;
               // copy just to 7C00 ...
               hlp_memcpy((char*)regs.r_esi, &exfb, sizeof(struct Disk_NewPB),
                          MEMCPY_PG0);
            } else {      // any other (including NTFS)
               // valid BPB? try to update "Hidden Sectors" value ...
               u16t bps = bpb->BPB_BytePerSect;
               if (bps==512 || bps==1024 || bps==2048 || bps==4096) {
                  if (di.StartSector<=FFFF) bpb->BPB_HiddenSec = (u32t)di.StartSector;
               } else
                  log_printf("warning! not a BPB!\n");
               // boot sector at 7C00
               hlp_memcpy((char*)regs.r_esi, bs, di.SectorSize, MEMCPY_PG0);
               // directory data just after the boot sector
               regs.r_es  = regs.r_esi + di.SectorSize >> 4;
               regs.r_edi = 0;
               regs.r_esi+= FIELDOFFSET(struct Boot_Record,BR_BPB);
               // update "offset to data" value
               if (VolFs==FST_FAT12 || VolFs==FST_FAT16)
                  regs.r_ebx = di.DataSector;
            }
            // drop to DPMI stack - this is safer in this case 
            regs.r_ss  = 0;
            regs.r_sp  = 0;
            /* expected location for FAT/exFAT boot ...
               NTLDR/BOOTMGR also want to be here ;) */
            regs.r_cs  = 0x2000;
         } else
         if (BootBPB.BPB_RootEntries!=bpb->BPB_RootEntries)
            log_printf("warning! BPB mismatch!\n");
   
         u32t rootlen = bpb->BPB_RootEntries * 32;

         // read/update root directory data for FAT12/16
         if ((VolFs==FST_FAT12 || VolFs==FST_FAT16) && rootlen && !bflags) {
            u32t nsec = rootlen/di.SectorSize;
            if (rootlen % di.SectorSize) nsec++;

            void *buf = malloc_th(di.SectorSize*nsec);
            /* read boot volume FAT12/16 root dir sectors and copy them to
               the original place */
            if (hlp_diskread(QDSK_VOLUME|srcvol, bpb->BPB_ResSectors +
               bpb->BPB_FATCopies*bpb->BPB_SecPerFAT, nsec, buf)!=nsec)
                  err = E_DSK_ERRREAD;
            else
               hlp_memcpy((char*)hlp_segtoflat(regs.r_es)+regs.r_edi, buf,
                  rootlen, MEMCPY_PG0);
            free(buf);
         }
      }
      free(bs);
   }
   /* verify that it fits between wanted CS and the loader/micro-FSD segment.
      Windows BOOTMGR is large, actually, so this can be fatal for PXE boot */
   if (!err) {
      u32t freetop = (bflags&BF_MICROFSD ? boot_info.filetab.ft_museg :
                      hlp_querybios(QBIO_QS16SEG)) << 4;
      if (freetop - ((u32t)regs.r_cs<<4) < ldrsize) err = E_SYS_EFBIG;
   }
   if (!err) {
      u32t rc = 1;
      // are we still a FSD boot?
      if ((bflags&BF_MICROFSD) && boot_info.minifsd_ptr)
         rc = hlp_memcpy((void*)hlp_segtoflat(boot_info.filetab.ft_mfsdseg),
            boot_info.minifsd_ptr, boot_info.filetab.ft_mfsdlen, MEMCPY_PG0);
      // copy BPB back to initial location (if available)
      if (!repvol && !(bflags&BF_NOMFSHVOLIO)==0)
         rc = hlp_memcpy((char*)hlp_segtoflat(regs.r_ds)+regs.r_esi, &BootBPB,
            bflags&BF_NEWBPB?sizeof(struct Disk_NewPB):sizeof(BootBPB), MEMCPY_PG0);
      // one of copy ops failed
      if (!rc) err = E_SYS_SOFTFAULT;
   }
   if (err) {
      hlp_memfree(ldrdata);
      return err;
   }

   // build data for the next loader
   hdata  = flags & QRST_NOHIST ? 0 : shl_history();
   hlen   = hdata ? strlen(hdata)+1 : 0;
   ilen   = ini ? strlen(ini)+1 : 0;
   trdata = 0;
   u32t rlen = 0;
   // environment copy
   if (envlen) rlen += Round16(envlen) + sizeof(transition_data);
   // cmd.exe history
   if (hlen)   rlen += Round16(hlen) + sizeof(transition_data);
   // ini file contents
   if (ilen)   rlen += Round16(ilen) + sizeof(transition_data);
   // current codepage
   if (SysCP && !(flags&QRST_NOCP)) rlen += 16 + sizeof(transition_data);
   // flags value
   if (flags) rlen += 16 + sizeof(transition_data);
   if (rlen) {
      trdata = hlp_memalloc(rlen, QSMA_RETERR|QSMA_NOCLEAR|QSMA_MAXADDR);
      if (trdata) {
         transition_data *td = (transition_data *)trdata;
         td->sign = 0;
         td->type = 0;
         // save env strings block
         if (envlen) add_trblock(td, env, envlen, TRND_ENV);
         // save shell history block
         if (hlen) {
            if (td->type) td = next_trblock(td);
            add_trblock(td, hdata, hlen, TRND_CMDHIST);
         }
         // save current codepage
         if (SysCP && !(flags&QRST_NOCP)) {
            if (td->type) td = next_trblock(td);
            add_trblock(td, &SysCP, 2, TRND_CODEPAGE);
         }
         // save INI file
         if (ilen) {
            if (td->type) td = next_trblock(td);
            add_trblock(td, ini, ilen, TRND_CFGFILE);
         }
         if (flags) {
            if (td->type) td = next_trblock(td);
            // should be after all "flags" use above
            flags &= QRST_SENDMASK;
            add_trblock(td, &flags, sizeof(flags), TRND_FLAGS);
         }
         if (td->type) td->type |= TRND_LAST;
      }
   }
   regs.r_ecx = (u32t)trdata;
   regs.r_edx = (u32t)bflags<<8 | bdisk;
#if 0
   if (repvol)
      log_printf("restart: cs=%04X ds:si=%04X:%04X es:di=%04X:%04X dx=%04X bx=%X len=%u\n",
         regs.r_cs, regs.r_ds, regs.r_esi, regs.r_es, regs.r_edi, regs.r_edx,
            regs.r_ebx, ldrsize);
#endif
   // clear screen
   vio_clearscr();
   // disable micro-FSD "terminate" call
   if (bflags&BF_MICROFSD) exptable.mfsd_noterm = 1;
   // shutdown self
   exit_prepare(SHUT_RESTART);
   /* copy new loader to the right place - it can overwrite micro-FSD in FAT
      cases, so do this after "shutdown" */
   memcpy((void*)hlp_segtoflat(regs.r_cs), ldrdata, ldrsize);
   // no need to free - we can`t return anyway ;)
   //hlp_memfree(ldrdata);
   // still micro-FSD boot?
   if (bflags&BF_MICROFSD) {
      /* fix moveton`s PXE broken filetable - use own cached copy. Just copy it
         to 3500:0 (address is safe enough for now since nobody except OS/4
         and QSINIT can use this API and both loaders fits between 1000 and 3500) */
      if (boot_info.filetab.ft_cfiles==6 && !boot_info.filetab.ft_muOpen) {
         regs.r_es  = FILETAB_SEG;
         regs.r_edi = 0;
         memcpy((u8t*)hlp_segtoflat(FILETAB_SEG), &boot_info.filetab,
                sizeof(struct filetable_s));
      }
      // patch loader size in the original filetable
      ((struct filetable_s*)(hlp_segtoflat(regs.r_es)+regs.r_edi))->
         ft_loaderlen = ldrsize;
   }
   // and launch it
   hlp_rmcallreg(-1, &regs, RMC_EXITCALL);
   // unreachable code
   return E_SYS_SOFTFAULT;
}

static qserr exit_loadfile(const char *path, u32t &bflen, void *&bf, u32t maxsize) {
   if (!path) return E_SYS_ZEROPTR;
   // no full path
   if (path[1]!=':') bf = hlp_freadfull(path, &bflen, 0);
   if (bf) {
      if (bflen>maxsize) {
         hlp_memfree(bf);
         return E_SYS_EFBIG;
      }
   } else {
      io_handle  fh;
      u64t      fsz;
      qserr     err;

      err = io_open(path, IOFM_OPEN_EXISTING|IOFM_READ|IOFM_SHARE_READ, &fh, 0);
      if (err) return err;
      err = io_size(fh, &fsz);
      if (!err && fsz>maxsize) err = E_SYS_EFBIG;
      if (!err) {
         bflen = fsz;
         bf    = hlp_memallocsig(bflen, "raw", QSMA_RETERR|QSMA_LOCAL);
         if (!bf) err = E_SYS_NOMEM; else
            if (io_read(fh, bf, bflen) != bflen)
               if (!(err = io_lasterror(fh))) err = E_SYS_SOFTFAULT;
      }
      io_close(fh);
      if (err) {
         if (bf) hlp_memfree(bf);
         return err;
      }
   }
   return 0;
}

/// exit and start another real mode code.
qserr _std exit_rawcode(const char *rmfile, u32t segofs) {
   if (hlp_hosttype()==QSHT_EFI) return E_SYS_EFIHOST;
   struct rmcallregs_s regs;
   u32t bflen = 0;
   void   *bf = 0;
   // use default if 0
   if (!segofs) segofs = 0x7C00;
   u32t flloc = (segofs>>16<<PARASHIFT) + (segofs & 0xFFFF),
        qsloc = hlp_querybios(QBIO_QS16SEG) << PARASHIFT;
   // verify target location
   if (flloc<0x600 || flloc>=qsloc) return E_SYS_INVPARM;
   // common load
   qserr  err = exit_loadfile(rmfile, bflen, bf, _512KB);
   if (err) return err;
   // file will erase a part of QSINIT code?
   if (flloc+bflen > qsloc) {
      hlp_memfree(bf);
      return E_SYS_EFBIG;
   }
   // setup DPMI
   memset(&regs, 0, sizeof(regs));
   regs.r_cs  = (u16t)(segofs>>16);
   regs.r_ip  = (u16t)segofs;
   /* send actual disk number, this allows to start NTLDR from CD in CD/DVD
      boot ;)
        "restart /raw:2000 ntldr" */
   regs.r_edx = boot_info.boot_disk;

   log_it(2,"launching %s raw code (%u bytes) at %04X:%04X\n", rmfile, bflen,
      regs.r_cs, regs.r_ip);
   // shutdown all and go boot
   vio_clearscr();
   exit_prepare(SHUT_RMCODE);
   // place to 0:07C00
   memcpy((char*)hlp_segtoflat(0)+flloc, bf, bflen);
   // one way call!
   hlp_rmcallreg(-1, &regs, RMC_EXITCALL);
   // who knows?
   memset(&regs, 0, sizeof(regs));
   hlp_rmcallreg(0x19, &regs, RMC_EXITCALL);
   // unreachable code
   return E_SYS_SOFTFAULT;
}

qserr _std exit_efiboot(const char *module, const char *args) {
   if (hlp_hosttype()!=QSHT_EFI) return E_SYS_BIOSHOST;
   if (!module) return E_SYS_ZEROPTR;
   u32t bflen = 0;
   void   *bf = 0;
   qserr  err = 0;

   if (module[1]!=':') {
      bflen = hlp_fexist(module);
      if (bflen!=FFFF && bflen>EFBM_SIZEMASK) return E_SYS_EFBIG;
      if (!bflen) err = E_SYS_NOFILE;
   }
   if (module[1]==':' || err) {
      err = exit_loadfile(module, bflen, bf, EFBM_SIZEMASK);
      if (err) return err;
   }
   log_it(2,"loading %s efi module (%u bytes)\n", module, bflen);
   err = call64(EFN_SETEFIMODULE, 0, 3, bf ? (char*)bf : module, args,
                bflen | (bf?EFBM_MEMORY:EFBM_FILE));
   if (err) {
      if (bf) hlp_memfree(bf);
      return err;
   }
   vio_clearscr();
   exit_prepare(SHUT_EFICODE);
   exit_pm32(0);
   // unreachable code
   return E_SYS_SOFTFAULT;
}

mod_addfunc exptable = { sizeof(mod_addfunc)/sizeof(void*)-1, 0, // number of entries
   mod_buildexps, mod_searchload, mod_unpack1, mod_unpack2, mod_unpack3,
   mod_freeexps, mem_alloc, mem_realloc, mem_free, freadfull, log_pushtm,
   sto_save, sto_flush, unzip_ldi, mod_startcb, mod_exitcb, log_memtable,
   hlp_memcpy, mod_loaded, sys_exfunc4, hlp_volinfo, io_fullpath,
   getcurdirw_dyn, io_remove, mem_freepool, sys_notifyexec, sys_notifyevent,
   0,0,0, mt_muxcapture, mt_waithandle, mt_muxrelease, env_copy, 0, 0, fptostr,
   fpu_stsave, fpu_strest, 0, io_mfs_open, io_mfs_read, io_mfs_close,
   io_open, io_read, io_write, io_seek, io_size, io_setsize, io_close,
   io_lasterror, bitfind, setbits, mempanic, mt_tlsget, mt_tlsaddr,
   mt_tlsset, 0, 0, 0, shl_history, sys_dccommit, gettimediff, hlp_freephys64,
   mem_zero};

void setup_loader_mt(void) {
   qserr   rc;
   if (exptable.in_mtmode) log_it(0, "????\n");
   exptable.in_mtmode = 1;

   if (mt_muxcreate(0, "__lxloader_mux__", &exptable.ldr_mutex) ||
      io_setstate(exptable.ldr_mutex, IOFS_DETACHED, 1) ||
         mt_muxcreate(0, "__microfsd_mux__", &exptable.mfs_mutex) ||
            io_setstate(exptable.mfs_mutex, IOFS_DETACHED, 1))
               log_it(0, "ldr/fsd mutexes err!\n");

   setup_fio_mt(&mod_secondary->dsk_canread, &mod_secondary->dsk_canwrite);
}

void setup_loader(void) {
   // export function table
   mod_secondary = &exptable;
   // flush log first time
   log_flush();
   // minor print
   log_printf("i/o delay value: %d\n", IODelay);
}
