//
// QSINIT "start" module
// volume functions
// -------------------------------------
//
// moved here from QSINIT binary

#include "qsutil.h"
#include "syslocal.h"
#include "qcl/sys/qsedinfo.h"
#include "dskinfo.h"
#include "qstask.h"
#include "qssys.h"
#include "qsinit.h"

extern vol_data *_extvol;
u16t               SysCP = 0; ///< system codepage (read value for START code)

static void _std notify_diskremove(sys_eventinfo *cbinfo) {
   if (cbinfo->eventtype==SECB_DISKREM) {
      u32t disk = cbinfo->info;
      // is disk present and emulated? then unmount all
      if ((hlp_diskmode(disk,HDM_QUERY)&HDM_EMULATED)!=0) hlp_unmountall(disk);
   } else
   if (cbinfo->eventtype==SECB_CPCHANGE) {
      codepage_info *cpi = (codepage_info*)cbinfo->data;
      SysCP = cpi->cpnum;
   }
}

void setup_cache(void) {
   sys_notifyevent(SECB_DISKREM|SECB_CPCHANGE|SECB_GLOBAL, notify_diskremove);
}

/** query disk text name.
    @param disk     disk number
    @param buffer   buffer for result, can be 0 for static. At least 8 bytes.
    @return 0 on error, buffer or static buffer ptr */
char* _std dsk_disktostr(u32t disk, char *buffer) {
   if (!buffer) {
      u64t *rc;
      mt_tlsaddr(QTLS_DSKSTRBUF, &rc);
      buffer = (char*)rc;
   }
   // drop known service flags
   disk&=~(QDSK_IAMCACHE|QDSK_IGNACCESS);

   if (disk&QDSK_FLOPPY) snprintf(buffer, 8, "fd%d", disk&QDSK_DISKMASK); else
   if (disk&QDSK_VOLUME) snprintf(buffer, 8, "%c:", 'A'+(disk&QDSK_DISKMASK)); else
   if ((disk&QDSK_DISKMASK)==disk)
      snprintf(buffer, 8, "hd%d", disk&QDSK_DISKMASK);
         else { *buffer=0; return 0; }
   return buffer;
}

u32t _std dsk_strtodisk(const char *str) {
   if (str && *str) {
      u32t disk, floppies, 
          disks = hlp_diskcount(&floppies);
      char   c1 = toupper(str[0]),
             c2 = toupper(str[1]);
      
      do {
         if ((c1=='H'||c1=='F') && isalnum(c2)) { // hd0 h0 hdd0 fd0 f0 fdd0
            int idx = isalpha(c2)?1:0;
            if (idx && c2!='D') break; else
            if (toupper(str[2])=='D') idx++;
            // check for 0..9
            if (!isdigit(str[1+idx])) break;
            disk = str2long(str+1+idx);
            if (disk>=(c1=='F'?floppies:disks)) break;
            if (c1=='F') disk|=QDSK_FLOPPY;
         } else
         if (c2==':' && isalnum(c1)) {            // 0: 1: A: B:
            if (isalpha(c1)) c1 = c1-'A'+'0';
            disk = c1-'0';
            if (disk>=DISK_COUNT) break;
            disk|=QDSK_VOLUME;
         } else break;
      
         return disk;
      } while (0);
   }
   return FFFF;
}

qserr _std dsk_fakechs(disk_geo_data *geo) {
   u32t shift, heads, spt;

   if (!geo) return E_SYS_ZEROPTR;
   // all parameters for calulation should be zero-filled!
   if (!geo->TotalSectors || geo->Heads>255 || geo->SectOnTrack>255)
      return E_SYS_INVPARM;
   // default sector size
   if (!geo->SectorSize) geo->SectorSize = 512;

   shift = bsf32(geo->SectorSize);
   if (shift!=bsr32(geo->SectorSize) || shift<9 || shift>12) return E_DSK_SSIZE;

   heads = geo->Heads ? geo->Heads : 1;
   spt   = geo->SectOnTrack ? geo->SectOnTrack : 17;
   // just deny all < 17 sectors?
   if (heads * spt > geo->TotalSectors) return E_SYS_TOOSMALL;

   while ((u64t)1024*spt*heads < geo->TotalSectors) {
      if (!geo->SectOnTrack && spt<63)
         { spt   = (spt   + 1 & 0xF0) * 2 - 1; continue; }
      if (!geo->Heads && heads<255)
         { heads = (heads + 1) * 2 - 1; continue; }
      break;
   }
   if (geo->TotalSectors / (spt*heads) > (u64t)FFFF) return E_SYS_TOOLARGE;

   geo->Cylinders   = geo->TotalSectors / (spt*heads);
   geo->Heads       = heads;
   geo->SectOnTrack = spt;

   return 0;
}

char* _std dsk_formatsize(u32t sectsize, u64t disksize, int width, char *buf) {
   static char *suffix[] = { "kb", "mb", "gb", "tb", "pb", "eb", "zb" }; // ;)
   char fmt[16];
   u64t size = disksize * (sectsize?sectsize:1);
   int idx = 0;
   size>>=10;
   while (size>=100000) { size>>=10; idx++; }
   if (width>4) sprintf(fmt, "%%%dd %%s", width-3);
   if (!buf) {
      u64t *rc;
      mt_tlsaddr(QTLS_DSKSZBUF, &rc);
      buf = (char*)rc;
   }
   sprintf(buf, width>4?fmt:"%d %s", (u32t)size, suffix[idx]);
   return buf;
}

u32t _std hlp_unmountall(u32t disk) {
   if (_extvol) {
      u32t umcnt=0, ii;
      for (ii=2; ii<DISK_COUNT; ii++) {
         vol_data *vdta = _extvol+ii;
         // mounted?
         if ((vdta->flags&VDTA_ON)!=0)
            if (vdta->disk==disk) { hlp_unmountvol(ii); umcnt++; }
      }
      return umcnt;
   }
   return 0;
}

qserr _std io_eject(u32t disk, u32t flags) {
   struct qs_diskinfo *di;
   qserr  res = 0;
   // verify that disk exists and emulated and delete is possible
   mt_swlock();
   di = hlp_diskstruct(disk, 0);
   if (!di) res = E_DSK_DISKNUM; else
      if ((di->qd_flags&HDDF_PRESENT)==0) res = E_DSK_DISKNUM; else
         if ((di->qd_flags&HDDF_HOTSWAP)==0) res = E_DSK_FIXED; else
            if (!di->qd_extremove) res = E_SYS_UNSUPPORTED;
   mt_swunlock();
   if (res) return res;

   // unmount in unlocked state (with possible popups)
   if (_extvol) {
      u32t ii;
      for (ii=2; ii<DISK_COUNT; ii++) {
         vol_data *vdta = _extvol+ii;
         // mounted?
         if ((vdta->flags&VDTA_ON)!=0)
            if (vdta->disk==disk)
               if ((res = io_unmount(ii, flags)))
                  if (res!=E_DSK_NOTMOUNTED) return res;
      }
   }
   // check it again (another thread can change anything)
   mt_swlock();
   di = hlp_diskstruct(disk, 0);
   if (di && (di->qd_flags&(HDDF_PRESENT|HDDF_HOTSWAP))==(HDDF_PRESENT|
      HDDF_HOTSWAP))
   {
      if (!di->qd_extremove) res = E_SYS_UNSUPPORTED; else
         res = di->qd_extremove(disk);
   } else
      res = E_DSK_DISKNUM;
   mt_swunlock();

   return res;
}

qserr _std hlp_diskflush(u32t disk) {
   struct qs_diskinfo *di;
   qserr rc = 0;

   if (disk & QDSK_VOLUME) {
      u8t drv = disk&QDSK_DISKMASK;
      if (drv>=DISK_COUNT) rc = E_DSK_DISKNUM; else {
         vol_data *vdta = _extvol + drv;
         mt_swlock();
         if (vdta->flags&VDTA_VFS) rc = E_DSK_NOTPHYS; else
            if (vdta->flags&VDTA_RO) rc = E_DSK_WP; else disk = vdta->disk;
         mt_swunlock();
      }
   }
   if (rc) return rc;

   mt_swlock();
   di = hlp_diskstruct(disk, 0);
   rc = E_SYS_UNSUPPORTED;

   if (!di || (di->qd_flags&HDDF_PRESENT)==0) rc = E_DSK_DISKNUM; else
      if (di->qd_flags&HDDF_HOTSWAP) {
         // call flush function if available ...
         if (di->qd_extflush) rc = di->qd_extflush(disk);
      }
   mt_swunlock();
   return rc;
}

qs_extdisk _std hlp_diskclass(u32t disk, const char *name) {
   struct qs_diskinfo *di;
   qs_extdisk       rcptr = 0;

   mt_swlock();
   di = hlp_diskstruct(disk, 0);
   if (di && di->qd_extptr)
      // "name" is for extension, now only "qs_extdisk" supported
      if (!name || strcmp(name,"qs_extdisk")==0) {
         u32t id = exi_classid(di->qd_extptr), mcnt;
         if (id) {
            mcnt = exi_methods(id);
            // can`t check compat., but at least, can compare number of methods!
            if (mcnt == sizeof(_qs_extdisk)/sizeof(void*)) 
               rcptr = (qs_extdisk)di->qd_extptr;
            else // something was supplied with other number of methods?
               log_it(2, "wrong extptr for disk %03X (%d funcs)\n", disk, mcnt);
         }
      }
   mt_swunlock();
   return rcptr;
}

u32t _std hlp_disktype(u32t disk) {
   // validate disk arg
   if ((disk&~(QDSK_FLOPPY|QDSK_VOLUME|QDSK_DISKMASK|QDSK_DIRECT))==0 &&
      (disk&(QDSK_FLOPPY|QDSK_VOLUME))!=(QDSK_FLOPPY|QDSK_VOLUME))
   {
      struct qs_diskinfo *qe = 0;
      u32t                rv = HDT_INVALID;

      if (disk&QDSK_VOLUME) {
         u8t vol = disk&QDSK_DISKMASK;
         disk    = FFFF;

         if (vol==DISK_LDR) rv = HDT_RAMDISK; else
         if (vol<DISK_COUNT && _extvol) {
            mt_swlock();
            if (_extvol[vol].flags&VDTA_ON) disk = _extvol[vol].disk;
         }
      } else
         mt_swlock();
      // skip B: & invalid volume, it also not locked in such cases
      if (disk!=FFFF) {
         disk &= QDSK_FLOPPY|QDSK_DISKMASK;
         qe = hlp_diskstruct(disk, 0);
         
         if (qe) {
            if (qe->qd_flags&HDDF_HOTSWAP) {
               qs_extdisk exc = hlp_diskclass(disk, 0);
               u32t  extstate = exc ? exc->state(EDSTATE_QUERY) : 0;
         
               rv = extstate&EDSTATE_RAM?HDT_RAMDISK:HDT_EMULATED;
            } else
            // filter cd-rom & real 1.44 (not a flash drive, reported as floppy)
            if (qe->qd_sectorsize==2048) rv = HDT_CDROM; else
               if (qe->qd_sectors<=2880*2) rv = HDT_FLOPPY; else
                  rv = HDT_REGULAR;
         }
         mt_swunlock();
      }
      return rv;
   }
   return HDT_INVALID;
}

void _std hlp_diskdump(printf_function pfn) {
   u32t   ii;
   if (pfn==0) pfn = log_printf;

   pfn("=== Disk table ===\n");
   // actually MAX_QS_DISK already includes MAX_QS_FLOPPY
   for (ii=0; ii<MAX_QS_DISK+MAX_QS_FLOPPY; ii++) {
      struct qs_diskinfo *di;
      char      *l1 = 0, *l2 = 0, *l3 = 0;
      u32t              disk = ii<MAX_QS_DISK ? ii : ii-MAX_QS_DISK|QDSK_FLOPPY;
      mt_swlock();
      di = hlp_diskstruct(disk, 0);
      if (di) {
         qs_extdisk  ep = 0;
         u32t       est = 0;
         if (di->qd_extptr) ep = (qs_extdisk)di->qd_extptr;
         if (ep) est = ep->state(EDSTATE_QUERY);

         l1 = sprintf_dyn("%02X: %s, sector:%u, CHS:%u/%u/%u, BIOS h:%u spt:%u disk:%02X type:%u\n",
            disk, dsk_formatsize(di->qd_sectorsize, di->qd_sectors, 0, 0),
            di->qd_sectorsize, di->qd_cyls, di->qd_heads, di->qd_spt,
            di->qd_chsheads, di->qd_chsspt, di->qd_biosdisk, di->qd_mediatype);
         l2 = sprintf_dyn("   %s%s%s%s%s%s%s%s%s\n", di->qd_flags&HDDF_LBAON?
            " LBA+":di->qd_flags&HDDF_LBAPRESENT?" LBA-":"",
            di->qd_flags&HDDF_REMOVABLE?" Removable":"",
            di->qd_flags&HDDF_HOTSWAP?" Virtual":"",
            di->qd_extread && di->qd_extwrite?" ExtRW":"", 
            di->qd_extremove?" Eject":"", 
            di->qd_extflush?" Flush":"",
            est&EDSTATE_RO?" R/o":"",
            est&EDSTATE_NOCACHE?" NoCache":"",
            est&EDSTATE_RAM?" Ramdisk":"");
         // drop second line if it`s empty ...
         if (strlen(l2)<=4) { free(l2); l2=0; }
         // get verbose disk information ...
         if (ep) {
            char *name = ep->getname(1);
            if (name) {
               l3 = sprintf_dyn("    %s\n", name);
               free(name);
            }
         }
      }
      mt_swunlock();
      if (l1) { pfn(l1); free(l1); }
      if (l2) { pfn(l2); free(l2); }
      if (l3) { pfn(l3); free(l3); }
   }
}
