#define Uses_MsgBox
#define Uses_TEvent

#include <tv.h>

#include "diskedit.h"
#include "ldrdlg.h"

TLdrBootDialog::TLdrBootDialog(const char *ldrname, int mode) :
       TDialog(TRect(16, 7, 63, 16), mode==LdrBootEFI?"UEFI application launch":
       "OS2LDR file launch"), TWindowInit(TLdrBootDialog::initFrame)
{
   TView *control;
   options |= ofCenterX | ofCenterY;
   helpCtx  = mode==LdrBootEFI?hcBootEFI:hcLdrName;
   runmode  = mode;

   bf_name  = new TInputLine(TRect(13, 2, 42, 3), 29);
   bf_name->helpCtx = helpCtx;
   insert(bf_name);
   insert(new THistory(TRect(42, 2, 45, 3), (TInputLine*)bf_name, 2));
   if (ldrname) setstrn(bf_name, ldrname, 64);

   insert(new TLabel(TRect(2, 2, 12, 3), "~F~ile name", bf_name));

   if (runmode!=LdrBootEFI) {
      bf_opts = new TCheckBoxes(TRect(3, 5, 27, 7), new TSItem("~r~aw mode",
                                new TSItem("~n~o data sending", 0)));
      bf_opts->helpCtx = hcLdrOpts;
      insert(bf_opts);
      insert(new TLabel(TRect(2, 4, 10, 5), "~O~ptions", bf_opts));
      if (mode==LdrBootRaw) bf_opts->press(0);
      bf_args = 0;
   } else {
      bf_opts = 0;
      bf_args = new TInputLine(TRect(13, 4, 45, 5), 255);
      bf_args->helpCtx = hcEfiArgs;
      insert(bf_args);
      insert(new TLabel(TRect(2, 4, 12, 5), "~A~rguments", bf_args));
   }

   control = new TButton(runmode==LdrBootEFI ? TRect(20, 6, 30, 8) :
                         TRect(33, 4, 43, 6), "~B~rowse", cmLdrBrowse, bfNormal);
   control->helpCtx = hcLdrBrowse;
   insert(control);
   control = new TButton(TRect(33, 6, 43, 8), "~G~o", cmLdrLaunch, bfDefault);
   control->helpCtx = hcLdrLoad;
   insert(control);

   selectNext(False);
}

void TLdrBootDialog::handleEvent(TEvent& event) {
   TDialog::handleEvent(event);
   if (event.what==evCommand) {
      switch (event.message.command)  {
         case cmLdrLaunch: {
            char *ldr = getstr(bf_name);
            if (ldr && *ldr) {
#ifdef __QSINIT__
               ldr = strdup(ldr);
               // getstr() buffer is static so we need strdup() above
               char *args = bf_args ? getstr(bf_args) : 0;
               TProgram::application->suspend();
               u32t rc = runmode==LdrBootEFI ? exit_efiboot(ldr, args) :
                  bf_opts->mark(0) ? exit_rawcode(ldr, 0) :
                  exit_restart(ldr, 0, 0, bf_opts->mark(1)?QRST_NOHIST|QRST_NOCP:0);
               TProgram::application->resume();
               TProgram::application->redraw();
               free(ldr);
#else
               u32t rc = 14;
#endif
               SysApp.PrintPTErr(rc);
            }
            clearEvent(event);
            // route Enter with the empty file name to the Browse command ...
            if (ldr && *ldr) break;
         }
         case cmLdrBrowse: {
            char *bn = getfname(1, runmode==LdrBootEFI?"Select EFI module":
               "Select loader file", runmode==LdrBootEFI?"A:\\*.EFI":"A:\\*");
            if (bn) bf_name->setData(bn);
            clearEvent(event);
            // move focus to "Go"
            selectNext(False);
            break;
         }
      }
   }
}

Boolean TLdrBootDialog::valid(ushort command) {
   Boolean rslt = TDialog::valid(command);
   if (rslt && (command == cmOK)) { }
   return rslt;
}
