#include "filedlg.h"
#define Uses_TFileDialog
#define Uses_MsgBox
#define Uses_TRect
#define Uses_TFileInputLine
#define Uses_TButton
#define Uses_TLabel
#define Uses_TFileList
#define Uses_THistory
#define Uses_TScrollBar
#define Uses_TEvent
#define Uses_TFileInfoPane
#define Uses_TScreen
#define Uses_TKeys
#include <tv.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>

#ifndef __BORLANDC__
struct  ftime   {
   unsigned    ft_tsec  : 5;   /* Two second interval */
   unsigned    ft_min   : 6;   /* Minutes */
   unsigned    ft_hour  : 5;   /* Hours */
   unsigned    ft_day   : 5;   /* Days */
   unsigned    ft_month : 4;   /* Months */
   unsigned    ft_year  : 7;   /* Year */
};
#endif

#define cpInfoPane2 "\x06\x1E"

class TFileInfoPane2 : public TFileInfoPane {
public:
   TFileInfoPane2(const TRect &bounds) : TFileInfoPane(bounds) {}
   virtual TPalette &getPalette() const;
   virtual void draw();
};

void TFileInfoPane2::draw() {
   TDrawBuffer  b;
   ushort   color;
   char      path[MAXPATH];

   // Prevents incorrect directory name display in info pane if wildCard
   // has already been expanded.
   strcpy(path, ((TAppFileDialog *)owner)->wildCard);
   if (!strchr(path, ':') && !strchr(path, '\\')) {
      strcpy(path, ((TAppFileDialog *)owner)->directory);
      strcat(path, file_block.name);
      fexpand(path);
   }

   color = getColor(0x01);
   b.moveChar(0, ' ', color, (ushort) size.x);
   if (strlen(path)+3 > size.x) {
      b.moveStr(3, path+(strlen(path)-size.x+4), color);
      b.moveStr(1, "..", color);
   } else {
      b.moveStr(1, path, color);
   }
   writeLine(0, 0, (ushort) size.x, 1, b);

   color = getColor(0x02);
   b.moveChar(0, ' ', color, (ushort) size.x);

   if (*(file_block.name) != EOS) {
      char    buf[24];
      ftime *time = (ftime *) &file_block.time;
      
      b.moveStr(1, months[time->ft_month], color);

      if (time->ft_day >= 10)
         itoa(time->ft_day, buf, 10);
      else {
         buf[0] = '0';
         itoa(time->ft_day, buf+1, 10);
      }
      b.moveStr(5, buf, color);
      b.putChar(7, ',');

      itoa(time->ft_year+1980, buf, 10);
      b.moveStr(8, buf, color);
      
      if (time->ft_hour >= 10)
         itoa(time->ft_hour, buf, 10);
      else {
         buf[0] = '0';
         itoa(time->ft_hour, buf+1, 10);
      }
      b.moveStr(14, buf, color);
      b.putChar(16, ':');

      if (time->ft_min >= 10)
         itoa(time->ft_min, buf, 10);
      else {
         buf[0] = '0';
         itoa(time->ft_min, buf+1, 10);
      }
      b.moveStr(17, buf, color);

      if (file_block.attr & FA_DIREC) {
         b.moveStr(21, "      <Directory>", color);
      } else {
         // watcom target cannot provide high dword via find_t
         unsigned long long fsz = file_block.size;
         snprintf(buf, 24, "%17Lu", fsz);
         b.moveStr(21, buf, color);
      }
      if (file_block.attr & FA_RDONLY) b.moveStr(42, "R/o", color);
      if (file_block.attr & FA_HIDDEN) b.moveStr(46, "Hid", color);
      if (file_block.attr & FA_SYSTEM) b.moveStr(50, "Sys", color);
      if (file_block.attr & FA_ARCH  ) b.moveStr(54, "Ar", color);
   }

   writeLine(0, 1, (ushort) size.x, 1, b);
   b.moveChar(0, ' ', color, (ushort) size.x);
   writeLine(0, 2, (ushort) size.x, (ushort)(size.y - 2), b);
}

TPalette &TFileInfoPane2::getPalette() const {
   static TPalette palette(cpInfoPane2, sizeof(cpInfoPane2)-1);
   return palette;
}

// remove all ~ and return clean string in a static(!) buffer
static char *unhotkey(char const *str) {
   static char buf[80];
   int len = 0;
   while (*str && len<sizeof(buf)-1)
      if (*str!='~') buf[len++] = *str++; else str++;
   buf[len] = 0;
   return buf;
}

TAppFileDialog::TAppFileDialog(const char *aWildCard, const char *aTitle,
   const char *inputName, ushort aOptions, uchar histId) :
   TDialog(TRect(10, 1, 69, 20 +
   (TScreen::screenHeight>=40?12:TScreen::screenHeight>=32?4:0)), aTitle),
   directory(newStr("")),
   TWindowInit(TAppFileDialog::initFrame)
{
   options |= ofCentered;
   strcpy(wildCard, aWildCard);

   fileName = new TFileInputLine(TRect(3, 3, 43, 4), MAXPATH);
   strcpy(fileName->data, wildCard);
   insert(fileName);

   int add_y = size.y - 19;

   /* remove all hotkeys in this dialog to allow most of Alt-? keys
      work as a disk selection */

   insert(new TLabel(TRect(2, 2, 3+cstrlen(inputName), 3), unhotkey(inputName), fileName));
   insert(new THistory(TRect(43, 3, 46, 4), fileName, histId));
   TScrollBar *sb = new TScrollBar(TRect(3, 14+add_y, 46, 15+add_y));
   insert(sb);
   insert(fileList = new TFileList(TRect(3, 6, 46, 14+add_y), sb));
   insert(new TLabel(TRect(2, 5, 8, 6), "Files", fileList));

   ushort opt = bfDefault;
   TRect r(47, 3, 57, 5);

   if ((aOptions & fdOpenButton) != 0) {
      insert(new TButton(r, unhotkey(TFileDialog::openText), cmFileOpen, opt));
      opt = bfNormal;
      r.a.y += 3;
      r.b.y += 3;
   }

   if ((aOptions & fdOKButton) != 0) {
      insert(new TButton(r, unhotkey(TFileDialog::okText), cmFileOpen, opt));
      opt = bfNormal;
      r.a.y += 3;
      r.b.y += 3;
   }

   if ((aOptions & fdReplaceButton) != 0) {
      r.b.x++;
      insert(new TButton(r, unhotkey(TFileDialog::replaceText), cmFileReplace, opt));
      r.b.x--;
      opt = bfNormal;
      r.a.y += 3;
      r.b.y += 3;
   }

   if ((aOptions & fdClearButton) != 0) {
      insert(new TButton(r, TFileDialog::clearText, cmFileClear, opt));
      r.a.y += 3;
      r.b.y += 3;
   }

   insert(new TButton(r, TFileDialog::cancelText, cmCancel, bfNormal));
   r.a.y += 3;
   r.b.y += 3;

   if ((aOptions & fdHelpButton) != 0) {
      insert(new TButton(r, TFileDialog::helpText, cmHelp, bfNormal));
      r.a.y += 3;
      r.b.y += 3;
   }

   insert(new TFileInfoPane2(TRect(1, 16+add_y, 58, 18+add_y)));

   selectNext(False);
   if ((aOptions & fdNoLoadDir) == 0) {
#ifdef QSINIT
      // make sure that currect directory is valid
      if (relativePath(wildCard))
         char curDir[MAXPATH];
         getCurDir(curDir);
         if (!isDir(curDir)) chdir("b:\\");
      }
#endif
      readDirectory();
   }
}

TAppFileDialog::~TAppFileDialog() {
   delete(char *)directory;
}

void TAppFileDialog::shutDown() {
   fileName = 0;
   fileList = 0;
   TDialog::shutDown();
}

static Boolean relativePath(const char *path) {
   if (path[0] != EOS && (path[0] == '\\' || path[1] == ':'))
      return False;
   else
      return True;
}

static void noWildChars(char *dest, const char *src) {
   while (*src != EOS) {
      if (*src != '?' && *src != '*') *dest++ = *src;
      src++;
   }
   *dest = EOS;
}

#if defined(__MSDOS__)
/* 'src' is cast to unsigned char * so that isspace sign extends it
   correctly. */
static void trim(char *dest, const char *src) {
   while (*src != EOS &&  isspace(uchar(*src))) src++;
   while (*src != EOS && !isspace(uchar(*src))) *dest++ = *src++;
   *dest = EOS;
}
#elif defined(__QSINIT__)
static void trim(char *dest, const char *src) {
   strcpy(dest, src);
   trimright(dest, " ");
}
#else
#define trim strcpy
#endif

void TAppFileDialog::getFileName(char *s) {
   char buf  [2*MAXPATH];
   char drive[MAXDRIVE], path[MAXDIR], name[MAXFILE], ext[MAXEXT];
   char TName[MAXFILE],  TExt[MAXEXT];

   if (fileName->data[0] == '.' && (fileName->data[1] == '\\' ||
       fileName->data[1] == '\0'))
   {
      getcwd(buf, sizeof(buf));
      if (fileName->data[2]) trim(buf + strlen(buf), fileName->data+1);
   } else {
      trim(buf, fileName->data);
   }
   if (relativePath(buf) == True) {
      strcpy(buf, directory);
      trim(buf + strlen(buf), fileName->data);
   }
   fexpand(buf);
   fnsplit(buf, drive, path, name, ext);
   //    printf("split %s\n drive=%s,path=%s,name=%s,ext=%s\n",buf,drive,path,name,ext);
   if ((name[0] == EOS || ext[0] == EOS) && !isDir(buf)) {
      fnsplit(wildCard, 0, 0, TName, TExt);
      if (name[0] == EOS && ext[0] == EOS)
         fnmerge(buf, drive, path, TName, TExt);
      else if (name[0] == EOS)
         fnmerge(buf, drive, path, TName, ext);
      else if (ext[0] == EOS) {
         if (isWild(name))
            fnmerge(buf, drive, path, name, TExt);
         else {
            fnmerge(buf, drive, path, name, 0);
            noWildChars(buf + strlen(buf), TExt);
         }
      }
   }
   strcpy(s, buf);
   //    printf("getFileName %s\n",s);
}

void TAppFileDialog::handleEvent(TEvent &event) {
   if (event.what == evKeyDown) {
      static ushort dcode[] = { kbAltA, kbAltB, kbAltC, kbAltD, kbAltE,
        kbAltF, kbAltG, kbAltH, kbAltI, kbAltJ, kbAltK, kbAltL, kbAltM,
        kbAltN, kbAltO, kbAltP, kbAltQ, kbAltR, kbAltS, kbAltT, kbAltU,
        kbAltV, kbAltW, kbAltX, kbAltY, kbAltZ, 0 };
      // select disk via Alt-? key ...
      for (int ii=0; dcode[ii]; ii++)
         if (event.keyDown.keyCode==dcode[ii]) {
            char *cwd = _getdcwd(ii, 0, 0);
            int    ok = cwd && isDir(cwd);
            if (!ok) {
               if (cwd) free(cwd);
               cwd = (char*)malloc(16);
               snprintf(cwd, 16, "%c:\\", ii+'A');
               ok  = isDir(cwd);
            }
            if (ok) {
               cwd = (char*)realloc(cwd, strlen(cwd) + 6);
               strcat(cwd, cwd[strlen(cwd)-1]=='\\' ? "*.*" : "\\*.*");
               setData(cwd);
               free(cwd);
               clearEvent(event);
               return;
            }
         }
   }

   TDialog::handleEvent(event);
   if (event.what == evCommand) {
      switch (event.message.command) {
      case cmFileOpen:
      case cmFileReplace:
      case cmFileClear:
         chdir(directory);
         endModal(event.message.command);
         clearEvent(event);
         break;
      default:
         break;
      }
   } else
   if (event.what == evBroadcast && event.message.command == cmFileDoubleClicked) {
      event.what = evCommand;
      event.message.command = cmOK;
      putEvent(event);
      clearEvent(event);
   }
}

void TAppFileDialog::readDirectory() {
   char curDir[MAXPATH];
   if (relativePath(wildCard)) {
      //      fileList->readDirectory( wildCard );
      getCurDir(curDir);
   } else {
      char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];
      fnsplit(wildCard, drive, dir, name, ext);
      strcpy(curDir,drive);
      strcat(curDir,dir);
      strcpy(wildCard,name);
      strcat(wildCard,ext);
   }
   delete(char *)directory;
   directory = newStr(curDir);
   fileList->readDirectory(directory, wildCard);
   //    printf("curDir=%s, directory=%p\n",curDir,directory);
}

void TAppFileDialog::setData(void *rec) {
   TDialog::setData(rec);
   if (*(char *)rec != EOS && isWild((char *)rec)) {
      valid(cmFileInit);
      fileName->select();
   }
}

void TAppFileDialog::getData(void *rec) {
   getFileName((char *)rec);
}

Boolean TAppFileDialog::checkDirectory(const char *str) {
   if (pathValid(str))
      return True;
   else {
      messageBox(TFileDialog::invalidDriveText, mfError | mfOKButton);
      fileName->select();
      return False;
   }
}

Boolean TAppFileDialog::valid(ushort command) {
   char fName[MAXPATH], drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];

   if (!TDialog::valid(command)) return False;

   // The base TDialog::valid() doesn't handle these.
   if (command == cmValid || command == cmCancel || command == cmFileClear)
      return True;

   getFileName(fName);
   if (isWild(fName)) {
      fnsplit(fName, drive, dir, name, ext);
      char path[MAXPATH];
      strcpy(path, drive);
      strcat(path, dir);
      if (checkDirectory(path)) {
         delete(char *)directory;
         directory = newStr(path);
         strcpy(wildCard, name);
         strcat(wildCard, ext);
         if (command != cmFileInit) fileList->select();
         fileList->readDirectory(directory, wildCard);
      }
      return False;
   }
   if (isDir(fName)) {
      if (checkDirectory(fName)) {
         delete(char *)directory;

         if (fName[strlen(fName)-1]!='\\') strcat(fName, "\\");

         directory = newStr(fName);
         if (command != cmFileInit) fileList->select();
         fileList->readDirectory(directory, wildCard);
      }
      return False;
   }
   if (validFileName(fName)) return True;

   messageBox(TFileDialog::invalidFileText, mfError | mfOKButton);
   return False;
}
