//
// QSINIT "vdisk" module
// PAE ramdisk creation and QSINIT i/o
//
#include "qsbase.h"
#include "doshlp.h"
#include "stdlib.h"
#include "qcl/qslist.h"
#include "vdproc.h"
#include "qcl/sys/qsedinfo.h"
#include "qsint.h"
#include "dskinfo.h"
#include "rmeinfo.h"
#include "biosdata.h"
#include "qsdump.h"

#define MAP_SIZE                  (_8MB)
#define PAGE_SIZE                 (_2MB)
#define PAGESIN4GB            (0x100000)   // # of pages in 4Gb
// read size, which SHOULD fit into mapped MAP_SIZE area, aligned to PAGE_SIZE
#define READ_IN_TIME              ((MAP_SIZE-PAGE_SIZE)/512 - 8)

static HD4_Header   *cdh = 0;   ///< active disk header
static dq_list     cdmem = 0;   ///< active disk memory list
static u64t      cdhphys = 0;   ///< disk header physical address
static u32t         chdd = 0;   ///< QS disk handle
static u32t      cdhsize = 0;   ///< disk size (in sectors)
static HD4_TabEntry *cde = 0;
static u32t     cdecount = 0;
static char     *maparea = 0;
static u64t     mapstart = 0;
static u64t    min4Gaddr = 0;   ///< lowest high memory address
static u32t    min4Gsize = 0;   ///< size of lowest high memory block (in pages)
static u32t     endofram = 0;
static int       hidemem = 0;
static u32t  classid_ext = 0;   ///< external info "qs_extdisk" compatible class
static qs_extdisk  eiptr = 0;
u32t          classid_vd = 0;   ///< qs_vdisk class id
extern struct
I15hListInfo biosfltinfo;

void  wbinvd(void);
#pragma aux wbinvd = "wbinvd" modify exact [];

static u32t io(u32t disk, u64t sector, u32t count, void *data, int write) {
   u32t ii, svcount;
   if (!cdh) return 0;
   // overflow check
   if (sector>=(u64t)cdhsize) return 0;
   if (sector+count>(u64t)cdhsize) count = cdhsize-sector;

   if (!endofram) endofram = sys_endofram();

   ii = 0;
   svcount = count;

   while (count && ii<cdecount) {
      if (cde[ii].hde_sectors>sector) {
         u32t toread = cde[ii].hde_sectors - sector, cdeofs = sector;
         if (toread>count) toread = count;
         // loop by 4Mb temp mapping
         while (toread) {
            u32t  cnt = toread>READ_IN_TIME ? READ_IN_TIME : toread,
                bytes = cnt * 512,
                  ofs = (cdeofs & 7) * 512;
            u64t addr = (u64t)cde[ii].hde_1stpage + (cdeofs>>3) << PAGESHIFT,
                eaddr = addr + bytes + ofs;

            char *ptr =  0;

            /* do not unmap areas below end of ram at all (because it was not
               mapped actually), for any other (must be only above 4Gb areas)
               use 8Mb map space */
            if (eaddr<=(u64t)endofram)
               ptr = (char*)pag_physmap(addr, bytes + ofs, 0);
            else {
               // unmap not matched 8Mb area
               if (maparea)
                  if (mapstart<=addr && eaddr>mapstart && eaddr<=mapstart+MAP_SIZE) ; else {
                     pag_physunmap(maparea);
                     maparea = 0;
                  }
               /* map 8Mb, aligned down to big page phys addr to allow page allocator
                  make 4 big pages for us (much faster) */
               if (!maparea) {
                  mapstart = addr & ~(u64t)(PAGE_SIZE-1);
                  maparea = pag_physmap(mapstart, MAP_SIZE, 0);
               }
               if (maparea) ptr = maparea + (addr - mapstart);
            }

            if (ptr)
               if (write) memcpy(ptr + ofs, data, bytes);
                  else memcpy(data, ptr + ofs, bytes);

            if (!ptr) return svcount - count;

            data    = (char*)data + bytes;
            toread -= cnt;
            count  -= cnt;
            cdeofs += cnt;
         }
         sector = 0;
      } else
         sector-=cde[ii].hde_sectors;
      ii++;
   }
   return svcount;
}

static u32t _std _read(u32t disk, u64t sector, u32t count, void *data) {
   return io(disk, sector, count, data, 0);
}

static u32t _std _write(u32t disk, u64t sector, u32t count, void *data) {
   return io(disk, sector, count, data, 1);
}

// io_eject() calls this
static qserr _std _remove(u32t disk) {
   if (disk!=chdd) return E_DSK_DISKNUM;
   return vdisk_free(0,0) ? 0 : E_SYS_SOFTFAULT;
}

// hlp_diskflush() calls this
static qserr _std _flush(u32t disk) {
   if (disk!=chdd) return E_DSK_DISKNUM;
   wbinvd();
   return 0;
}

static void fill_header(u32t pages) {
   if (cdh) {
      cdhsize = pages<<PAGESHIFT-9;

      memset(cdh, 0, PAGESIZE);
      cdh->h4_sign     = EXPDATA_SIGN;
      cdh->h4_version  = 1;

      cdh->h4_cyls     = 1024;
      cdh->h4_spt      = 17;
      cdh->h4_heads    = 31;

      while (cdh->h4_cyls*cdh->h4_spt*cdh->h4_heads < cdhsize) {
         if (cdh->h4_spt<63) {
            cdh->h4_spt   = (cdh->h4_spt   + 1 & 0xF0) * 2 - 1; continue;
         }
         if (cdh->h4_heads<255) {
            cdh->h4_heads = (cdh->h4_heads + 1 & 0xF0) * 2 - 1; continue;
         }
         break;
      }
      cdh->h4_cyls     = cdhsize / (cdh->h4_spt*cdh->h4_heads);

      cdh->h4_pages    = pages;
      cdh->h4_tabofs   = sizeof(HD4_Header);
      cdh->h4_tabsize  = 1;

      cde      = (HD4_TabEntry*)(cdh+1);
      cdecount = 0;
   }
}

static void wipe_fat32bs(u8t vol) {
   disk_volume_data vi;
   if (hlp_volinfo(vol, &vi)==FST_FAT32) dsk_emptysector(vi.Disk, vi.StartSector, 1);
}

static void set_lvmname(u32t index, str_list* vol_names, u32t *pl_rc) {
   if (vol_names)
      if (vol_names->count>index && vol_names->item[index][0])
         *pl_rc = lvm_setname(chdd, index, LVMN_PARTITION|LVMN_VOLUME,
            vol_names->item[index]);
}

static void set_gptname(u32t index, str_list* vol_names, u32t *pd_rc) {
   if (vol_names)
      if (vol_names->count>index && vol_names->item[index][0]) {
         dsk_gptpartinfo di;
         *pd_rc = dsk_gptpinfo(chdd, index, &di);
         if (!*pd_rc) {
            char *src = vol_names->item[index];
            u32t  len = strlen(src), ii;
            // a bit dumb :)
            if (len>35) len = 35;
            for (ii=0; ii<len; ii++) di.Name[ii] = ff_convert(src[ii], 1);
            di.Name[ii] = 0;
            *pd_rc  = dsk_gptpset(chdd, index, &di);
         }
      }
}

static void mount_persistent(u32t index) {
   char ltr = 0;
   u8t  vol;

   if (sto_dword(STOKEY_LVMMOUNT)) ltr = lvm_ismounted(chdd,index);
   ltr = toupper(ltr);
   // filter out A: B: and *:
   if (ltr && ltr<'C') ltr = 0;
   // don`t remount existing ...
   if (ltr)
      if (io_ismounted(ltr-'A',0)) ltr = 0;

   vol = ltr ? ltr-'A': 0;
   vol_mount(&vol, chdd, index, 0);
}

static u32t make_partition(u32t index, u32t flags, char letter, u32t *pl_rc,
                           str_list* vol_labels)
{
   u32t rc = 0;
   letter  = toupper(letter);
   // set LVM drive letter
   if (*pl_rc==0 && letter && (isalpha(letter) || letter=='*'))
      *pl_rc = lvm_assignletter(chdd, index, letter, 1);
   // and format at last
   if ((flags&VFDF_NOFMT)==0) {
      u8t    vol = sto_dword(STOKEY_LVMMOUNT)&&letter>='C' ? letter-'A' : 0;
      u32t usize = 0;
      /* forced drive letter will unmount already mounted to it, this is wrong
         for LVMMOUNT logic */
      if (vol)
         if (io_ismounted(vol,0)) vol = 0;
      // mount & format.
      rc = vol_mount(&vol, chdd, index, 0);
      // "always-FAT32"
      if (!rc && (flags&VFDF_FAT32)!=0) {
         disk_volume_data vi;
         u32t          csize = _64KB;
         hlp_volinfo(vol, &vi);
         // making too big for FAT16 number of clusters
         if (vi.TotalSectors) {
            while (vi.TotalSectors * (u64t)vi.SectorSize / csize <= _64KB)
               if (csize > vi.SectorSize) csize>>=1; else break;
            if (vi.TotalSectors * (u64t)vi.SectorSize / csize > _64KB)
               usize = csize;
         }
      }
      if (!rc) rc = vol_formatfs(vol, flags&VFDF_HPFS?"HPFS":
         flags&VFDF_JFS?"JFS":"FAT", DFMT_QUICK|DFMT_NOALIGN, usize, 0);
      // volume label is available ...
      if (!rc && vol_labels && vol_labels->count>index && vol_labels->item[index][0]) {
         qserr vlres = io_vollabel(vol, vol_labels->item[index]);
         if (vlres) log_printf("Pt index %u label \"%s\" error %X\n", index,
            vol_labels->item[index], vlres);
      }
      // un-format FAT32 formatted partition
      if (!rc && (flags&VFDF_NOFAT32)!=0) wipe_fat32bs(vol);

   }
   return rc;
}

static int use_lowmem(HD4_Header *dh) {
   if (dh->h4_sign==EXPDATA_SIGN && dh->h4_version==1) {
      HD4_TabEntry *tab = (HD4_TabEntry*)(dh+1);
      u32t ii;
      // enum page ranges
      for (ii=0; ii<dh->h4_tabsize; ii++)
         if (tab[ii].hde_1stpage<PAGESIN4GB) return 1;
   }
   return 0;
}

/// QSINIT exit hook
static void _std qs_shutdown(sys_eventinfo *info) {
   // reboot can drop last memory changes without this (bad for PERSIST)
   if (cdh) wbinvd();
   /* install own int15h memory list only for OS/4 kernel (excluding old
      versions) or at RESTART with disk using memory above 4Gb only */
   if (cdh && /*(info->info==SHUT_BOOTOS2 && sto_dword(STOKEY_MPDATA16) ||
       info->info==SHUT_RESTART && !use_lowmem(cdh))*/ !use_lowmem(cdh) && hidemem)
   {
      struct I15hListInfo *fd;
      HD4_TabEntry       *tab = (HD4_TabEntry*)(cdh+1);
      AcpiMemInfo         *ml = hlp_int15mem();
      ptr_list             lp = NEW(ptr_list);
      u16t             dstseg = int12mem() << 6;
      u32t         ii, tablen, codesize, idx;
      int              update = 0;

      log_printf("ramdisk: %X sectors\n", cdhsize);

      fd   = (struct I15hListInfo*)hlp_segtoflat(dstseg);
      // we can reuse existing code because revision is the same!
      if (fd->I15_sign==MEMLIST_SIGN && fd->I15_revno==biosfltinfo.I15_revno) {
         update   = 1;
         codesize = fd->I15_codesize;
         log_it(3, "reuse table at %04X\n", dstseg);
      } else {
         dstseg   = int12mem() - 1 << 6;
         fd       = (struct I15hListInfo*)hlp_segtoflat(dstseg);
         codesize = biosfltinfo.I15_codesize;
      }
      // move to own list
      for (ii=0; ml[ii].LengthHigh + ml[ii].LengthLow; ii++) {
         AcpiMemInfo *ne = (AcpiMemInfo*)malloc_th(sizeof(AcpiMemInfo));
         u64t       base = *(u64t*)&ml[ii].BaseAddrLow,
                     len = *(u64t*)&ml[ii].LengthLow;
         // patch 640k entry to be safe ..
         if (base+len == (dstseg<<4) + _1KB) ml[ii].LengthLow -=_1KB;

         memcpy(ne, ml+ii, sizeof(AcpiMemInfo));
         lp->add(ne);
      }

      for (ii=0; ii<cdh->h4_tabsize; ii++) {
         u64t rb   = (u64t)tab[ii].hde_1stpage << PAGESHIFT,
              rlen = (u64t)tab[ii].hde_sectors << 9,
              rend = rb + rlen;
         // header must be bound to the first block
         if (cdhphys+PAGESIZE == rb) {
            rb   -= PAGESIZE;
            rlen += PAGESIZE;
         }
         // the same table entry can be affected by multiple disk ranges
         for (idx=0; idx<lp->count(); ) {
            AcpiMemInfo *me = (AcpiMemInfo*)lp->value(idx);
            u64t       base = *(u64t*)&me->BaseAddrLow, orgbase = base,
                        len = *(u64t*)&me->LengthLow, orglen = len,
                        end = base + len;

            if (rb>base) {
               if (rb<end)
                  if (rend >= end) len = rb - base; else {
                     // range in the middle of the block ??
                     AcpiMemInfo *ne = (AcpiMemInfo*)malloc_th(sizeof(AcpiMemInfo));
                     *(u64t*)&ne->BaseAddrLow = base;
                     *(u64t*)&ne->LengthLow   = rb - base;
                     ne->AcpiMemAttr = me->AcpiMemType;
                     lp->insert(idx, ne, 1);
                     log_it(3, "insert %09LX %09LX\n", base, rb - base);
                     base = rend;
                     len  = end - rend;
                  }
            } else
            if (rend > base)
               if (rend >= end) len = 0; else {
                  base = rend;
                  len  = end - base;
               }

            if (!len) log_it(3, "del %09LX %09LX\n", orgbase, orglen);
               else
            if (orglen!=len || orgbase!=base) {
               log_it(3, "%09LX %09LX -> %09LX %09LX\n", orgbase, orglen,
                   base, len);
               *(u64t*)&me->BaseAddrLow = base;
               *(u64t*)&me->LengthLow   = len;
            }
            // remove entry or go next ...
            if (!len) { free(me); lp->del(idx,1); }
               else idx++;
         }
      }
      tablen = lp->count();
      // we have reserved 1kb below the current int12h memory border
      if (codesize + tablen*sizeof(AcpiMemInfo) > _1KB)
         log_printf("list too long(%u)!\n", tablen);
      else {
         struct I15hListVec *vt;
         AcpiMemInfo      *dtab;

         if (!update) memcpy(fd, &biosfltinfo, biosfltinfo.I15_codesize);
         /* during update here can be another version of this code with the
            same revision (i.e. interrupt table match) */
         vt   = (struct I15hListVec*)((u8t*)fd + fd->I15_inttab);
         dtab = (AcpiMemInfo*)((u8t*)fd + fd->I15_codesize);

         log_it(3, "mtable at %X, %u entries\n", dtab, tablen);

         for (ii=0; ii<tablen; ii++, dtab++)
            memcpy(dtab, lp->value(ii), sizeof(AcpiMemInfo));

         fd->I15_tablen   = tablen;
         fd->I15_state    = 1;
         fd->I15_revno    = 2;
         //fd->I15_diskpage = (u32t)(cdhphys>>PAGESHIFT);
         // replace interrupt vectors!
         if (!update) {
            for (ii=0; vt[ii].I15v_intnum!=0xFF; ii++) {
               u32t ptr = MAKEFAR16(dstseg,vt[ii].I15v_offset), orgv,
                   vofs = vt[ii].I15v_intnum*4;
            
               hlp_memcpy(&orgv, (void*)vofs, 4, MEMCPY_PG0|MEMCPY_DI);
               hlp_memcpy((void*)vofs, &ptr, 4, MEMCPY_PG0|MEMCPY_DI);
               vt[ii].I15v_oldvec = orgv;
            
               log_it(3, "int %02Xh %04X:%04X->%04X:%04X\n", vt[ii].I15v_intnum,
                  orgv>>16, orgv&0xFFFF, ptr>>16, ptr&0xFFFF);
            
            }
            // memory size in the BIOS data ...
            ii = dstseg >> 6;
            hlp_memcpy((void*)(0x400 + FIELDOFFSET(struct BiosDataArea,BDA_MemSize)),
               &ii, 2, MEMCPY_PG0|MEMCPY_DI);
            // eBDA size
            fd->I15_reserved[0] = (u8t)(640 - ii);
            //log_it(3, "eBDA size %u\n", fd->I15_reserved[0]);
         }
      }

      free(ml);
      lp->freeitems(0,FFFF);
      DELETE(lp);
   }
}

/// query number of available pages ...
static void get_memory(u32t *lopages, u32t *hipages) {
   pcmem_entry  *me, *pe;
   u32t         low = 0,
               high = 0;
   me = sys_getpcmem(0);
   if (me) {
      pe = me;
      while (pe->pages) {
         u32t owner = pe->flags&PCMEM_TYPEMASK;
         if (pe->start>=_4GBLL && pe->pages>=_4MB>>PAGESHIFT &&
            (owner==PCMEM_FREE || owner==PCMEM_QSINIT))
         {
            // address fits in 44 bits?
            if (pe->start>>PAGESHIFT<=(u64t)FFFF) {
               high += pe->pages;
               // find lowest page
               if (min4Gaddr && pe->start<min4Gaddr || !min4Gaddr) {
                  min4Gaddr = pe->start;
                  min4Gsize = pe->pages;
               }
            }
         }
         if (pe->start<_4GBLL && pe->pages>=_1MB>>PAGESHIFT &&
            (owner==PCMEM_FREE || owner==PCMEM_QSINIT))
               low += pe->pages;
         pe++;
      }
      free(me);
   }
   if (lopages) *lopages = low;
   if (hipages) *hipages = high;
}

static qserr check_pae(void) {
   qserr rc = 0;
   u32t  pm = sys_pagemode();
   if (pm==QPGM_FLAT) {
      rc = pag_enable(PGRN_RAMDISK);
      if (rc) {
         log_printf("set PAE (%d)\n", rc);
         return rc;
      }
   } else
   // PAE mode already set - ask everybody to free the high memory
   if (pm & QPGM_PAE)
      sys_notifyexec(SECB_LOWMEM, QSMA_HIGH);

   return sys_pagemode()&QPGM_PAE ? 0 : E_SYS_UNSUPPORTED;
}

/// free all addresses in the list and delete list itself ...
static void free_phys_list(dq_list ml) {
   u32t cnt = ml->count(), ii;

   for (ii=0; ii<cnt; ii++) {
      qserr rc = hlp_memfreephys(ml->value(ii));
      if (rc)
         log_printf("error %X releasing mem at %LX\n", rc, ml->value(ii));
   }
   DELETE(ml);
}

/** check for disk presence at the start of 4th Gb.
    Note that dskmem can be filled even when error returned, zero entry is the
    disk header then any successful allocations until first fail.

    @param [out]    pdh        returning disk header (on success, mapped)
    @param [in,out] dskmem     list for *allocated* disk ranges above 4Gb
    @retval 0                  success, returns mapped disk header (must be released in caller)
    @retval E_SYS_NOTFOUND     no disk
    @retval E_SYS_NOMEM        no memory above 4Gb
    @retval E_SYS_UNSUPPORTED  no PAE
    @retval E_SYS_BADFMT       disk present, but uses low pages
    @retval E_SYS_ARANGE       memory above 4Gb is used by someone */
static qserr vdisk_exists(HD4_Header **pdh, dq_list dskmem) {
   qserr  rc;
   dskmem->clear();
   if (pdh) *pdh = 0;
   if (!min4Gaddr) return E_SYS_NOMEM;
   // unable to set mode?
   rc = check_pae();
   if (!rc) {
      // note that allocation can round down the returned pointer
      u64t mptr = hlp_memalloc64(PAGESIZE, "vhdr", QSMA_LOCEXACT|QSMA_HIGH|
                                 QSMA_NOCLEAR, min4Gaddr, &rc);
      if (mptr) {
         HD4_Header *dh;
         dskmem->add(mptr);
         // map start of the memory above 4Gb
         dh = (HD4_Header*)pag_physmap(min4Gaddr, PAGESIZE, 0);
         if (dh) {
            if (dh->h4_sign==EXPDATA_SIGN && dh->h4_version==1) {
               HD4_TabEntry *tab = (HD4_TabEntry*)(dh+1);
               u32t           ii;
               int        lowmem = 0;

               for (ii=0; ii<dh->h4_tabsize; ii++)
                  if (tab[ii].hde_1stpage<PAGESIN4GB) { lowmem = 1; break; }
                  else {
                     mptr = hlp_memalloc64(tab[ii].hde_sectors*512, "vdsk",
                               QSMA_LOCEXACT|QSMA_HIGH|QSMA_NOCLEAR,
                               (u64t)tab[ii].hde_1stpage<<PAGESHIFT, &rc);
                     if (mptr) dskmem->add(mptr); else break;
                  }
               if (lowmem) {
                  log_printf("Disk found, but uses low memory!\n");
                  if (!rc) rc = E_SYS_BADFMT;
               }
               if (!rc && pdh) *pdh = dh;
            } else
               rc = E_SYS_NOTFOUND;
            // unmap if error or not used
            if (rc || !pdh) pag_physunmap(dh);
         }
hlp_memprint64(0);
      }
   }
   return rc;
}


qserr _exicc vdisk_init(EXI_DATA, u32t minsize, u32t maxsize, u32t flags,
                       u32t divpos, char letter1, char letter2, u32t *disk)
{
   qserr         rc;
   u32t          ii, 
         persist_ok = 0;
   u32t     lopages, hipages;
   // check below that only one of them is set
   u32t      fsbits = flags&(VFDF_FAT32|VFDF_HPFS|VFDF_JFS);
   // disk was created?
   if (cdh) return E_SYS_INITED;
   // invalid flags
   if ((flags&(VFDF_NOHIGH|VFDF_NOLOW))==(VFDF_NOHIGH|VFDF_NOLOW) ||
       (flags&(VFDF_EMPTY|VFDF_SPLIT))==(VFDF_EMPTY|VFDF_SPLIT) ||
       (flags&(VFDF_EMPTY|VFDF_GPT))==(VFDF_EMPTY|VFDF_GPT) ||
       (flags&(VFDF_FAT32|VFDF_NOFAT32))==(VFDF_FAT32|VFDF_NOFAT32) ||
       bsr32(fsbits)!=bsf32(fsbits) ||
       (flags&VFDF_EXACTSIZE)!=0 && !minsize ||
       (flags&VFDF_SPLIT)!=0 && !divpos ||
       (flags&VFDF_EMPTY)!=0 && divpos ||
       (flags&VFDF_PERCENT)!=0 && divpos>=100) return E_SYS_INVPARM;
   // allow NOFAT32 with HPFS but clears it to not confuse the code below
   if (flags&(VFDF_HPFS|VFDF_JFS)) flags&=~VFDF_NOFAT32;
   // disk memory now hidden only via key
   hidemem = flags&VFDF_HIDEMEM?1:0;
   /* lock the entire creation process (this thread cannot be killed until
      mt_sysunlock()) */
   mt_syslock();

   get_memory(&lopages, &hipages);

   if (flags&VFDF_NOHIGH) hipages = 0;
   if (flags&VFDF_NOLOW)  lopages = 0;
   // limit usage memory above 4Gb
   if (hipages && maxsize) {
      maxsize <<= 20 - PAGESHIFT;
      if (hipages>maxsize) hipages = maxsize;
   }
   minsize <<= 20 - PAGESHIFT;
   // adjust size before persistence check
   if ((flags&VFDF_EXACTSIZE) && hipages>minsize) hipages = minsize;

   if ((flags&VFDF_PERSIST) && hipages) {
      HD4_Header *edh;
      dq_list    dmem = NEW_G(dq_list);
      qserr      perr = vdisk_exists(&edh,dmem);
      // fill it back from the header, but only on size match!
      if (perr) {
         if (perr==E_SYS_NOTFOUND) log_printf("persist not found\n");
            else log_printf("persist err %X\n", perr);
      } else
      if (!minsize || edh->h4_pages+1>=minsize) {
         cdh        = edh;
         cdmem      = dmem;
         cdhphys    = min4Gaddr;
         cdhsize    = edh->h4_pages<<PAGESHIFT-9;
         cde        = (HD4_TabEntry*)(edh+1);
         cdecount   = edh->h4_tabsize;
         persist_ok = 1;
         rc = 0;
      } else
         log_printf("persist !match: %X %X %X\n", edh->h4_pages, hipages, minsize);
      // any error - free reserved memory ...
      if (!cdh) free_phys_list(dmem);
   }
   if (!cdh)
   do {
      rc = 0;
      // min size specified, check memory below 4Gb
      if (minsize) {
         if (hipages<minsize) {
            if (hipages+lopages<minsize) {
               log_printf("min %d pages, but avail: lo %d, high %d\n",
                  minsize, lopages, hipages);
               rc = E_SYS_NOMEM;
               break;
            }
            lopages = minsize - hipages;
         } else {
            lopages = 0;
            if (flags&VFDF_EXACTSIZE) hipages = minsize;
         }
      } else lopages = 0;

      if (hipages+lopages==0) {
         log_printf("no high memory?\n");
         rc = E_SYS_NOMEM;
         break;
      }
      // check split border
      if (divpos) {
         if ((flags&VFDF_PERCENT)==0) divpos<<=20-PAGESHIFT; else
            divpos = (u64t)(hipages+lopages) * divpos / 100;
         if (divpos>=hipages+lopages) divpos = 0;
         if (divpos) log_printf("split on %d Mb\n", divpos>>20-PAGESHIFT);
      }
      // physical memory list to free ...
      if (!cdmem) cdmem = NEW_G(dq_list);
      // turn on PAE paging - if high pages used
      if (hipages) {
         u32t    aflags = QSMA_LOCEXACT | QSMA_HIGH,
                  hicnt = 0,
                   lcnt = 100;
         u64t    wanted = min4Gaddr;
         rc = check_pae();
         if (rc) break;
         // zero mem?
         if ((flags&VFDF_ZEROMEM)==0) aflags|=QSMA_NOCLEAR;

         /* try to get block at the start of 4Gb, else just largest, but
            below 2Tb (address>>9 should fits in dword) */
         // and limit it to 100 loops since high memory mgr is a new code ;)
         while (hipages && lcnt--) {
            u64t blocksize, mptr;
            int      first = (aflags&QSMA_LOCEXACT)==QSMA_LOCEXACT;

            rc = hlp_memavail64(&blocksize, 0, 0, aflags, wanted);

            if (rc) {
               // first time we try min4Gaddr location, it *can* fail 
               if (!first) break;
            } else
            if (blocksize>=PAGESIZE*2) {
               u32t pages = (u32t)(blocksize>>PAGESHIFT);
               if (pages > hipages) pages = hipages;
            
               mptr = hlp_memalloc64((u64t)pages<<PAGESHIFT, "vdsk",
                                     aflags, wanted, &rc);
               if (mptr) {
                  hipages -= pages;
                  hicnt   += pages;
                  cdmem->add(mptr);
               }
            } else
            // largest block < 8k?
            if (!first) break;
            // after the first entry ...
            if (first) {
               aflags &= ~QSMA_LOCEXACT;
               aflags |= QSMA_LOCBELOW;
               wanted  = _2TBLL;
            }
         }
         // ignore all allocation errors ...
         rc = 0;
         /* some high pages still unprocessed? try to use low memory if
            disk size was forced */
         if (hipages && minsize) {
            log_printf("hipages rem: %d\n", hipages);
            lopages += hipages;
         }
         // walk over collected memory
         for (ii=0; ii<cdmem->count(); ii++) {
            u64t  addr = cdmem->value(ii);
            u32t pages = (u32t)(hlp_memgetsize64(addr)>>PAGESHIFT);

            // map header on first block
            if (!cdh) {
               cdh     = (HD4_Header*)pag_physmap(addr, PAGESIZE, 0);
               cdhphys = addr;
               if (!cdh) {
                  log_printf("Header map err (hi) - %LX\n", addr);
                  rc = E_SYS_SOFTFAULT;
                  break;
               }
               fill_header(--hicnt + lopages);
               addr += PAGESIZE;
               pages--;
            }
            // add block to list in header
            cde[cdecount].hde_1stpage = (u32t)(addr>>PAGESHIFT);
            cde[cdecount].hde_sectors = pages<<3;
            cdecount++;
         }
         if (rc) break;
      }
      // low memory (system alloc)
      mt_swlock();
      while (lopages) {
         u32t  pages, addr, res;
         hlp_memavail(&pages, 0, QSMA_PHYS);
         pages>>=PAGESHIFT;
      
         if (pages>lopages) pages = lopages;
         addr = (u32t)hlp_memallocsig(pages<<PAGESHIFT, "disk", QSMA_MAXADDR|
                                      QSMA_RETERR|QSMA_PHYS|
                                      (flags&VFDF_ZEROMEM?0:QSMA_NOCLEAR));
         if (!addr) break;
         // push to the list
         cdmem->add(addr);
         // mark memory as used by ram disk (hide it from OS/2)
         res = sys_markmem(addr, pages, PCMEM_QSINIT|PCMEM_HIDE);
         if (res)
            log_printf("warning! mark failed, addr %LX - %d\n", addr, res);
         // map header on first block
         if (!cdh) {
            // low addr must be mapped in any case (else we hang! :)
            cdh     = (HD4_Header*)pag_physmap(addr, PAGESIZE, 0);
            cdhphys = addr;
            if (!cdh) {
               log_printf("Header map err (lo) - %08X\n", (u32t)addr);
               rc = E_SYS_SOFTFAULT;
               break;
            }
            fill_header(--lopages);
            addr += PAGESIZE;
            pages--;
         }
         // add block to list in header
         cde[cdecount].hde_1stpage = (u32t)(addr>>PAGESHIFT);
         cde[cdecount].hde_sectors = pages<<3;
         lopages -= pages;
         cdecount++;
      }
      mt_swunlock();

      if (rc) break;
      // some low pages still unprocessed? strange...
      if (lopages) {
         log_printf("lopages rem: %d\n", lopages);
         rc = E_SYS_NOMEM;
         break;
      }
   } while (0);

   if (!rc) {
      s32t           drvnum;
      struct qs_diskinfo dp;
      memset(&dp, 0, sizeof(dp));
      dp.qd_sectors     = cdhsize;
      dp.qd_cyls        = cdh->h4_cyls;
      dp.qd_heads       = cdh->h4_heads;
      dp.qd_spt         = cdh->h4_spt;
      dp.qd_extread     = _read;
      dp.qd_extwrite    = _write;
      dp.qd_extremove   = _remove;
      dp.qd_extflush    = _flush;
      dp.qd_sectorsize  = 512;
      dp.qd_sectorshift = 9;
      // if class is ok - create instance for our`s disk
      if (classid_ext) {
         eiptr = (qs_extdisk)exi_createid(classid_ext, EXIF_SHARED);
         dp.qd_extptr = eiptr;
      }
      // install new "hdd"
      drvnum = hlp_diskadd(&dp, HDDA_BIOSNUM);
      if (drvnum<0) {
         log_printf("hlp_diskadd err!\n");
         rc = E_DSK_MOUNTERR;
      } else {
         u32t  d_rc = 0, l_rc = 0;
         chdd = drvnum;
         // # of disk slices
         cdh->h4_tabsize = cdecount;

         log_it(3, "vdisk ranges: %d, bios fake num %02X\n", cdecount,
            hlp_diskbios(chdd,1));
         for (ii=0; ii<cdecount; ii++)
            log_it(3, "%2d. %010LX - %d pages\n", ii+1,
               (u64t)cde[ii].hde_1stpage<<PAGESHIFT, cde[ii].hde_sectors>>3);

         if (!persist_ok) {
            str_list *vol_names = 0,
                    *vol_labels = 0;
            // allows custom LVM names in form {RAM/V}DISKNAME = pt1 name, pt2 name
            char       *vdnames = getenv("RAMDISKNAME"),
                      *vdlabels = getenv("RAMDISKLABEL");
            if (!vdnames) vdnames = getenv("VDISKNAME");
            if (!vdlabels) vdlabels = getenv("VDISKLABEL");
            // custom LVM volume names
            if (vdnames)  vol_names  = str_split(vdnames, ",");
            if (vdlabels) vol_labels = str_split(vdlabels, ",");

            // wipe mbr sector
            dsk_newmbr(chdd, DSKBR_CLEARALL);
            // wipe LVM sector (to prevent of using previous disk data after "ramdisk delete")
            dsk_emptysector(chdd, cdh->h4_spt-1, 1);

            // init mbr/gpt & first lvm dlat (on mbr)
            d_rc = flags & VFDF_GPT ? dsk_gptinit(chdd) : dsk_ptinit(chdd, 1);
            // create partition(s) and format it
            if (!d_rc) {
               // set disk LVM name
               if ((flags & VFDF_GPT)==0)
                  l_rc = lvm_setname(chdd, 0, LVMN_DISK, "PAE_RAM_DISK");

               if ((flags&VFDF_EMPTY)==0) {
                  u64t  start, size;
                  // div pos in megabytes
                  divpos >>= 20-PAGESHIFT;
                  // create a single partition
                  d_rc = dsk_ptalign(chdd, 0, divpos?divpos:100, DPAL_CHSSTART|
                     DPAL_CHSEND|(divpos?0:DPAL_PERCENT), &start, &size);
                  if (flags & VFDF_GPT) {
                     if (!d_rc) d_rc = dsk_gptcreate(chdd, start, size, DFBA_PRIMARY,
                        CGUID_OS2DATA);
                     if (!d_rc) d_rc = dsk_gptactive(chdd, 0);
                  } else {
                     if (!d_rc) d_rc = dsk_ptcreate(chdd, start, size, DFBA_PRIMARY, 12);
                     // and make it active
                     if (!d_rc) d_rc = dsk_setactive(chdd, 0);
                  }
                  // format partition & assign LVM drive letter
                  if (!d_rc) d_rc = make_partition(0, flags, letter1, &l_rc, vol_labels);
                  // set custom (non-empty) volume name
                  if (vol_names)
                     if (flags & VFDF_GPT) set_gptname(0, vol_names, &d_rc);
                        else set_lvmname(0, vol_names, &l_rc);
                  /* second partition processing */
                  if (divpos) {
                     d_rc = dsk_ptalign(chdd, 0, 100, DPAL_CHSSTART|DPAL_CHSEND|DPAL_PERCENT,
                        &start, &size);
                     if (!d_rc)
                        if (flags & VFDF_GPT)
                           d_rc = dsk_gptcreate(chdd, start, size, DFBA_PRIMARY, CGUID_OS2DATA);
                        else
                           d_rc = dsk_ptcreate(chdd, start, size, DFBA_PRIMARY, 12);
                     // format partition & assign LVM drive letter
                     if (!d_rc) d_rc = make_partition(1, flags, letter2, &l_rc, vol_labels);
                     // set custom (non-empty) volume name
                     if (vol_names)
                        if (flags & VFDF_GPT) set_gptname(1, vol_names, &d_rc);
                           else set_lvmname(1, vol_names, &l_rc);
                  }
               }
            }
            log_printf("ram disk init: dmgr rc = %X, lvm rc = %X\n", d_rc, l_rc);
            if (vol_labels) free(vol_labels);
            if (vol_names) free(vol_names);
         } else {
            log_printf("ram disk init: persistent!\n");
            /* mount partition(s) to 1st available drive letter, but
               ignore result. This should help for ones, who assume it as C:
               in any case */
            if ((flags&VFDF_EMPTY)==0) {
               mount_persistent(0);
               if (divpos) mount_persistent(1);
            }
         }

         if (disk) *disk = chdd;
         /* save header page number to storage key for "bootos2" to prevent
            static linking of this DLL to it (vdisk.dll uses partmgr, i.e.
            in case of static linking we will load both VDISK & PARTMGR on
            every boot) */
         {
            char dname[16];
            sto_savedword(STOKEY_VDPAGE, cdhphys>>PAGESHIFT);

            dsk_disktostr(chdd, dname);
            sto_save(STOKEY_VDNAME, dname, strlen(dname)+1, 1);
         }
      }
   }
   if (rc) vdisk_free(0,0); else
      sys_notifyevent(SECB_QSEXIT|SECB_GLOBAL, qs_shutdown);

   mt_sysunlock();

   return persist_ok && !rc ? E_SYS_EXIST : rc;
}

u32t _exicc vdisk_clean(EXI_DATA) {
   if (chdd) return E_SYS_INITED; else {
      qserr       rc;
      dq_list   dmem;
      // deny thread kill here ...
      mt_syslock();
      // just makes list at least once, to setup min4Gsize
      if (!min4Gsize) get_memory(0,0);

      dmem = NEW(dq_list);
      rc   = vdisk_exists(0,dmem);
      // 1st tab entry must be above 4Gb ...
      if (!rc || rc==E_SYS_BADFMT || rc==E_SYS_ARANGE) {
         u32t ii, cnt = dmem->count();
         // at least we heave something to do!
         for (ii=0; ii<cnt; ii++) {
            u32t   clean_size = PAGE_SIZE;
            u64t     blk_addr = dmem->value(ii),
                     blk_size = hlp_memgetsize64(blk_addr);
            void   *clean_ptr;

            if (blk_size<clean_size) clean_size = (u32t)blk_size;
         
            clean_size &= ~PAGEMASK;
            clean_ptr   = pag_physmap(blk_addr, clean_size, 0);
            if (clean_ptr) {
               memset(clean_ptr, 0, clean_size);
               pag_physunmap(clean_ptr);
               log_printf("ram disk clean(%LX,%X)!\n", blk_addr, clean_size);
               /* clean header AND the 1st range then return "no error",
                  if first range is not mapped - return E_SYS_ARANGE
                  from vdisk_exists() to the user */
               if (ii==1) { rc = 0; break; }
            } else {
               rc = E_SYS_SOFTFAULT;
               break;
            }
         }
      }
      free_phys_list(dmem);
      mt_sysunlock();

      return rc;
   }
}


u32t _exicc vdisk_free(EXI_DATA) {
   u32t     rc = 1;
   // deny this thread killing for a while ...
   mt_syslock();
   // remove installed "HDD"
   if (chdd) {
      // unmount all volumes
      hlp_unmountall(chdd);
      /* wipe MBR and DLAT sectors else DMGR code can catch them in the next
         disk with other size */
      dsk_newmbr(chdd,DSKBR_CLEARALL|DSKBR_LVMINFO);
      // remove disk
      if (!hlp_diskremove(chdd)) rc = 0;
      chdd = 0;
      // delete disk name from storage
      sto_del(STOKEY_VDNAME);
   }
   // remove callback
   sys_notifyevent(0, qs_shutdown);
   // delete lost external info class instance
   if (eiptr) { DELETE(eiptr); eiptr = 0; }

   // unmark hidden memory, but only it present in the list of owned blocks 
   if (cdmem) {
      pcmem_entry *me = sys_getpcmem(0);
      u32t         ii = 0;
      // this will also unmark "mem hide" results if blocks was merged ...
      while (me[ii].pages) {
         if (me[ii].flags==(PCMEM_QSINIT|PCMEM_HIDE))
            if (cdmem->indexof(me[ii].start,0)>=0)
               sys_markmem(me[ii].start, me[ii].pages, PCMEM_QSINIT);
         ii++;
      }
      free(me);
   }
   // free all owned physical memory
   if (cdmem) { free_phys_list(cdmem); cdmem = 0; }
   // delete storage key with disk header phys page number
   sto_del(STOKEY_VDPAGE);
   // critical part finished ...
   mt_sysunlock();
   // there is no disk header, return 0
   if (!cdh) return 0;
   // unmap high memory
   if (maparea) pag_physunmap(maparea);
   maparea  = 0;
   mapstart = 0;
   // unmap header
   pag_physunmap(cdh);
   cdh      = 0;

   cdhphys  = 0;
   cdhsize  = 0;
   return rc;
}

qserr _exicc vdisk_query(EXI_DATA, disk_geo_data *geo, u32t *disk,
                         u32t *sectors, u32t *seclo, u32t *physpage)
{
   qserr rc = cdh?0:E_SYS_NONINITOBJ;
   if (!rc) {
      if (disk)     *disk     = chdd;
      if (sectors)  *sectors  = cdhsize;
      if (physpage) *physpage = cdhphys>>PAGESHIFT;
      if (geo) {
         geo->Cylinders    = cdh->h4_cyls;
         geo->Heads        = cdh->h4_heads;
         geo->SectOnTrack  = cdh->h4_spt;
         geo->SectorSize   = 512;
         geo->TotalSectors = cdhsize;
      }
      if (seclo) {
         u32t ii;
         for (ii=0, *seclo=0; ii<cdecount; ii++)
            if (cde[ii].hde_1stpage < _4GBLL>>PAGESHIFT) *seclo+=cde[ii].hde_sectors;
      }
   } else {
      if (geo) memset(geo, 0, sizeof(disk_geo_data));
      if (sectors) *sectors  = 0;
      if (seclo) *seclo = 0;
   }
   return rc;
}

static qserr _exicc vdisk_setgeo(EXI_DATA, disk_geo_data *geo) {
   qserr rc = 0;
   if (!geo) rc = E_SYS_ZEROPTR; else
   if (!cdh) rc = E_SYS_NONINITOBJ; else
   if (geo->TotalSectors!=cdhsize || geo->SectorSize!=512) rc = E_SYS_INVPARM; else
   // check CHS (at least minimal)
   if (!geo->Heads || geo->Heads>255 || !geo->SectOnTrack ||
      geo->SectOnTrack>255 || !geo->Cylinders || (u64t)geo->Cylinders *
         geo->Heads * geo->SectOnTrack > geo->TotalSectors) rc = E_SYS_INVPARM;
   else {
      // copy CHS
      cdh->h4_cyls  = geo->Cylinders;
      cdh->h4_heads = geo->Heads;
      cdh->h4_spt   = geo->SectOnTrack;
      // if disk is mounted - update CHS value in system info
      if (chdd) {
         struct qs_diskinfo *qdi = hlp_diskstruct(chdd, 0);
         // this is WRONG! we patching it directly in the system data
         if (qdi) {
            qdi->qd_cyls  = geo->Cylinders;
            qdi->qd_heads = geo->Heads;
            qdi->qd_spt   = geo->SectOnTrack;
         }
      }
   }
   return rc;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// external info class
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

typedef struct {
   qs_vdisk       inst;
} doptdata;

static void _std dopt_init(void *instance, void *data) {
   doptdata *dp = (doptdata*)data;
   // use global class mutex as an easieast way of sync
   dp->inst = exi_createid(classid_vd, EXIF_SHARED);
}

static void _std dopt_done(void *instance, void *data) {
   doptdata *dp = (doptdata*)data;
   exi_free(dp->inst);
   memset(dp, 0, sizeof(doptdata));
}

static qserr _exicc dopt_getgeo(EXI_DATA, disk_geo_data *geo) {
   doptdata *dp = (doptdata*)data;
   qserr rc = 0;
   if (!geo) rc = E_SYS_ZEROPTR; else
      if (!cdh) rc = E_SYS_NONINITOBJ; else
         rc = dp->inst->query(geo, 0, 0, 0, 0);
   return rc;
}

static qserr _exicc dopt_setgeo(EXI_DATA, disk_geo_data *geo) {
   doptdata *dp = (doptdata*)data;
   return dp->inst->setgeo(geo);
}

static char* _exicc dopt_getname(EXI_DATA, int verbose) {
   u32t sectors, seclo;
   doptdata *dp = (doptdata*)data;
   qserr     rc = dp->inst->query(0, 0, &sectors, &seclo, 0);
   if (rc) return 0;
   if (verbose)
      return sprintf_dyn("PAE ram disk. %u low, %u high pages", seclo>>3,
         sectors-seclo>>3);
   return strdup("PAE ram disk");
}

static u32t _exicc dopt_state(EXI_DATA, u32t state) {
   //doptdata *dp = (doptdata*)data;
   // set r/o is not supported, cache denied for this disk
   return EDSTATE_NOCACHE|EDSTATE_RAM;
}

static void *qs_vdisk_fn[] = { vdisk_init, vdisk_load, vdisk_free, vdisk_query, vdisk_setgeo, vdisk_clean };

static void *qs_extdisk_fn[] = { dopt_getgeo, dopt_setgeo, dopt_getname, dopt_state };

int unregister_class(void) {
   int err = 0;
   if (classid_vd)
      if (exi_unregister(classid_vd)) classid_vd = 0; else err++;
   if (classid_ext)
      if (exi_unregister(classid_ext)) classid_ext = 0; else err++;
   return err?0:1;
}

int register_class(void) {
   if (!classid_ext) classid_ext = exi_register("qs_vdisk_ext", qs_extdisk_fn,
      sizeof(qs_extdisk_fn)/sizeof(void*), sizeof(doptdata), 0, dopt_init, dopt_done, 0);

   if (!classid_vd) classid_vd = exi_register("qs_vdisk", qs_vdisk_fn,
      sizeof(qs_vdisk_fn)/sizeof(void*), 0, EXIC_GMUTEX, 0, 0, 0);

   if (!classid_ext || !classid_vd) { unregister_class(); return 0; }
   return 1;
}
