// some rare device paths are still missing ...
#include "qsloader.h"
#include "efilib.h"

#define ALIGN_SIZE(a)   ((a % MIN_ALIGNMENT_SIZE) ? MIN_ALIGNMENT_SIZE - (a % MIN_ALIGNMENT_SIZE) : 0)

typedef struct {
   char       *str;
   u32t        len;
   u32t     maxlen;
} POOL_PRINT;

extern EFI_GUID   UnknownDevice,
              EfiGlobalVariable;

char *guidtostr(EFI_GUID *Guid) {
   static char gstr[48];    // !!
   gstr[0] = 0;
   if (Guid) GuidToStr(gstr, Guid);
   return gstr;
}

int _snprint(char *buf, u32t count, const char *fmt, ptrdiff_t *argp);

void CatPrint(POOL_PRINT *Str, char *fmt, ...) {
   static char sbuf[1024];
   int len = _snprint(sbuf, sizeof(sbuf), fmt, (ptrdiff_t*)&fmt + 1);
   // no realloc!
   if (Str->maxlen - Str->len < len + 1) {
      u32t avail = Str->maxlen - Str->len;
      if (avail>1) {
         RtCopyMem(Str->str + Str->len, sbuf, avail-1);
         Str->len += avail-1;
      }
      return;
//      Str->str     = (char*)realloc(Str->str, Str->maxlen + len*2 + 1);
//      Str->maxlen += len*2;
   }
   RtCopyMem(Str->str + Str->len, sbuf, len);
   Str->len += len;
}

static void _DevPathPci(POOL_PRINT *Str, void *DevPath) {
   PCI_DEVICE_PATH  *Pci = DevPath;
   CatPrint(Str, "Pci(%X,%X)", Pci->Device, Pci->Function);
}

static void _DevPathPccard(POOL_PRINT *Str, void *DevPath) {
   PCCARD_DEVICE_PATH  *Pccard = DevPath;   
   CatPrint(Str, "Pccard(Func %X)", Pccard->FunctionNumber);   
}

static void _DevPathMemMap(POOL_PRINT *Str, void *DevPath) {
   MEMMAP_DEVICE_PATH  *MemMap = DevPath;   
   CatPrint(Str, "MemMap(%d:%LX-%LX)", MemMap->MemoryType,
      MemMap->StartingAddress, MemMap->EndingAddress);
}

static void _DevPathController(POOL_PRINT *Str, void *DevPath) {
   CONTROLLER_DEVICE_PATH *Controller = DevPath;
   CatPrint(Str, "Ctrl(%d)", Controller->Controller);
}

static void _DevPathVendor(POOL_PRINT *Str, void *DevPath) {
   VENDOR_DEVICE_PATH                  *Vendor;
   char                                  *Type;
   UNKNOWN_DEVICE_VENDOR_DEVICE_PATH   *UnknownDevPath;

   Vendor = DevPath;
   switch (DevicePathType(&Vendor->Header)) {
      case HARDWARE_DEVICE_PATH:  Type = "Hw";    break;
      case MESSAGING_DEVICE_PATH: Type = "Msg";   break;
      case MEDIA_DEVICE_PATH:     Type = "Media"; break;
      default:                    Type = "?";     break;
   }                            

   CatPrint(Str, "Ven%s(%s", Type, guidtostr(&Vendor->Guid));
   if (RtCompareMem(&Vendor->Guid, &UnknownDevice, 16) == 0) {
      // GUID used by EFI to enumerate an EDD 1.1 device
      UnknownDevPath = (UNKNOWN_DEVICE_VENDOR_DEVICE_PATH *)Vendor;
      CatPrint(Str, ":%02X)", UnknownDevPath->LegacyDriveLetter);
   } else {
      CatPrint(Str, ")");
   }
}

static void _DevPathAcpi(POOL_PRINT *Str, void *DevPath) {
   ACPI_HID_DEVICE_PATH *Acpi = DevPath;

   if ((Acpi->HID & PNP_EISA_ID_MASK) == PNP_EISA_ID_CONST) {
      CatPrint(Str, "Acpi(PNP%04X,%X)", EISA_ID_TO_NUM (Acpi->HID), Acpi->UID);
   } else {
      CatPrint(Str, "Acpi(%08X,%X)", Acpi->HID, Acpi->UID);
   }
}

static void _DevPathAtapi(POOL_PRINT *Str, void *DevPath) {
   ATAPI_DEVICE_PATH  *Atapi = DevPath;

   CatPrint(Str, "Ata(%s,%s)", 
      Atapi->PrimarySecondary ? "Secondary" : "Primary",
      Atapi->SlaveMaster ? "Slave" : "Master");
}

static void _DevPathScsi(POOL_PRINT *Str, void *DevPath) {
   SCSI_DEVICE_PATH   *Scsi = DevPath;
   CatPrint(Str, "Scsi(Pun%X,Lun%X)", Scsi->Pun, Scsi->Lun);
}

static void _DevPathFibre(POOL_PRINT *Str, void *DevPath) {
   FIBRECHANNEL_DEVICE_PATH  *Fibre = DevPath;
   CatPrint(Str, "Fibre(WWN%LX,Lun%LX)", Fibre->WWN, Fibre->Lun);    
}

static void _DevPath1394(POOL_PRINT *Str, void *DevPath) {
   F1394_DEVICE_PATH  *F1394 = DevPath;
   // Guid has format of IEEE-EUI64 ??
   CatPrint(Str, "1394(%016LX)", &F1394->Guid);
}

static void _DevPathUsb(POOL_PRINT *Str, void *DevPath) {
   USB_DEVICE_PATH  *Usb = DevPath;
   CatPrint(Str, "Usb(%X,%X)", Usb->ParentPortNumber, Usb->InterfaceNumber); 
}

static void _DevPathFvFile(POOL_PRINT *Str, void *DevPath) {
   MEDIA_FW_VOL_FILEPATH_DEVICE_PATH  *FvFile = DevPath;
   CatPrint(Str, "FvFile(%s)", guidtostr(&FvFile->NameGuid));
}

static void _DevPathFv(POOL_PRINT *Str, void *DevPath) {
   MEDIA_FW_VOL_DEVICE_PATH  *Fv = DevPath;
   CatPrint(Str, "Fv(%s)", guidtostr(&Fv->NameGuid));
}

static void _DevPathUsbClass(POOL_PRINT *Str, void *DevPath) {
   USB_CLASS_DEVICE_PATH  *UsbClass = DevPath;

   CatPrint(Str, "Usb Class(%X,%X,%X,%X,%X)", 
      UsbClass->VendorId, UsbClass->ProductId, UsbClass->DeviceClass,
         UsbClass->DeviceSubClass, UsbClass->DeviceProtocol);
}

static void _DevPathI2O(POOL_PRINT *Str, void *DevPath) {
   I2O_DEVICE_PATH  *I2O = DevPath;
   CatPrint(Str, "I2O(%X)", I2O->Tid);
}

static void _DevPathMacAddr(POOL_PRINT *Str, void *DevPath) {
   MAC_ADDR_DEVICE_PATH    *MAC;
   UINTN                   HwAddressSize;
   UINTN                   Index;

   MAC = DevPath;

   HwAddressSize = sizeof(EFI_MAC_ADDRESS);
   if (MAC->IfType == 0x01 || MAC->IfType == 0x00) HwAddressSize = 6;
   CatPrint(Str, "Mac(");

   for(Index = 0; Index < HwAddressSize; Index++) {
       CatPrint(Str, "%02X",MAC->MacAddress.Addr[Index]);
   }
   CatPrint(Str, ")");
}

static void _DevPathIPv4(POOL_PRINT *Str, void *DevPath) {
   IPv4_DEVICE_PATH *IP = DevPath;
   CatPrint(Str, "IPv4(%d.%d.%d.%d:%d)",IP->RemoteIpAddress.Addr[0],
      IP->RemoteIpAddress.Addr[1], IP->RemoteIpAddress.Addr[2],
         IP->RemoteIpAddress.Addr[3], IP->RemotePort);
}

static void _DevPathIPv6(POOL_PRINT *Str, void *DevPath) {
   IPv6_DEVICE_PATH  *IP = DevPath;
   CatPrint(Str, "IP-v6");
}

static void _DevPathiSCSI(POOL_PRINT *Str, void *DevPath) {
   ISCSI_DEVICE_PATH_WITH_NAME *ISCSIDevPath = DevPath;
   UINT16                      Options;

   CatPrint(Str, "iSCSI(%s,%X,%LX,", ISCSIDevPath->TargetName,
      ISCSIDevPath->TargetPortalGroupTag, ISCSIDevPath->Lun);

   Options = ISCSIDevPath->LoginOption;
   CatPrint(Str, "%s,", (Options>>1 & 1) != 0 ? "CRC32C" : "None");
   CatPrint(Str, "%s,", (Options>>3 & 1) != 0 ? "CRC32C" : "None");
   if (((Options >> 11) & 0x0001) != 0) CatPrint(Str, "%s,", "None"); else
      if (((Options >> 12) & 0x0001) != 0) CatPrint(Str, "%s,", "CHAP_UNI");
         else CatPrint(Str, "%s,", "CHAP_BI");
   CatPrint(Str, "%s)", ISCSIDevPath->NetworkProtocol==0 ? "TCP" : "reserved");
}

static void _DevPathSata(POOL_PRINT *Str, void *DevPath) {
   SATA_DEVICE_PATH *Sata = DevPath;
   CatPrint(Str, "Sata(%X,%X,%X)", Sata->HBAPortNumber,
      Sata->PortMultiplierPortNumber, Sata->Lun);
}

static void _DevPathInfiniBand (POOL_PRINT *Str, void *DevPath) {
   INFINIBAND_DEVICE_PATH  *InfiniBand = DevPath;
   CatPrint(Str, "InfiniBand");
}

static void _DevPathUart (OUT POOL_PRINT *Str, void *DevPath) {
   UART_DEVICE_PATH  *Uart = DevPath;
   char             Parity;

   switch (Uart->Parity) {
      case 0  : Parity = 'D'; break;
      case 1  : Parity = 'N'; break;
      case 2  : Parity = 'E'; break;
      case 3  : Parity = 'O'; break;
      case 4  : Parity = 'M'; break;
      case 5  : Parity = 'S'; break;
      default : Parity = 'x'; break;
   }
   if (Uart->BaudRate == 0) CatPrint(Str, "Uart(DEFAULT %c",Parity);	
      else CatPrint(Str, "Uart(%d %c",Uart->BaudRate,Parity);
   if (Uart->DataBits == 0) CatPrint(Str, "D");
      else CatPrint(Str, "%d",Uart->DataBits);

   switch (Uart->StopBits) {
      case 0  : CatPrint(Str, "D)");   break;
      case 1  : CatPrint(Str, "1)");   break;
      case 2  : CatPrint(Str, "1.5)"); break;
      case 3  : CatPrint(Str, "2)");   break;
      default : CatPrint(Str, "x)");   break;
   }
}

static void _DevPathHardDrive(OUT POOL_PRINT *Str, void *DevPath) {
   HARDDRIVE_DEVICE_PATH   *Hd;

   Hd = DevPath;
   switch (Hd->SignatureType) {
      case SIGNATURE_TYPE_MBR:
         CatPrint(Str, "HD(%d,MBR,%08X,", Hd->PartitionNumber,
             *((UINT32 *)(&(Hd->Signature[0]))));
         break;
      case SIGNATURE_TYPE_GUID:
         CatPrint(Str, "HD(%d,GPT,%s,", Hd->PartitionNumber,
             guidtostr((EFI_GUID*)&(Hd->Signature[0])));
         break;
      default:
         CatPrint(Str, "HD(%d,%d,0,", Hd->PartitionNumber, Hd->SignatureType);
         break;
   }
   CatPrint(Str, "%LX,%LX)", Hd->PartitionStart, Hd->PartitionSize);
}

static void _DevPathCDROM(OUT POOL_PRINT *Str, void *DevPath) {
   CDROM_DEVICE_PATH  *Cd = DevPath;
   CatPrint(Str, "CDROM(%X,%LX,%LX)", Cd->BootEntry, Cd->PartitionStart,
      Cd->PartitionSize);
}

static void _DevPathFilePath(POOL_PRINT *Str, void *DevPath) {
   FILEPATH_DEVICE_PATH  *Fp = DevPath;
   CatPrint(Str, "%S", Fp->PathName);
}

static void _DevPathMediaProtocol(OUT POOL_PRINT *Str, void *DevPath) {
   MEDIA_PROTOCOL_DEVICE_PATH  *MediaProt = DevPath;
   CatPrint(Str, "%s", guidtostr(&MediaProt->Protocol));
}

static void _DevPathBssBss(OUT POOL_PRINT *Str, void *DevPath) {
   BBS_BBS_DEVICE_PATH  *Bss = DevPath;
   char                *Type;

   switch (Bss->DeviceType) {
      case BBS_TYPE_FLOPPY:           Type = "Floppy";    break;
      case BBS_TYPE_HARDDRIVE:        Type = "Harddrive"; break;
      case BBS_TYPE_CDROM:            Type = "CDROM";     break;
      case BBS_TYPE_PCMCIA:           Type = "PCMCIA";    break;
      case BBS_TYPE_USB:              Type = "Usb";       break;
      case BBS_TYPE_EMBEDDED_NETWORK: Type = "Net";       break;
      default:                        Type = "?";         break;
   }

   CatPrint(Str, "Bss-%s(%s)", Type, Bss->String);
}

static void _DevPathEndInstance(OUT POOL_PRINT *Str, void *DevPath) {
   CatPrint(Str, ",");
}

static void _DevPathNodeUnknown(OUT POOL_PRINT *Str, void *DevPath) {
   CatPrint(Str, "?");
}

typedef void (*OutFunc)(POOL_PRINT *, void *);

struct {
   u8t          Type;
   u8t       SubType;
   OutFunc  Function;
} DevPathTable[] = {
   HARDWARE_DEVICE_PATH,  HW_PCI_DP,                        _DevPathPci,
   HARDWARE_DEVICE_PATH,  HW_PCCARD_DP,                     _DevPathPccard,
   HARDWARE_DEVICE_PATH,  HW_MEMMAP_DP,                     _DevPathMemMap,
   HARDWARE_DEVICE_PATH,  HW_VENDOR_DP,                     _DevPathVendor,
   HARDWARE_DEVICE_PATH,  HW_CONTROLLER_DP,                 _DevPathController,
   ACPI_DEVICE_PATH,      ACPI_DP,                          _DevPathAcpi,
   MESSAGING_DEVICE_PATH, MSG_ATAPI_DP,                     _DevPathAtapi,
   MESSAGING_DEVICE_PATH, MSG_SCSI_DP,                      _DevPathScsi,
   MESSAGING_DEVICE_PATH, MSG_FIBRECHANNEL_DP,              _DevPathFibre,
   MESSAGING_DEVICE_PATH, MSG_1394_DP,                      _DevPath1394,
   MESSAGING_DEVICE_PATH, MSG_USB_DP,                       _DevPathUsb,
   MESSAGING_DEVICE_PATH, MSG_USB_CLASS_DP,                 _DevPathUsbClass,    
   MESSAGING_DEVICE_PATH, MSG_SATA_DP,                      _DevPathSata,
   MESSAGING_DEVICE_PATH, MSG_I2O_DP,                       _DevPathI2O,
   MESSAGING_DEVICE_PATH, MSG_MAC_ADDR_DP,                  _DevPathMacAddr,
   MESSAGING_DEVICE_PATH, MSG_IPv4_DP,                      _DevPathIPv4,
   MESSAGING_DEVICE_PATH, MSG_IPv6_DP,                      _DevPathIPv6,
   MESSAGING_DEVICE_PATH, MSG_INFINIBAND_DP,                _DevPathInfiniBand,
   MESSAGING_DEVICE_PATH, MSG_UART_DP,                      _DevPathUart,
   MESSAGING_DEVICE_PATH, MSG_VENDOR_DP,                    _DevPathVendor,
   MESSAGING_DEVICE_PATH, MSG_ISCSI_DP,                     _DevPathiSCSI,
   MEDIA_DEVICE_PATH,     MEDIA_HARDDRIVE_DP,               _DevPathHardDrive,
   MEDIA_DEVICE_PATH,     MEDIA_CDROM_DP,                   _DevPathCDROM,
   MEDIA_DEVICE_PATH,     MEDIA_VENDOR_DP,                  _DevPathVendor,
   MEDIA_DEVICE_PATH,     MEDIA_FILEPATH_DP,                _DevPathFilePath,
   MEDIA_DEVICE_PATH,     MEDIA_PROTOCOL_DP,                _DevPathMediaProtocol,
   MEDIA_DEVICE_PATH,     MEDIA_FV_DP,                      _DevPathFv,
   MEDIA_DEVICE_PATH,     MEDIA_FV_FILEPATH_DP,             _DevPathFvFile,
   BBS_DEVICE_PATH,       BBS_BBS_DP,                       _DevPathBssBss,
   END_DEVICE_PATH_TYPE,  END_INSTANCE_DEVICE_PATH_SUBTYPE, _DevPathEndInstance,
   0,                     0,                          NULL
};

int DevPathToStr(char *buf, u32t buflen, EFI_DEVICE_PATH_PROTOCOL *DevPath) {
   EFI_DEVICE_PATH_PROTOCOL  *DevPathNode;
   POOL_PRINT            Str;
   OutFunc          DumpNode;
   UINTN               Index, NewSize;

   if (!DevPath || !buf || !buflen) return 0;

   Str.str    = buf;
   Str.len    = 0;
   Str.maxlen = buflen;
   // process each device path node
   DevPathNode = DevPath;
   while (!IsDevicePathEnd(DevPathNode)) {
      // find the handler to dump this device path node
      DumpNode = NULL;
      for (Index = 0; DevPathTable[Index].Function; Index += 1) {
         if (DevicePathType(DevPathNode) == DevPathTable[Index].Type &&
            DevicePathSubType(DevPathNode) == DevPathTable[Index].SubType) {
            DumpNode = DevPathTable[Index].Function;
            break;
         }
      }
      // if not found, use a generic function
      if (!DumpNode) DumpNode = _DevPathNodeUnknown;
      // put a path seperator in if needed
      if (Str.len  &&  DumpNode != _DevPathEndInstance) {
          CatPrint (&Str, "/");
      }
      // print this node of the device path
      DumpNode (&Str, DevPathNode);
      // next device path node
      DevPathNode = NextDevicePathNode(DevPathNode);
   }
   buf[Str.len] = 0;
   return 1;
}
