//
// QSINIT EFI loader
// disk & boot partition i/o
//
#include "qsloader.h"
#include "efilib.h"
#include "dskinfo.h"

#define BIOS_MAX_CYLINDERS         1024
#define BIOS_MAX_NUMHEADS           255
#define BIOS_MAX_SECTORSPERTRACK     63

extern u8t             *mfsd_openname;
//extern EFI_DEVICE_PATH    *DevicePath;

char                          DevPath[1024];

static EFI_BLOCK_IO        **BlockDev = 0;
static EFI_HANDLE          **BlockHdl = 0;
static u32t             BlockDevCount;

static BOOLEAN get_partition(EFI_HANDLE *dev, u32t *bootdisk, u64t *ptstart, u64t *ptlen) {
   EFI_DEVICE_PATH *dp = DevicePathFromHandle(dev);
   *bootdisk = FFFF;
   if (dp) {
      EFI_DEVICE_PATH *cp = dp;
      u32t          cdrom = 0, ii;
      while (1) {
         UINTN  type = DevicePathType(cp);
         // log_it(3, "type %d subtype %d\n", type, cp->SubType);

         if (type==MESSAGING_DEVICE_PATH && cp->SubType==MSG_ATAPI_DP) cdrom = 1;
            else
         if (type==MEDIA_DEVICE_PATH && (cp->SubType==MEDIA_HARDDRIVE_DP ||
             cp->SubType==MEDIA_CDROM_DP))
         {
            HARDDRIVE_DEVICE_PATH *pte = (HARDDRIVE_DEVICE_PATH*)cp;
            CDROM_DEVICE_PATH     *cde = (CDROM_DEVICE_PATH*)cp;
            log_it(3, "pt start %LX end %LX\n", pte->PartitionStart, pte->PartitionSize-1);

            for (ii=0; ii<BlockDevCount; ii++) {
               EFI_DEVICE_PATH *pdp = DevicePathFromHandle(BlockHdl[ii]);
               if (pdp)
                  if (RtCompareMem(pdp, dp, (u8t*)pte-(u8t*)dp)==0) {
                     // log_it(3, "block dev %u!\n", ii);
                     *bootdisk = (u32t)ii;
                     *ptstart  = cp->SubType==MEDIA_CDROM_DP?cde->PartitionStart:
                                 pte->PartitionStart;
                     *ptlen    = cp->SubType==MEDIA_CDROM_DP?cde->PartitionSize:
                                 pte->PartitionSize;
                     return 1;
                  }
            }
            log_it(3, "boot block dev not found!\n");
            return 0;
         }
         if (IsDevicePathEnd(cp)) break;
      
         cp = NextDevicePathNode(cp);
      }
      // try another way
      if (cdrom) {
         EFI_BLOCK_IO *BlkIo = 0;
         EFI_STATUS       st = BS->HandleProtocol(dev, &BlockIoProtocol, (VOID*)&BlkIo);
         if (!EFI_ERROR(st) && BlkIo) {
            EFI_BLOCK_IO_MEDIA  *Media = BlkIo->Media;

            for (ii=0; ii<BlockDevCount; ii++)
               if (BlkIo==BlockDev[ii]) {
                  *bootdisk = (u32t)ii;
                  *ptstart  = 0;
                  *ptlen    = Media->LastBlock+1;
                  return 1;
               }
         }
      }
   }
   return 0;
}

u32t dsk_init(exp32_table *etab, struct qs_diskinfo *qdd) {
   EFI_STATUS         st;
   UINTN          hcount;
   EFI_HANDLE   *handles;
   EFI_BLOCK_IO   *BlkIo;
   u32t        ii, dplen;
   char           *apath,
                *LdrPath;
   // init self path here
   DevPathToStr(DevPath, sizeof(DevPath), LoadedImage->FilePath);
   for (ii = (u32t)strlen(DevPath); ii>0 && DevPath[ii]!='/' && DevPath[ii]!='\\'; ii--);
   DevPath[ii] = 0;
   LdrPath = DevPath + ii + 1;
   if (ii && (DevPath[ii-1] == '\\' || DevPath[ii-1] == '/')) DevPath[ii-1] = 0;

   // empty by default
   etab->qd_fdds      = 0;
   etab->qd_hdds      = 0;

   st = LibLocateHandle(ByProtocol, &BlockIoProtocol, 0, &hcount, &handles);
   if (EFI_ERROR(st) || !hcount) return 0;

   dplen    = (u32t)strlen(DevPath) + 1;
   BlockDev = (EFI_BLOCK_IO**)AllocateZeroPool(sizeof(EFI_BLOCK_IO*) * hcount);
   BlockHdl = (EFI_HANDLE**)  AllocateZeroPool(sizeof(EFI_HANDLE*)   * hcount +
                                               dplen + strlen(LdrPath) + 1);
   if (!BlockDev || !BlockHdl) return 0;
   // cvt path to asc
   apath  = (char*)BlockHdl + sizeof(EFI_HANDLE*) * hcount;
   strcpy(apath, DevPath);
   etab->startdir = (u32t)(u64t)apath;
   apath += dplen;
   strcpy(apath, LdrPath);
   etab->ldrfname = (u32t)(u64t)apath;

   BlockDevCount = 0;

   // enum BLOCK_IO devices
   for(ii=0; ii<hcount; ii++) {
      EFI_BLOCK_IO_MEDIA  *Media;
      struct qs_diskinfo     *di = qdd + MAX_QS_FLOPPY + BlockDevCount;
      // Get the BLOCK_IO Protocol Interface to the device
      st = BS->HandleProtocol(handles[ii], &BlockIoProtocol, (VOID*)&BlkIo);
      // Handle does not support the BLOCK_IO protocol
      if (EFI_ERROR(st) || BlkIo==0) continue;

      Media = BlkIo->Media;

      if (!Media->LogicalPartition && Media->MediaPresent && Media->BlockSize>=512 &&
         Media->BlockSize<=4096)
      {
         RtZeroMem(di, sizeof(struct qs_diskinfo));
         di->qd_sectors     = Media->LastBlock+1;

         di->qd_sectorsize  = Media->BlockSize;
         // bit shift must be possible (i.e. only 512, 1024, 2048 & 4096 sector size)
         di->qd_sectorshift = bsf64(Media->BlockSize);
         if (1<<di->qd_sectorshift != Media->BlockSize) continue;

         di->qd_biosdisk    = 0x80 + (u8t)BlockDevCount;
         di->qd_mediatype   = Media->ReadOnly ? 2 : 3;
         di->qd_flags       = HDDF_LBAON|HDDF_PRESENT|HDDF_LBAPRESENT;
         if (Media->RemovableMedia) di->qd_flags|=HDDF_REMOVABLE;
         // CHS dances
         di->qd_cyls        = 1024;
         di->qd_heads       = 31;
         di->qd_spt         = 17;
         while (di->qd_cyls*di->qd_spt*di->qd_heads < di->qd_sectors) {
            if (di->qd_spt<63) {
               di->qd_spt   = (di->qd_spt   + 1 & 0xF0) * 2 - 1; continue; 
            }
            if (di->qd_heads<255) {
               di->qd_heads = (di->qd_heads + 1 & 0xF0) * 2 - 1; continue; 
            }
            break;
         }
         di->qd_cyls        = (u32t)(di->qd_sectors / (di->qd_spt*di->qd_heads));
         di->qd_chsheads    = di->qd_heads;
         di->qd_chsspt      = di->qd_spt;

         BlockDev[BlockDevCount] = BlkIo;
         BlockHdl[BlockDevCount] = handles[ii];
         // disk limit reached
         if (++BlockDevCount == MAX_QS_DISK-MAX_QS_FLOPPY) break;
      }
   }
   // Free the resources allocated from pool.
   FreePool(handles);

   etab->qd_hdds = BlockDevCount;


   st = BS->HandleProtocol (LoadedImage->DeviceHandle, &BlockIoProtocol, (VOID*)&BlkIo);
   if (!EFI_ERROR(st))
      get_partition(LoadedImage->DeviceHandle, &etab->bootdisk, &etab->bootstart, &etab->bootlen);

   return BlockDevCount;
}

// result is reversed in comparition with common practice, i.e. 0 - success, 1 - error
u32t dsk_io(struct qs_diskinfo *di, u64t start, u32t sectors, u32t writeflag) {
   u32t  idx;
   // check for all requirements
   if (!di || !sectors || !BlockDev || !Header32) return 1;

   idx = di->qd_biosdisk & 0x7F;
   //log_it(2, "dsk_io(%u,%LX,%u,%u)\n", idx, start, sectors, writeflag);
   //if (writeflag) return 1;

   while (idx<BlockDevCount) {
      EFI_BLOCK_IO_MEDIA  *Media = BlockDev[idx]->Media;
      UINT32             MediaId = Media->MediaId;
      UINTN               iosize = sectors<<di->qd_sectorshift;
      EFI_STATUS   st;
      u32t         rc;

      if (iosize>DISKBUF_SIZE) break;

      st = (writeflag ? BlockDev[idx]->WriteBlocks : BlockDev[idx]->ReadBlocks)
         (BlockDev[idx], MediaId, start, iosize, (void*)(u64t)Header32->DiskBufPM);
      rc = EFI_ERROR(st)?1:0;
      // call Flush() on removable
      if (rc && writeflag && Media->RemovableMedia)
         BlockDev[idx]->FlushBlocks(BlockDev[idx]);

      return rc;
   }
   return 1;
}

void dsk_term(void) {
   if (BlockDev) { FreePool(BlockDev); BlockDev = 0; }
   if (BlockHdl) { FreePool(BlockHdl); BlockHdl = 0; }
}

static EFI_FILE_IO_INTERFACE    *Vol = 0;
static EFI_FILE_HANDLE        FsRoot = 0;
static EFI_FILE_HANDLE    FileHandle = 0;

u16t mfs_open (u32t plen) {
   static CHAR16    FName[132 + MFSD_NAME_LEN];
   EFI_STATUS      Status;
   UINTN           ii, jj;
   u32t        *PFileSize = (u32t*)(u64t)plen;

   *PFileSize = FFFF;

   // first time use - init boot partition volume
   if (!Vol) {
      // Open the volume for the device where the EFI OS Loader was loaded from.
      Status = BS->HandleProtocol (LoadedImage->DeviceHandle, &FileSystemProtocol, (VOID*)&Vol);
      if (EFI_ERROR(Status)) { Vol = 0; return 1; }

      //DumpHandleProtocols("Loaded image:", LoadedImage->DeviceHandle);

      Status = Vol->OpenVolume (Vol, &FsRoot);
      if (EFI_ERROR(Status)) { Vol = 0; return 2; }

      if (!DevPath || strlen(DevPath)>128) { Vol = 0; return 3; }
   }
   // if file asking with lead '/' - assume direct path, else \EFI\BOOT dir
   if (mfsd_openname[0]!='/' && mfsd_openname[0]!='\\') {
      atow(DevPath, FName, 0);
      ii = wcslen(FName);
      FName[ii++] = '\\';
   } else
      ii = 0;
   jj = 0;
   // char -> char16
   while ((FName[ii++] = mfsd_openname[jj++]));

   Status = FsRoot->Open(FsRoot, &FileHandle, FName, EFI_FILE_MODE_READ, 0);
   if (EFI_ERROR(Status)) {
      log_it(3, "cannot open \"%S\"\n", FName);
      return 4;
   } else {
      static u8t buffer[sizeof(EFI_FILE_INFO)+512];
      EFI_FILE_INFO *fi = (EFI_FILE_INFO*)&buffer;
      UINTN       bsize = sizeof(EFI_FILE_INFO)+512;
      u16t           rc = 0;

      fi->FileSize = 0;
      Status = FileHandle->GetInfo(FileHandle, &GenericFileInfo, &bsize, fi);

      if (EFI_ERROR(Status)) rc = 5; else
         if (fi->FileSize>=_1GB) rc = 6; else
            *PFileSize = (u32t)fi->FileSize;
      // Print(L"File %s, rc %d, size %d\n", FName, (u32t)rc, *PFileSize);
      if (rc) {
         FileHandle->Close(FileHandle);
         FileHandle = 0;
      }
      return rc;
   }
}

u32t mfs_read (u32t offset, u32t buf, u32t readsize) {
   if (FileHandle) {
      u8t    *bufptr = (u8t*)(u64t)buf;
      UINTN   rdsize = readsize;
      EFI_STATUS  st = FileHandle->SetPosition(FileHandle, offset);
      if (EFI_ERROR(st)) return 0;

      st = FileHandle->Read(FileHandle, &rdsize, bufptr);
      if (EFI_ERROR(st)) return 0;

      return (u32t)rdsize;
   }
   return 0;
}

u32t mfs_close(void) {
   if (FileHandle) {
      FileHandle->Close(FileHandle);
      FileHandle = 0;
   }
   return 0;
}

u32t mfs_term(void) {
   mfs_close();
   // trying to be polite and close _all_ handles
   if (FsRoot) { FsRoot->Close(FsRoot); FsRoot = 0; }
   Vol = 0;
   return 0;
}
