//
// QSINIT
// 64-bit EFI loader
//
#ifndef QS_EFI_LOADER
#define QS_EFI_LOADER

#pragma warning(disable:4554)  // disable warning about shift presedence

#include "qsrtq.h"
#include "efnlist.h"
#include "qsexp.h"
#include "cpudef.h"
#include "qsinit.h"

/* key_status() bits, just copied here from QSINIT`s vio.h, because qsinit, 
   at last, able to receive only what we know here ;) */
#define KEY_SHIFT       0x000001
#define KEY_ALT         0x000002
#define KEY_CTRL        0x000004
#define KEY_NUMLOCK     0x000008
#define KEY_CAPSLOCK    0x000010
#define KEY_SCROLLLOCK  0x000020
#define KEY_SHIFTLEFT   0x000040
#define KEY_SHIFTRIGHT  0x000080
#define KEY_CTRLLEFT    0x000100
#define KEY_CTRLRIGHT   0x000200
#define KEY_ALTLEFT     0x000400
#define KEY_ALTRIGHT    0x000800
#define KEY_SYSREQ      0x001000
#define KEY_INSERT      0x002000

struct qs_diskinfo;

void   key_init(void);
u16t   key_read(void);
u16t   key_wait(u32t seconds, u32t *status);
u32t   key_status(void);
u8t    key_pressed(void);
u16t   tm_calibrate(void);
u32t   tm_counter(void);
u64t   tm_getdate(void);
void   tm_setdate(u32t dostime);
u16t   mfs_open (u32t plen);
u32t   mfs_read (u32t offset, u32t buf, u32t readsize);
u32t   mfs_close(void);
u32t   mfs_term (void);
u32t   vio_charout(u8t ch);
u32t   vio_strout(const char *str);
void   vio_clearscr(void);
void   vio_setpos(u32t line, u32t col);
void   vio_getpos(u32t *line, u32t *col);
void   vio_setcolor(u16t color);
u8t    vio_getmode(u32t *cols, u32t *lines);
void   vio_init(void);
void   vio_setmode(u32t lines);
void   vio_conlock(int lock);
int    vio_setmodeex(u32t cols, u32t lines);
void   vio_showcursor(u32t state);
u32t   vio_enummodes(struct console_mode **mi);
u32t   grf_enummodes(struct graphic_mode **mi, u64t *vmemptr, u64t *vmemsize);
u32t   grf_setmode(u32t mode);
u32t   grf_blit(u32t x, u32t y, u32t dx, u32t dy, void *buf, u32t pitch);
u32t   grf_clear(u32t x, u32t y, u32t dx, u32t dy, u32t color);
void   vio_term(void);
u32t   dsk_init(exp32_table *etab, struct qs_diskinfo *qdd);
u32t   dsk_io(struct qs_diskinfo *di, u64t start, u32t sectors, u32t writeflag);
void   dsk_term(void);
void   int15mem(void *Buf, u32t exitbootservices, u32t *ldrmem, u32t *ldrsize);
void   exit_poweroff(u32t actiontype);
u64t   hlp_setreg(u32t idx, u64t value);
void   hlp_memmove(void *dst, void *src, u64t length, u32t usecli);
void   hlp_memset (void *dst, u64t value, u64t length, u32t valsize);

int    Load32(void);
u32t   Exec32(void);
void   ErrorExit(const char *msg);

void   IntsOn(void);
void   IntsOff(void);
u32t   IntsEnabled(void);
u16t   GetCodeSelector(void);
u16t   GetTaskReg(void);

int    SysConfig(void);
int    VioConfig(void);
void   DumpConfigTable(void);

u32t   XcptOn(void);
u32t   TimerIrqsOn(u32t* apicaddr, u8t tmr_vector, u8t spr_vector);
u32t   LoadExitModule(char *module, char *args, u32t modtype);

u64t   ReadTSC(void);
void   CpuHlt(void);

void   ReplaceGDT(void *buffer);

/** dump handle protocols to QSINIT log
    @param info    short header info (without EOL)
    @param eh      handle */
void   DumpHandleProtocols(const char *info, EFI_HANDLE *eh);
int    DevPathToStr(char *buf, u32t buflen, EFI_DEVICE_PATH_PROTOCOL *DevPath);
void   GuidToStr (char *buf, EFI_GUID *guid);

EFI_STATUS OpenProtocolByHandle(EFI_HANDLE handle, EFI_GUID *protocol,
                                void **interface);

/// stricmp for A..Z chars
int    RtStrICmp (const CHAR16 *s1, const CHAR16 *s2);
/// strnicmp for A..Z chars
int    RtStrNICmp(const CHAR16 *s1, const CHAR16 *s2, UINTN len);

CHAR16*strto16   (const char *src, UINTN maxlen);
void   atow      (const char *src, CHAR16 *dst, UINTN maxlen);

extern exp32_table          *Header32;
extern u32t                    *lapic;
extern EFI_HANDLE             SelfImg;
extern UINT32                SafeMode;  // "safe mode" active (no exceptions, etc...)
extern u8t                  BootMount;
extern UINT32                  Mem640;  // 1Mb actual memory size
extern EFI_LOADED_IMAGE  *LoadedImage;
extern char                 LdiName[];
extern char                 IniName[];

#endif // QS_EFI_LOADER
