//
// QSINIT
// 64-bit runtime functions
//
#ifndef QS_RTF64
#define QS_RTF64

#include "qstypes.h"
#include "qserr.h"
#include "efi.h"

int       strnicmp  (const char *s1, const char *s2, u32t len);
char*     strcpy    (char *dst, const char *src);
u64t      strlen    (const char *s);

CHAR16*   wcscpy    (CHAR16 *dst, const CHAR16 *src);
CHAR16*   wcscat    (CHAR16 *dst, const CHAR16 *src);
u64t      wcslen    (const CHAR16 *str);

// don`t mix with cl embedded _bittest64(u64t*,u32t) and _bittest()
u32t      _bittestq (u64t src, u32t bit);
u32t      _bittestd (u32t src, u32t bit);
#define _bittest _bittestd
u32t      _bitsetq  (u64t *dst, u32t bit);
u32t      _bitset   (u32t *dst, u32t bit);
u32t      _bitresetq(u64t *dst, u32t bit);
u32t      _bitreset (u32t *dst, u32t bit);

s32t      bsf64     (u64t value);
s32t      bsr64     (u64t value);

u8t*      memchr    (u8t*mem, u8t chr, u64t buflen);
u8t*      memchrnb  (u8t*mem, u8t chr, u64t buflen);
u16t*     memchrw   (u16t*mem, u16t chr, u64t buflen);
u16t*     memchrnw  (u16t*mem, u16t chr, u64t buflen);
u32t*     memchrd   (u32t*mem, u32t chr, u64t buflen);
u32t*     memchrnd  (u32t*mem, u32t chr, u64t buflen);
u64t*     memchrq   (u64t*mem, u64t chr, u64t buflen);
u64t*     memchrnq  (u64t*mem, u64t chr, u64t buflen);

void      memset    (void *dst, int c, u64t length);
void      memsetw   (u16t *dst, u16t value, u64t length);
void      memsetd   (u32t *dst, u32t value, u64t length);
void      memsetq   (u64t *dst, u64t value, u64t length);

/** push string to the QSINIT log
    @param level   log level = 0..3
    @param fmt     format - in EFI Lib Print() syntax */
int       log_it    (int level, const char *fmt, ...);

/** normal snprint.
    This is QSINIT printf code, but w/o floaing point; %p is 64-bit,
    but 64-bit integers are still only via %Lu/%Ld/%LX */
int       snprintf  (char *buf, u32t count, const char *fmt, ...);
// note that target is the default EFI console, not QSINIT console!
int       printf    (const char *fmt, ...);

/// allocate pages
ptrdiff_t page_alloc(u32t npages);
/// free pages
u64t      page_free (ptrdiff_t addr, u32t npages);
/** allocate pages.
    @param bound           type of allocation (limited by memory address or
                           not, see PGALLOCEX_* in efnlist.inc)
    @param [in,out] addr   wanted address on input, result on exit
    @param pages           number of pages to allocate
    @return UEFI error code */
u64t      page_getex(u32t bound, ptrdiff_t *addr, u32t pages);

#endif // QS_RTF64
