;
; QSINIT EFI loader
; 32<->64 switching and minor code
;
                extrn   Ret64Addr:fword                         ;
                extrn   Init32Addr:fword                        ;
                extrn   Stack32  :fword                         ;
                extrn   FuncTable:qword                         ;
                extrn   FuncCount:dword                         ; # of entries

                .data
SaveRSP         dq      0
SaveSS          dw      0
SaveDS          dw      0
SaveES          dw      0
SaveFS          dw      0
SaveGS          dw      0

                .code


;----------------------------------------------------------------
;u16t GetCodeSelector(void);
GetCodeSelector proc    public
                mov     rax, cs                                 ;
                ret                                             ;
GetCodeSelector endp
;----------------------------------------------------------------
;u16t GetTaskReg(void);
GetTaskReg      proc    public
                str     ax
                ret
GetTaskReg      endp
;----------------------------------------------------------------
ReadGDTR        proc    public
                sgdt    fword ptr [rcx]                         ;
                ret                                             ;
ReadGDTR        endp

WriteGDTR       proc    public
ifdef __JWASM__
                lgdt    [rcx]                                   ;
else
                lgdt    fword ptr [rcx]                         ;
endif
                ret                                             ;
WriteGDTR       endp
;----------------------------------------------------------------
ReadIDTR        proc    public
                sidt    fword ptr [rcx]                         ;
                ret                                             ;
ReadIDTR        endp
;----------------------------------------------------------------
IntsOff         proc    public
                cli                                             ;
                ret                                             ;
IntsOff         endp

IntsOn          proc    public
                sti                                             ;
                ret                                             ;
IntsOn          endp
;----------------------------------------------------------------
;u32t IntsEnabled(void);
IntsEnabled     proc    public                                  ;
                pushfq                                          ;
                pop     rax                                     ;
                shr     eax,9                                   ;
                and     eax,1                                   ;
                ret                                             ;
IntsEnabled     endp                                            ;
;----------------------------------------------------------------
;u64t hlp_setreg(u32t idx, u64t value);
hlp_setreg      proc    public
                xor     rax,rax                                 ; accept CR2..CR4
                and     rcx,0Fh                                 ; and CR8 now
                cmp     cl,2                                    ;
                jc      @@hsetr_exit                            ;
                cmp     cl,8                                    ;
                jz      @@hsetr_process                         ;
                cmp     cl,5                                    ;
                jnc     @@hsetr_exit                            ;
@@hsetr_process:
                mov     rax,rdx                                 ;
                sub     cl,2                                    ;
                inc     rax                                     ; -1?
                jrcxz   @@hsetr_cr2                             ;
                loop    @@hsetr_01                              ; using loop as
                mov     rax,cr3                                 ; dec here to leave
                jz      @@hsetr_exit                            ; ZF unchanged
                mov     cr3,rdx                                 ;
                jmp     @@hsetr_exit                            ;
@@hsetr_01:
                loop    @@hsetr_02                              ;
                mov     rax,cr4                                 ;
                jz      @@hsetr_exit                            ;
                mov     cr4,rdx                                 ;
                jmp     @@hsetr_exit                            ;
@@hsetr_02:
                mov     rax,cr8                                 ;
                jz      @@hsetr_exit                            ;
                mov     cr8,rdx                                 ;
                jmp     @@hsetr_exit                            ;
@@hsetr_cr2:
                mov     rax,cr2                                 ;
                jz      @@hsetr_exit                            ;
                mov     cr2,rdx                                 ;
@@hsetr_exit:
                ret                                             ;
hlp_setreg      endp
;----------------------------------------------------------------
;void hlp_memmove(void *dst, void *src, u64t length, u32t usecli);
hlp_memmove     proc    public
                push    rdi                                     ;
                push    rsi                                     ;
                pushfq                                          ;
                mov     rdi,rcx                                 ;
                mov     rcx,r8                                  ;
                jrcxz   @@move0                                 ;
                or      r9,r9                                   ;
                jz      @@move_nocli                            ;
                cli                                             ;
@@move_nocli:
                mov     rsi,rdx                                 ;
                cmp     rsi,rdi                                 ;
                jnc     @@move_forward                          ;
                lea     rax,[rsi+rcx]                           ;
                cmp     rdi,rax                                 ;
                jnc     @@move_forward                          ;
                std                                             ; move backward
                add     rsi,rcx                                 ;
                add     rdi,rcx                                 ;
                mov     rax,rcx                                 ;
                and     rcx,7                                   ;
                dec     rsi                                     ;
                dec     rdi                                     ;
            rep movsb                                           ;
                mov     rcx,rax                                 ;
                shr     rcx,3                                   ;
                jz      @@move0                                 ;
                sub     rsi,7                                   ;
                sub     rdi,7                                   ;
            rep movsq                                           ;
                jmp     @@move0                                 ;
@@move_forward:
                cld                                             ; move forward
@@move_align:
                test    rdi,7                                   ; make sure data is aligned
                jz      @@move_aligned                          ;
                movsb                                           ;
                dec     rcx                                     ;
                jz      @@move0                                 ;
                jmp     @@move_align                            ;
@@move_aligned:
                mov     rax,rcx                                 ;
                shr     rcx,3                                   ;
                and     al,7                                    ;
            rep movsq                                           ;
                mov     cl,al                                   ;
            rep movsb                                           ;
@@move0:
                popfq                                           ;
                pop     rsi                                     ;
                pop     rdi                                     ;
                cld                                             ;
                xor     rax,rax                                 ;
                ret                                             ;
hlp_memmove     endp                                            ;

;----------------------------------------------------------------
SaveSegRegs     proc    public                                  ; this is 
                mov     SaveDS,ds                               ; REALLY
                mov     SaveES,es                               ; required
                mov     SaveFS,fs                               ;
                mov     SaveGS,gs                               ;
                ret                                             ;
SaveSegRegs     endp                                            ;

RestSegRegs     proc    public
                mov     ds,SaveDS                               ;
                mov     es,SaveES                               ;
                mov     fs,SaveFS                               ;
                mov     gs,SaveGS                               ;
                ret                                             ;
RestSegRegs     endp
;----------------------------------------------------------------
IODelayLoop     proc    public                                  ;
                call    ReadTSC                                 ; rax - start counter
                mov     ecx,20000                               ;
                align   4                                       ;
@@delayloop:
                dec     ecx                                     ;
                jnz     @@delayloop                             ;
                mov     rcx,rax                                 ;
                call    ReadTSC                                 ;
                sub     rax,rcx                                 ;
                ret                                             ;
IODelayLoop     endp                                            ;

ReadTSC         proc    public                                  ;
                rdtsc                                           ;
                shl     rax,32                                  ; edx:eax into
                shrd    rax,rdx,32                              ; rax
                ret                                             ;
ReadTSC         endp

CpuHlt          proc    public                                  ;
                sti                                             ;
                hlt                                             ;
                ret                                             ;
CpuHlt          endp                                            ;

                public  Ret32Point
Exec32          proc    public
                push    rbx                                     ;
                push    rsi                                     ;
                push    rdi                                     ;
                push    r12                                     ;
                push    r13                                     ;
                push    r14                                     ;
                push    r15                                     ;
                push    rbp                                     ;
                mov     SaveRSP,rsp                             ;
                mov     SaveSS,ss                               ;
                xor     rdx,rdx                                 ;
                mov     edx,dword ptr Stack32                   ;
                mov     ss,word ptr Stack32+4                   ;
                mov     rsp,rdx                                 ;
                push    rax                                     ; save rsp
                cli                                             ; YES!
                jmp     fword ptr [Init32Addr]                  ;
Ret32Point      label   byte                                    ;
                mov     ss,SaveSS                               ;
                mov     rsp,SaveRSP                             ; restore stack
                call    RestSegRegs                             ;
                pop     rbp                                     ;
                pop     r15                                     ;
                pop     r14                                     ;
                pop     r13                                     ;
                pop     r12                                     ;
                pop     rdi                                     ;
                pop     rsi                                     ;
                pop     rbx                                     ;
                sti                                             ;
                ret                                             ;
Exec32          endp                                            ;


; callback from 32-bit code
;----------------------------------------------------------------
; eax - function, stack - parameters in 64-bit calling conv.
;
Callback32      proc    public
                xor     rcx,rcx                                 ;
                mov     fs,rcx                                  ;
                mov     gs,rcx                                  ;
                dec     ecx                                     ; just make sure
                and     rsp,rcx                                 ;
;                call    RestSegRegs                             ;
;                mov     ss,SaveSS                               ; try to use loader ds/es/ss
                sti                                             ; they are FLAT, so it looks safe

                mov     rdx,offset FuncCount                    ;
                cmp     eax,[rdx]                               ;
                jnc     @@cb32_erridx                           ;
                cdqe                                            ;
                mov     rdx,offset FuncTable                    ;
                mov     rax,[rdx+rax*8]                         ;

                mov     rcx,[rsp]                               ; get top 4
                mov     rdx,[rsp+8]                             ; parameters from
                mov     r8,[rsp+16]                             ; stack
                mov     r9,[rsp+24]                             ;

;                clts                                           ; it called already in 32-bit
                                                                ; code after cli
                call    rax                                     ;
@@cb32_ret:
                mov     rdx,rax                                 ; make edx:eax
                shr     rdx,32                                  ; 64-bit result
                cli                                             ;
                mov     ss,word ptr Stack32+4                   ;
                jmp     fword ptr [Ret64Addr]                   ; go back home! :)
@@cb32_erridx:
                xor     rax,rax
                dec     rax
                jmp     @@cb32_ret

Callback32      endp


                end
